/*
 * (c) 2025 Jens Mueller
 *
 * JBASICODE
 *
 * Application Start
 */

package jbasicode;

import java.awt.EventQueue;
import java.io.BufferedInputStream;
import java.io.InputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.PrintStream;
import java.text.MessageFormat;
import java.text.ParseException;
import java.util.Properties;
import java.util.ResourceBundle;
import javax.swing.UIManager;
import jbasicode.ui.UIUtil;
import jbasicode.ui.MainFrm;


public class Main
{
  public static final String APPNAME     = "JBASICODE";
  public static final String VERSION     = "1.1";
  public static final String COPYRIGHT   = "(c) 2025 Jens M\u00FCller";
  public static final String PROP_PREFIX = "jbasicode.";

  private static final String SETTINGS_DIR  = "jbasicode";
  private static final String SETTINGS_FILE = "jbc_config.xml";
  private static final String PRG_PATH_FILE = "recent_prg_path.txt";

  private static Properties     properties     = new Properties();
  private static ResourceBundle resourceBundle = null;
  private static File           settingsDir    = null;
  private static File           settingsFile   = null;
  private static File           prgPathFile    = null;


  public static void createSettingsDirIfNotExists()
  {
    if( settingsDir != null ) {
      if( !settingsDir.exists() ) {
	settingsDir.mkdirs();
      }
    }
  }


  public static String emptyToNull( String text )
  {
    if( text != null ) {
      if( text.isEmpty() ) {
	text = null;
      }
    }
    return text;
  }


  public static boolean getBooleanProperty(
				String  propKey,
				boolean defaultValue )
  {
    String text = getProperty( propKey );
    return text != null ? Boolean.parseBoolean( text.trim() ) : defaultValue;
  }


  public static int getIntProperty( String propKey, int defaultValue )
  {
    int    rv   = defaultValue;
    String text = getProperty( propKey );
    if( text != null ) {
      try {
	rv = Integer.parseInt( text.trim() );
      }
      catch( NumberFormatException ex ) {}
    }
    return rv;
  }


  public static File getProgramPathFile()
  {
    return prgPathFile;
  }


  public static String getProperty( String propKey )
  {
    return properties.getProperty( PROP_PREFIX + propKey );
  }


  public static Properties getProperties()
  {
    return properties;
  }


  public static File getSettingsDir()
  {
    return settingsDir;
  }


  public static File getSettingsFile()
  {
    return settingsFile;
  }


  public static String getText( String textKey )
  {
    String rv = textKey;
    if( resourceBundle != null ) {
      try {
	rv = resourceBundle.getString( textKey );
      }
      catch( Exception ex ) {
	ex.printStackTrace( System.err );
      }
    }
    return rv;
  }


  public static void main( String[] args )
  {
    // load resources
    try {
      resourceBundle = ResourceBundle.getBundle( "resources.jbasicode" );
    }
    catch( Exception ex ) {
      ex.printStackTrace( System.err );
    }

    // parse arguments
    boolean helpFlag    = false;
    boolean runFlag     = false;
    boolean versionFlag = false;
    String  prgFileName = null;
    try {
      for( String arg : args ) {
	if( arg.startsWith( "-" ) ) {
	  if( arg.equalsIgnoreCase( "-H" ) ) {
	    helpFlag = true;
	  } else if( arg.equalsIgnoreCase( "-R" ) ) {
	    runFlag = true;
	  } else if( arg.equalsIgnoreCase( "-V" ) ) {
	    versionFlag = true;
	  } else {
	    throw new ParseException(
			MessageFormat.format(
				getText( "error.unknown_option" ),
				arg ),
			0 );
	  }
	} else {
	  if( prgFileName != null ) {
	    throw new ParseException(
			Main.getText( "error.more_than_one_file_specified" ),
			0 );
	  }
	  prgFileName = arg;
	}
      }
      if( helpFlag ) {
	printVersion( System.out );
	printUsage( System.out );
	System.exit( 0 );
      } else if( versionFlag ) {
	printVersion( System.out );
	System.exit( 0 );
      }
      if( runFlag && (prgFileName == null) ) {
	throw new ParseException(
			Main.getText( "error.run_flag_without_file" ),
			0 );
      }
    }
    catch( ParseException ex ) {
      printVersion( System.err );
      System.err.println();
      System.err.println( ex.getMessage() );
      printUsage( System.err );
      System.exit( -1 );
    }

    // read settings file
    Exception propEx     = null;
    String    appDataDir = emptyToNull( System.getenv( "APPDATA" ) );
    if( appDataDir != null ) {
      settingsDir = new File( appDataDir, SETTINGS_DIR );
    } else {
      String homeDir = emptyToNull( System.getProperty( "user.home" ) );
      if( homeDir != null ) {
	settingsDir = File.separatorChar == '/' ?
				new File( homeDir, "." + SETTINGS_DIR )
				: new File( homeDir, SETTINGS_DIR );
      } else {
	settingsDir = File.separatorChar == '/' ?
				new File( homeDir, "." + SETTINGS_DIR )
				: new File( homeDir, SETTINGS_DIR );
      }
    }
    if( settingsDir != null ) {
      settingsFile = new File( settingsDir, SETTINGS_FILE );
      prgPathFile  = new File( settingsDir, PRG_PATH_FILE );
    }
    if( settingsFile != null ) {
      if( settingsFile.exists() ) {
	try( InputStream in = new BufferedInputStream(
				new FileInputStream( settingsFile ) ) )
	{
	  properties.loadFromXML( in );
	}
	catch( Exception ex ) {
	  propEx = ex;
	}
      }
    }

    // Look&Feel
    try {
      UIManager.put( "swing.boldMetal", Boolean.FALSE );
      UIManager.setLookAndFeel( UIManager.getSystemLookAndFeelClassName() );
    }
    catch( Exception ex ) {}

    // possibly show error message
    if( (settingsFile != null) && (propEx != null) ) {
      StringBuilder buf = new StringBuilder( 256 );
      buf.append( settingsFile.getPath() );
      buf.append( ":\n" );
      buf.append( getText( "error.read_settings" ) );
      String msg = emptyToNull( propEx.getMessage() );
      if( msg != null ) {
	buf.append( '\n' );
	buf.append( msg );
      }
      UIUtil.fireShowErrorMsg( null, buf.toString() );
    }

    // open main window
    final String  prgFileName1 = prgFileName;
    final boolean runFlag1     = runFlag;
    EventQueue.invokeLater(
		()->(new MainFrm(
				prgFileName1,
				runFlag1 )).setVisible( true ) );
  }


  public static void setProperty( String propKey, boolean value )
  {
    setProperty( propKey, String.valueOf( value ) );
  }


  public static void setProperty( String propKey, int value )
  {
    setProperty( propKey, String.valueOf( value ) );
  }


  public static void setProperty( String propKey, String value )
  {
    properties.setProperty(
			PROP_PREFIX + propKey,
			value != null ? value : "" );
  }


	/* --- Private Methods --- */

  private static void printUsage( PrintStream out )
  {
    out.println();
    out.println( Main.getText( "label.cmd_line" ) );
    out.println( "java -jar jbasicode.jar [-h] [-v] [[-r] file]" );
    out.println( "  -h:   " + Main.getText( "option.show_this_help" ) );
    out.println( "  -v:   " + Main.getText( "option.show_prg_version" ) );
    out.println( "  -r:   " + Main.getText( "option.run_prg" ) );
    out.println( "  file: " + Main.getText( "option.file_to_be_opened" ) );
  }


  private static void printVersion( PrintStream out )
  {
    out.println( APPNAME + "\u0020" + VERSION );
    out.println( COPYRIGHT );
  }


	/* --- Constructor --- */

  private Main()
  {
    // empty
  }
}
