/*
 * (c) 2025 Jens Mueller
 *
 * JBASICODE
 *
 * Create BASICODE Audio Data
 */

package jbasicode.bc;

import java.io.ByteArrayOutputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.StringReader;
import jbasicode.Main;
import jbasicode.etc.ExtByteArrayOutputStream;
import jbasicode.etc.IOUtil;


public class BCAudioCreator
{
  private static final int SAMPLE_HIGHT_VALUE = 0xE0;

  private int sampleRate;
  private int bit1halfWaveSamples;


  public BCAudioCreator( int sampleRate )
  {
    this.sampleRate          = sampleRate;
    this.bit1halfWaveSamples = sampleRate / 4800;
  }


  public void writeDataWAV( File file, InputStream in ) throws IOException
  {
    ByteArrayOutputStream aBuf = new ByteArrayOutputStream( 0x4000 );

    boolean status   = false;
    int     blockNum = 0;
    while( writeBlock( aBuf, 0x01, blockNum++, in, 1024 ) ) {
      status = true;
    }
    if( !status ) {
      throw new IOException( Main.getText( "error.no_content_to_save" ) );
    }
    writeWAV( file, aBuf );
  }


  public void writeProgramWAV(
			File   file,
			String prgText ) throws IOException
  {
    // convert characters to ASCII bytes with 0Dh as line end
    ExtByteArrayOutputStream bBuf = new ExtByteArrayOutputStream( 0x4000 );
    if( prgText != null ) {
      try( BufferedReader reader = new BufferedReader(
					new StringReader( prgText ) ); )
      {
	String line = reader.readLine();
	while( line != null ) {
	  boolean state = false;
	  int     len   = line.length();
	  for( int i = 0; i < len; i++ ) {
	    char ch = line.charAt( i );
	    if( (ch >= 0x04) && (ch <= 0xFF) ) {
	      bBuf.write( ch );
	      state = true;
	    }
	  }
	  if( state ) {
	    bBuf.write( 0x0D );
	  }
	  line = reader.readLine();
	}
      }
    }
    if( bBuf.size() == 0 ) {
      throw new IOException( Main.getText( "error.no_content_to_save" ) );
    }

    // create audio data
    ByteArrayOutputStream aBuf = new ByteArrayOutputStream( 0x4000 );
    writeBlock( aBuf, 0x02, -1, bBuf.createInputStream(), bBuf.size() );
    writeWAV( file, aBuf );
  }


	/* --- private methods --- */

  private static void writeASCII(
				OutputStream out,
				String       text ) throws IOException
  {
    int len = text.length();
    for( int i = 0; i < len; i++ ) {
      out.write( text.charAt( i ) );
    }
  }


  private void writeBit(
			OutputStream out,
			boolean      value  ) throws IOException
  {
    if( value ) {
      writeSamples( out, this.bit1halfWaveSamples, false );
      writeSamples( out, this.bit1halfWaveSamples, true );
      writeSamples( out, this.bit1halfWaveSamples, false );
      writeSamples( out, this.bit1halfWaveSamples, true );
    } else {
      writeSamples( out, this.bit1halfWaveSamples * 2, false );
      writeSamples( out, this.bit1halfWaveSamples * 2, true );
    }
  }


  private void writeByte(
			OutputStream out,
			int          value  ) throws IOException
  {
    writeBit( out, false );			// start bit
    for( int i = 0; i < 8; i++ ) {		// 8 data bits
      writeBit( out, (value & 0x01) != 0 );
      value >>= 1;
    }
    writeBit( out, true );			// 2 stop bits
    writeBit( out, true );
  }


  private boolean writeBlock(
			OutputStream out,
			int          blockType,
			int          blockNum,
			InputStream  in,
			int          blockSize ) throws IOException
  {
    boolean rv = false;
    int b = in.read();
    if( b >= 0 ) {

      // pulse tone
      for( int i = 0; i < 6000; i++ ) {
	writeBit( out, true );
      }

      // block type byte
      blockType ^= 0x80;
      int cks = blockType;
      writeByte( out, blockType );

      // block number
      if( blockNum >= 0 ) {
	blockNum ^= 0x80;
	cks ^= blockNum;
	writeByte( out, blockNum );
      }

      // file content
      b ^= 0x80;
      cks ^= b;
      writeByte( out, b );
      --blockSize;
      b = in.read();
      while( b >= 0 ) {
	b ^= 0x80;
	cks ^= b;
	writeByte( out, b );
	--blockSize;
	b = in.read();
      }

      // fill remaing block
      b = 0x84;
      while( blockSize > 0 ) {
	cks ^= b;
	writeByte( out, b );
	--blockSize;
      }

      // block end byte
      cks ^= 0x83;
      writeByte( out, 0x83 );

      // check byte
      writeByte( out, cks );

      // trailing pulse tone
      for( int i = 0; i < 1200; i++ ) {
	writeBit( out, true );
      }

      // trailing pause
      for( int i = 0; i < this.sampleRate; i++ ) {
	out.write( 0 );
      }
      rv = true;
    }
    return rv;
  }


  private static void writeInt2LE(
				OutputStream out,
				int          value ) throws IOException
  {
    out.write( value & 0xFF );
    out.write( (value >> 8) & 0xFF );
  }


  private static void writeInt4LE(
				OutputStream out,
				long         value ) throws IOException
  {
    writeInt2LE( out, (int) value );
    writeInt2LE( out, (int) (value >> 16) );
  }


  private void writeSamples(
			OutputStream out,
			int          count,
			boolean      value  ) throws IOException
  {
    for( int i = 0; i < count; i++ ) {
      out.write( value ? SAMPLE_HIGHT_VALUE : 0 );
    }
  }


  private void writeWAV(
			File                  file,
			ByteArrayOutputStream audioData ) throws IOException
  {
    BufferedOutputStream out = null;
    try {
      out = new BufferedOutputStream( new FileOutputStream( file ) );

      int audioDataLen = audioData.size();
      writeASCII( out, "RIFF" );
      writeInt4LE( out, audioDataLen + 36 );	// remaining file length
      writeASCII( out, "WAVEfmt\u0020" );
      writeInt4LE( out, 16 );			// remaining fmt block length
      writeInt2LE( out, 1 );			// pcm
      writeInt2LE( out, 1 );			// channels
      writeInt4LE( out, this.sampleRate );	// sample rate
      writeInt4LE( out, this.sampleRate );	// bytes per second
      writeInt2LE( out, 1 );			// frame size in bytes
      writeInt2LE( out, 8 );			// sample size in bits
      writeASCII( out, "data" );
      writeInt4LE( out, audioDataLen );		// remaining data length
      audioData.writeTo( out );
      out.close();
      out = null;
    }
    finally {
      IOUtil.closeSilently( out );
    }
  }
}
