/*
 * (c) 2025 Jens Mueller
 *
 * JBASICODE
 *
 * Runtime environment for execution of BASICODE programs
 */

package jbasicode.bc;

import java.awt.AWTError;
import java.awt.Toolkit;
import java.io.BufferedOutputStream;
import java.io.CharArrayWriter;
import java.io.Closeable;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.PushbackInputStream;
import java.text.CharacterIterator;
import java.text.MessageFormat;
import java.text.ParseException;
import java.text.StringCharacterIterator;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.Stack;
import java.util.TreeSet;
import javax.sound.midi.MidiChannel;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.MidiUnavailableException;
import javax.sound.midi.Synthesizer;
import jbasicode.Main;
import jbasicode.etc.IOUtil;
import jbasicode.ui.DebugFld;
import jbasicode.ui.MainFrm;
import jbasicode.ui.UIUtil;
import jbasicode.ui.VarView;


public class BCRuntime extends Thread
{
  public static final int     DEFAULT_SPEED                        = 10;
  public static final boolean DEFAULT_GOTO950_WITHOUT_CLS          = false;
  public static final boolean DEFAULT_WARN_REDECLARATION           = true;
  public static final boolean DEFAULT_WARN_VAR_NOT_INITIALIZED     = true;
  public static final boolean DEFAULT_WAIT_DURATION_WHEN_SOUND_OFF = false;

  public static final String PROP_DATA_FILE_EOR = "data_file.end_of_record";
  public static final String PROP_GOTO950_WITHOUT_CLS
					= "goto950_without_cls";
  public static final String PROP_SPEED = "speed";
  public static final String PROP_WAIT_DURATION_WHEN_SOUND_OFF
					= "wait_duration_when_sound_off";
  public static final String PROP_WARN_REDECLARATION
					= "warn.redeclaration";
  public static final String PROP_WARN_VAR_NOT_INITIALIZED
					= "warn.var_not_initialized";

  public static final Object PRINT_CONCAT_ITEM = new Object();
  public static final Object PRINT_TAB_ITEM    = new Object();

  public static final String VALUE_EOR_CR   = "\\r";
  public static final String VALUE_EOR_CRNL = "\\r\\n";
  public static final String VALUE_EOR_NL   = "\\n";

  public enum EOR { CR, CRNL, NL };


  private static final int MAX_STACK_DEPTH = 1000;


  private static class PrintTabItem
  {
    private Number value;

    public PrintTabItem( Number value )
    {
      this.value = value;
    }
  };


  private static class UserFunction
  {
    private String argName;
    private String funcExpr;

    private UserFunction( String argName, String funcExpr )
    {
      this.argName  = argName;
      this.funcExpr = funcExpr;
    }
  };


  private File recentDirFile = null;

  private MainFrm                  mainFrm;
  private volatile DebugFld        debugFld;
  private BCScreen                 bcScreen;
  private String                   prgText;
  private ProgramParser            prgParser;
  private Map<String,DimVar>       dimVars;
  private Map<String,Object>       simpleVars;
  private Map<String,UserFunction> functions;
  private Set<String>              calledFunctions;
  private Stack<BCStackItem>       stack;
  private Map<Integer,Closeable>   dataStreams;
  private EOR                      dataFileEOR;
  private volatile boolean         audioEnabled;
  private boolean                  midiAvailable;
  private MidiChannel              midiChannel;
  private Synthesizer              synthesizer;
  private StringBuffer             printer;
  private Random                   random;
  private Object                   waitMonitor;
  private int                      speedBrakeCounter;
  private volatile int             speedBrakeCntInit;
  private volatile boolean         warnRedeclaration;
  private volatile boolean         warnVarNotInitialized;


  public BCRuntime(
		MainFrm mainFrm,
		String  prgText,
		File    prgFile,
		boolean audioEnabled )
  {
    super( "BASICODE Executor" );
    this.mainFrm           = mainFrm;
    this.bcScreen          = mainFrm.getBCScreen();
    this.debugFld          = null;
    this.prgText           = prgText;
    this.prgParser         = new ProgramParser( mainFrm, this );
    this.dimVars           = new HashMap<>();
    this.simpleVars        = new HashMap<>();
    this.functions         = new HashMap<>();
    this.calledFunctions   = new TreeSet<>();
    this.stack             = new Stack<>();
    this.dataStreams       = new HashMap<>();
    this.dataFileEOR       = getEndOfRecord();
    this.audioEnabled      = audioEnabled;
    this.midiAvailable     = true;
    this.midiChannel       = null;
    this.synthesizer       = null;
    this.waitMonitor       = new Object();
    this.printer           = null;
    this.random            = new Random( System.currentTimeMillis() );
    this.speedBrakeCounter = 0;
    declareVarCC();
    if( (recentDirFile == null) && (prgFile != null) ) {
      recentDirFile = prgFile.getParentFile();
    }
    this.bcScreen.updScreenSize();
    setSpeed( getSpeedSetting() );
    settingsChanged();
  }


  public Object callUserFunction(
			String            funcName,
			ExprParser        exprParser,
			CharacterIterator iter ) throws ParseException
  {
    Object rv = null;
    if( this.calledFunctions.contains( funcName ) ) {
      throw new ParseException(
			ProgramParser.createMsg(
				"bc.illegal_recursive_call",
				funcName ),
			iter.getIndex() );
    }
    try {
      this.calledFunctions.add( funcName );

      UserFunction function = this.functions.get( funcName );
      if( function == null ) {
	throw new ParseException(
			ProgramParser.createMsg(
				"bc.function_not_found", funcName ),
			iter.getIndex() );
      }
      Object argValue = null;
      Object orgValue = null;
      if( function.argName != null ) {
	ExprParser.parseLeftParenthesis( iter );
	if( function.argName.endsWith( "$" ) ) {
	  argValue = exprParser.parseStringExpr( iter );
	} else {
	  argValue = exprParser.parseNumericExpr( iter );
	}
	ExprParser.parseRightParenthesis( iter );
	synchronized( this.simpleVars ) {
	  orgValue = this.simpleVars.get( function.argName );
	  this.simpleVars.put( function.argName, argValue );
	}
      }
      boolean warnEnabled = this.prgParser.getWarningsEnabled();
      try {
	this.prgParser.setWarningsEnabled( false );
	rv = exprParser.parseNumericExpr(
			new StringCharacterIterator( function.funcExpr ) );
      }
      catch( ParseException ex ) {
	throw new ParseException( ex.getMessage(), iter.getIndex() );
      }
      finally {
	this.prgParser.setWarningsEnabled( warnEnabled );
	if( function.argName != null ) {
	  synchronized( this.simpleVars ) {
	    if( orgValue != null ) {
	      this.simpleVars.put( function.argName, orgValue );
	    } else {
	      this.simpleVars.remove( function.argName );
	    }
	  }
	}
      }
    }
    finally {
      this.calledFunctions.remove( funcName );
    }
    return rv;
  }


  public void checkDebugAction(
			int        lineIdx,
			SourceLine srcLine,
			int        instrBegIdx ) throws InterruptedException
  {
    DebugFld debugFld = this.debugFld;
    if( debugFld != null ) {
      debugFld.checkDebugAction(
				lineIdx,
				srcLine,
				instrBegIdx,
				this.stack.size() );
    }
  }


  public void checkSpeedBrake() throws InterruptedException
  {
    int speedBrakeCntInit = this.speedBrakeCntInit;
    if( speedBrakeCntInit > 0 ) {
      if( this.speedBrakeCounter > 0 ) {
	--this.speedBrakeCounter;
      } else {
	Thread.sleep( 10 );
	this.speedBrakeCounter = speedBrakeCntInit;
      }
    }
  }


  public static Object createPrintTabItem( Number value )
  {
    return new PrintTabItem( value );
  }


  public String createStackItemsView()
  {
    StringBuilder buf = new StringBuilder( 256 );
    for( int i = this.stack.size() - 1; i >= 0; --i ) {
      int len = buf.length();
      if( len > 0 ) {
	buf.append( ", " );
      }
      if( len > 200 ) {
	buf.append( "..." );
	break;
      }
      buf.append( this.stack.get( i ) );
    }
    return buf.toString();
  }


  public List<VarView> createVarViewList()
  {
    List<VarView> list = new ArrayList<>();
    synchronized( this.simpleVars ) {
      for( String varName : this.simpleVars.keySet() ) {
	list.add( new VarView( varName, this.simpleVars.get( varName ) ) );
      }
    }
    synchronized( this.dimVars ) {
      for( DimVar dimVar : this.dimVars.values() ) {
	list.add( dimVar.createVarView() );
      }
    }
    try {
      Collections.sort( list );
    }
    catch( ClassCastException ex ) {}
    return list;
  }


  public void debugStopFired()
  {
    synchronized( this.waitMonitor ) {
      try {
	this.waitMonitor.notifyAll();
      }
      catch( IllegalMonitorStateException ex ) {}
    }
  }


  public void declareDimVar(
			String       varName,
			List<Number> dimensions,
			int          srcIndex ) throws ParseException
  {
    boolean exists = this.dimVars.containsKey( varName );
    switch( dimensions.size() ) {
      case 1:
	int dim = dimensions.get( 0 ).intValue();
	checkDimValue( dim, srcIndex );
	synchronized( this.dimVars ) {
	  this.dimVars.put( varName, new DimVar( varName, dim ) );
	}
	break;
      case 2:
	int dim1 = dimensions.get( 0 ).intValue();
	int dim2 = dimensions.get( 1 ).intValue();
	checkDimValue( dim1, srcIndex );
	checkDimValue( dim2, srcIndex );
	synchronized( this.dimVars ) {
	  this.dimVars.put( varName, new DimVar( varName, dim1, dim2 ) );
	}
	break;
      default:
	throw new ParseException(
			ProgramParser.createMsg(
				"bc.illegal_dimension_count",
				dimensions.size() ),
			srcIndex );
    }
    if( exists ) {
      putRedeclarationWarning( varName, srcIndex );
    }
  }


  public void declareUserFunction(
			String funcName,
			String argName,
			String funcExpr,
			int    srcIndex ) throws ParseException
  {
    if( this.functions.containsKey( funcName ) ) {
      putRedeclarationWarning( funcName, srcIndex );
    }
    synchronized( this.functions) {
      this.functions.put( funcName, new UserFunction( argName, funcExpr ) );
    }
  }


  public void doEnd() throws EndException
  {
    throw new EndException();
  }


  public void doFor( ForStackItem stackItem )
  {
    this.stack.push( stackItem );
  }


  public void doGosubGoto(
			int         bcLineNum,
			BCStackItem gosubStackItem,
			int         srcIndex ) throws
						EndException,
						InterruptedException,
						GotoException,
						JumpException,
						ParseException
  {
    if( (bcLineNum == 20) || (bcLineNum >= 1000) ) {
      if( gosubStackItem != null ) {
	if( this.stack.size() >= MAX_STACK_DEPTH ) {
	  throw new ParseException(
			Main.getText( "bc.stack_overflow" ),
			srcIndex );
	}
	this.stack.push( gosubStackItem );
      }
      if( bcLineNum >= 1000 ) {
	throw new GotoException( srcIndex, bcLineNum );
      }
      doGOTO20( srcIndex );	// init and jump to line 1010
    }
    switch( bcLineNum ) {
      case 20:			// init and jump to line 1010
	doGOTO20( srcIndex );
	break;
      case 100:			// set text mode and clear screen
	doGOSUB100( srcIndex );
	break;
      case 110:			// set cursor position
	doGOSUB110( srcIndex );
	break;
      case 120:			// get cursor position
	doGOSUB120();
	break;
      case 150:			// highlighted text output
	doGOSUB150( srcIndex );
	break;
      case 200:			// inkey
	doGOSUB200();
	break;
      case 210:			// inchar
	doGOSUB210();
	break;
      case 220:			// get character at screen position
	doGOSUB220( srcIndex );
	break;
      case 250:			// beep
	doGOSUB250( srcIndex );
	break;
      case 260:			// get a random number
	doGOSUB260();
	break;
      case 270:			// get free memory
	doGOSUB270();
	break;
      case 280:			// enable or disable stop key
	doGOSUB280( srcIndex );
	break;
      case 300:			// convert a number to string
	doGOSUB300( srcIndex );
	break;
      case 310:			// format a number
	doGOSUB310( srcIndex );
	break;
      case 330:			// convert string to uppercase
	doGOSUB330( srcIndex );
	break;
      case 350:			// printer output
	doGOSUB350( srcIndex );
	break;
      case 360:			// printer carriage return
	doGOSUB360();
	break;
      case 400:			// sound
	doGOSUB400( srcIndex );
	break;
      case 450:			// pause
	doGOSUB450( srcIndex );
	break;
      case 500:			// open file
	doGOSUB500( srcIndex );
	break;
      case 540:			// read file
	doGOSUB540( srcIndex );
	break;
      case 560:			// write file
	doGOSUB560( srcIndex );
	break;
      case 580:			// close file
	doGOSUB580( srcIndex );
	break;
      case 600:			// set graphics mode and clear screen
	doGOSUB600( srcIndex );
	break;
      case 620:			// set pixel in graphics mode
	doGOSUB620( srcIndex );
	break;
      case 630:			// draw line in graphics mode
	doGOSUB630( srcIndex );
	break;
      case 650:			// draw text in graphics mode
	doGOSUB650( srcIndex );
	break;
      case 950:			// end program
	doGOTO950();
	break;
    }
    if( gosubStackItem == null ) {
      doReturn( srcIndex );
    }
  }


  public void doInput(
		String       text,
		String       varName,
		List<Number> indices,
		int          srcIndex )
	throws EndException, InterruptedException, ParseException
  {
    if( text == null ) {
      text = "? ";
    } else {
      text += "\u0020";
    }
    this.bcScreen.print( text );
    for(;;) {
      String line = this.bcScreen.readLineFromKeyboard( this.debugFld );
      if( line.isEmpty() ) {
	/*
	 * no input -> keep old value in variable or,
	 * if variable does not exist,
	 * initialize it with default value
	 */
	if( !this.simpleVars.containsKey( varName ) ) {
	  synchronized( this.simpleVars ) {
	    this.simpleVars.put(
			varName,
			ExprParser.getDefaultVarValue( varName ) );
	  }
	}
	break;
      }
      if( varName.endsWith( "$" ) ) {
	setVarValue( varName, indices, line, srcIndex );
	break;
      }
      line         = line.trim();
      Number value = null;
      try {
	value = Integer.valueOf( line );
      }
      catch( NumberFormatException ex ) {}
      if( value == null ) {
	try {
	  value = Double.valueOf( line );
	}
	catch( NumberFormatException ex ) {}
      }
      if( value != null ) {
	setVarValue( varName, indices, value, srcIndex );
	break;
      }
      this.bcScreen.print( Main.getText( "bc.invalid_number" ) );
      this.bcScreen.println();
      this.bcScreen.print( "? " );
    }
  }


  public void doNext(
		String       varName,
		List<Number> indices,
		int          srcIndex ) throws JumpException, ParseException
  {
    if( this.stack.isEmpty() ) {
      throw new ParseException(
			Main.getText( "bc.next_without_for" ),
			srcIndex );
    }
    BCStackItem stackItem = this.stack.pop();
    if( !(stackItem instanceof ForStackItem) ) {
      throw new ParseException(
			Main.getText( "bc.return_without_gosub" ),
			srcIndex );
    }
    ForStackItem forStackItem = (ForStackItem) stackItem;
    String       orgVarName   = forStackItem.getVarName();
    List<Number> orgIndices   = forStackItem.getVarIndices();
    boolean      varMatches   = true;
    if( varName != null ) {
      varMatches = varName.equals( orgVarName );
      if( (indices != null) && (orgIndices != null) ) {
	int n = indices.size();
	if( n == orgIndices.size() ) {
	  for( int i = 0; i < n; i++ ) {
	    if( indices.get( i ).intValue()
				!= orgIndices.get( i ).intValue() )
	    {
	      varMatches = false;
	      break;
	    }
	  }
	} else {
	  varMatches = false;
	}
      } else if( (indices != null) && (orgIndices == null) ) {
	varMatches = false;
      } else if( (indices == null) && (orgIndices != null) ) {
	varMatches = false;
      }
      if( !varMatches ) {
	throw new ParseException(
			Main.getText( "bc.var_mismatch" ),
			srcIndex );
      }
    }
    boolean loop  = true;
    Number  vVar  = (Number) getVarValue(
					orgVarName,
					orgIndices,
					srcIndex );
    Number  vEnd  = forStackItem.getEndValue();
    Number  vStep = forStackItem.getStepSize();
    if( (vVar instanceof Integer)
	&& (vEnd instanceof Integer)
	&& (vStep instanceof Integer) )
    {
      vVar = Integer.valueOf( vVar.intValue() + vStep.intValue() );
      if( vStep.intValue() < 0 ) {
	if( vVar.intValue() < vEnd.intValue() ) {
	  loop = false;
	}
      } else {
	if( vVar.intValue() > vEnd.intValue() ) {
	  loop = false;
	}
      }
    } else {
      vVar = Double.valueOf( vVar.doubleValue() + vStep.doubleValue() );
      if( vStep.doubleValue() < 0.0 ) {
	if( vVar.doubleValue() < vEnd.doubleValue() ) {
	  loop = false;
	}
      } else {
	if( vVar.doubleValue() > vEnd.doubleValue() ) {
	  loop = false;
	}
      }
    }
    setVarValue( orgVarName, orgIndices, vVar, srcIndex );
    if( loop ) {
      this.stack.push( stackItem );
      throw new JumpException( stackItem );
    }
  }


  public void doReturn( int srcIndex ) throws JumpException, ParseException
  {
    if( this.stack.isEmpty() ) {
      throw new ParseException(
			Main.getText( "bc.return_without_gosub" ),
			srcIndex );
    }
    BCStackItem stackItem = this.stack.pop();
    if( stackItem instanceof GosubStackItem ) {
      throw new JumpException( stackItem );
    }
    throw new ParseException(
			Main.getText( "bc.return_without_gosub" ),
			srcIndex );
  }


  public void fireAbort()
  {
    this.prgParser.fireAbort();
    interrupt();
  }


  public static EOR getDefaultEndOfRecord()
  {
    return EOR.CR;
  }


  public static EOR getEndOfRecord()
  {
    EOR    rv   = getDefaultEndOfRecord();
    String text = Main.getProperty( PROP_DATA_FILE_EOR );
    if( text != null ) {
      switch( text ) {
	case VALUE_EOR_CR:
	  rv = EOR.CR;
	  break;
	case VALUE_EOR_CRNL:
	  rv = EOR.CRNL;
	  break;
	case VALUE_EOR_NL:
	  rv = EOR.NL;
	  break;
      }
    }
    return rv;
  }


  public ProgramParser getProgramParser()
  {
    return this.prgParser;
  }


  public String getProgramText()
  {
    return this.prgText;
  }


  public static int getSpeedSetting()
  {
    return Main.getIntProperty( PROP_SPEED, DEFAULT_SPEED );
  }


  /*
   * Retrieves the value of a simple and only simple variable,
   * If the variable does not exist it will be created with a default value.
   */
  public Object getVarValue( String varName, int srcIndex )
  {
    Object value = this.simpleVars.get( varName );
    if( value == null ) {
      value = ExprParser.getDefaultVarValue( varName );
      synchronized( this.simpleVars ) {
	this.simpleVars.put( varName, value );
      }
      putVarNotInitializedWarning( varName, srcIndex );
    }
    return value;
  }


  // Retrieves the value of a simple or dimensioned variable
  public Object getVarValue(
			String       varName,
			List<Number> indices,
			int          srcIndex ) throws ParseException
  {
    Object value = null;
    if( indices != null ) {
      if( indices.isEmpty() ) {
	value = getVarValue( varName, srcIndex );
      } else {
	value = getDimVar( varName, srcIndex ).getValue(
							this,
							indices,
							srcIndex );
      }
    } else {
      value = getVarValue( varName, srcIndex );
    }
    return value;
  }


  public static boolean getWaitDurationWhenSoundOffSetting()
  {
    return Main.getBooleanProperty(
			PROP_WAIT_DURATION_WHEN_SOUND_OFF,
			DEFAULT_WAIT_DURATION_WHEN_SOUND_OFF );
  }


  public static boolean getWarnRedeclarationSetting()
  {
    return Main.getBooleanProperty(
			PROP_WARN_REDECLARATION,
			DEFAULT_WARN_REDECLARATION );
  }


  public static boolean getWarnVarNotInitializedSetting()
  {
    return Main.getBooleanProperty(
			PROP_WARN_VAR_NOT_INITIALIZED,
			DEFAULT_WARN_VAR_NOT_INITIALIZED );
  }


  public void print( List<Object> items )
  {
    if( !this.bcScreen.isGraphicsMode() ) {
      boolean nl = true;
      for( Object o : items ) {
	if( o == PRINT_CONCAT_ITEM ) {
	  nl = false;
	} else if( o == PRINT_TAB_ITEM ) {
	  StringBuilder buf = new StringBuilder();
	  buf.append( '\u0020' );
	  int col = this.bcScreen.getCursorCol() + 1;
	  while( (col % 13) != 0 ) {
	    buf.append( '\u0020' );
	    col++;
	  }
	  this.bcScreen.print( buf );
	  nl = false;
	} else if( o instanceof PrintTabItem ) {
	  int col    = this.bcScreen.getCursorCol();
	  int newCol = ((PrintTabItem) o).value.intValue() - 1;
	  if( newCol < col ) {
	    this.bcScreen.println();
	    col = 0;
	  }
	  if( col < newCol ) {
	    StringBuilder buf = new StringBuilder();
	    for( int i = col; i < newCol; i++ ) {
	      buf.append( '\u0020' );
	    }
	    this.bcScreen.print( buf );
	  }
	  nl = false;
	} else {
	  String text = null;
	  if( o instanceof Number ) {
	    text = formatNumber( (Number) o, true );
	  } else {
	    text = o.toString();
	  }
	  if( text != null ) {
	    this.bcScreen.print( text );
	  }
	  nl = true;
	}
      }
      if( nl ) {
	this.bcScreen.println();
      }
    }
  }


  public void putRedeclarationWarning( String name, int srcIndex )
  {
    if( this.warnRedeclaration ) {
      this.prgParser.putWarning(
		ProgramParser.createMsg( "bc.redeclaration", name ),
		srcIndex );
    }
  }


  public void putVarNotInitializedWarning( String varName, int srcIndex )
  {
    if( this.warnVarNotInitialized ) {
      this.prgParser.putWarning(
		ProgramParser.createMsg( "bc.var_not_initialized", varName ),
		srcIndex );
    }
  }


  public void setAudioEnabled( boolean audioEnabled )
  {
    this.audioEnabled = audioEnabled;
  }


  public void setDebugFld( DebugFld debugFld )
  {
    this.debugFld = debugFld;
  }


  public void setSpeed( int value )
  {
    if( (value >= 1) && (value <= 9) ) {
      this.speedBrakeCntInit = (int) Math.round( Math.exp( value ) ) - 1;
    } else {
      this.speedBrakeCntInit = 0;
    }
  }


  // Set the value of a simple or dimensioned variable
  public void setVarValue(
			String       varName,
			List<Number> indices,
			Object       value,
			int          srcIndex ) throws ParseException
  {
    if( varName.endsWith( "$" ) ) {
      if( !(value instanceof String) ) {
	throw new ParseException(
			Main.getText( "bc.type_mismatch" ),
			srcIndex );
      }
    } else {
      if( !(value instanceof Number) ) {
	throw new ParseException(
			Main.getText( "bc.type_mismatch" ),
			srcIndex );
      }
    }
    if( indices != null ) {
      if( indices.isEmpty() ) {
	synchronized( this.simpleVars ) {
	  this.simpleVars.put( varName, value );
	}
      } else {
	getDimVar( varName, srcIndex ).setValue( indices, value, srcIndex );
      }
    } else {
      synchronized( this.simpleVars ) {
	this.simpleVars.put( varName, value );
      }
    }
  }


  public void settingsChanged()
  {
    this.warnRedeclaration     = getWarnRedeclarationSetting();
    this.warnVarNotInitialized = getWarnVarNotInitializedSetting();
  }


	/* --- Runnable --- */

  @Override
  public void run()
  {
    String msg = null;
    try {
      this.mainFrm.setBCBreakKeyEnabled( true );
      this.bcScreen.setBgColorNum( bcScreen.getBgColorNumSetting() );
      this.bcScreen.setFgColorNum( bcScreen.getFgColorNumSetting() );
      this.bcScreen.clear();
      this.prgParser.parse( this.prgText );
      if( this.prgParser.hasBeenAborted() ) {
	msg = Main.getText( "bc.program_aborted" );
      } else {
	msg = Main.getText( "bc.program_finished" );
      }
    }
    catch( InterruptedException ex ) {
      msg = Main.getText( "bc.program_aborted" );
    }
    catch( Exception ex ) {
      CharArrayWriter cWriter = new CharArrayWriter();
      PrintWriter     pWriter = new PrintWriter( cWriter );
      ex.printStackTrace( pWriter );
      pWriter.close();
      msg = cWriter.toString();
    }
    finally {
      for( Closeable c : this.dataStreams.values() ) {
	IOUtil.closeSilently( c );
      }
      if( this.synthesizer != null ) {
	this.synthesizer.close();
      }
      this.dataStreams.clear();
      this.mainFrm.bcThreadFinished( msg );
    }
  }


	/* --- Private Methodes --- */

  private static void checkDimValue(
				int value,
				int index ) throws ParseException
  {
    if( value < 0 ) {
      throw new ParseException(
			ProgramParser.createMsg(
				"bc.illegal_arg_value",
				value ),
			index );
    }
  }


  private void declareVarCC()
  {
    DimVar ccVar      = new DimVar( "CC", 1 );
    ccVar.setValue( 0, bcScreen.getFgColorNumSetting() );
    ccVar.setValue( 1, bcScreen.getBgColorNumSetting() );
    synchronized( this.simpleVars ) {
      this.dimVars.put( "CC", ccVar );
    }
  }


  // Set text mode and Clear screen
  private void doGOSUB100( int srcIndex )
  {
    this.bcScreen.setFgColorNum( getVarCCasInt( 0, srcIndex ) );
    this.bcScreen.setBgColorNum( getVarCCasInt( 1, srcIndex ) );
    this.bcScreen.setGraphicsMode( false );
    this.bcScreen.clear();
  }


  // Set cursor position
  private void doGOSUB110( int srcIndex )
  {
    /*
     * Setting the color here is an optional feature
     * that does not need to be supported,
     * but is supported by many BASICODE implementations.
     */
    this.bcScreen.setFgColorNum( getVarCCasInt( 0, srcIndex ) );
    this.bcScreen.setBgColorNum( getVarCCasInt( 1, srcIndex ) );
    this.bcScreen.setCursorPos(
			getVarVEasInt( srcIndex ),
			getVarHOasInt( srcIndex ) );
  }


  // Get cursor position
  private void doGOSUB120()
  {
    synchronized( this.simpleVars ) {
      this.simpleVars.put( "HO", this.bcScreen.getCursorCol() );
      this.simpleVars.put( "VE", this.bcScreen.getCursorRow() );
    }
  }


  // Highlighted text output
  private void doGOSUB150( int srcIndex )
  {
    this.bcScreen.print(
		String.format(
			"   %s   ",
			getStringVarSR( srcIndex ) ),
		getVarCCasInt( 0, srcIndex),	// bg color used as fg color
		getVarCCasInt( 1, srcIndex) );	// fg color used as bg color
  }


  // Inkey
  private void doGOSUB200()
  {
    setKeyboardVars( this.bcScreen.getCurKeyChar() );
  }


  // Inchar
  private void doGOSUB210() throws InterruptedException
  {
    setKeyboardVars(
	this.bcScreen.readCharFromKeyboard( this.debugFld, -1 ) );
  }


  // Get character at screen position
  private void doGOSUB220( int srcIndex )
  {
    int ch = this.bcScreen.getCharAt(
			getVarVEasInt( srcIndex ),
			getVarHOasInt( srcIndex ) );
    if( ch < 0 ) {
      ch = 0;
    }
    int value = Character.toUpperCase( ch );
    synchronized( this.simpleVars ) {
      this.simpleVars.put( "IN", value );
      this.simpleVars.put( "CN", ch - value );
    }
  }


  // Beep
  private void doGOSUB250( int srcIndex ) throws InterruptedException
  {
    if( this.audioEnabled ) {
      try {
	Toolkit.getDefaultToolkit().beep();
      }
      catch( AWTError e ) {}
    }
  }


  // Get a random number
  private void doGOSUB260()
  {
    synchronized( this.simpleVars ) {
      this.simpleVars.put( "RV", this.random.nextDouble() );
    }
  }


  /*
   * Get free memory
   *
   * This methid calculates the string space, per string its lengh + 1 byte,
   * and returns the difference to 0x7FFF.
   */
  private void doGOSUB270()
  {
    int stringSpace = 0;
    for( Object o : this.simpleVars.values() ) {
      if( o instanceof String ) {
	int len = ((String) o).length();
	if( len > 0 ) {
	  stringSpace = stringSpace + len + 1;
	}
      }
    }
    for( DimVar dimVar : this.dimVars.values() ) {
      int len = dimVar.calcStringSpace();
      if( len > 0 ) {
	stringSpace = stringSpace + len + 1;
      }
    }
    int freeSpace = 0x7FFF - stringSpace;
    if( freeSpace < 0 ) {
      freeSpace = 0;
    }
    synchronized( this.simpleVars ) {
      this.simpleVars.put( "FR", freeSpace );
    }
  }


  // Disable or enable stop key
  private void doGOSUB280( int srcIndex )
  {
    this.mainFrm.setBCBreakKeyEnabled(
		((Number) getVarValue( "FR", srcIndex )).intValue() == 0 );
  }


  // Convert a number to string
  private void doGOSUB300( int srcIndex )
  {
    synchronized( this.simpleVars ) {
      this.simpleVars.put(
		"SR$",
		formatNumber(
			(Number) getVarValue( "SR", srcIndex ),
			false ) );
    }
  }


  // Format a number
  private void doGOSUB310( int srcIndex )
  {
    String text = null;
    int    ct   = ((Number) getVarValue( "CT", srcIndex )).intValue();
    if( ct > 0 ) {
      if( ct > 20 ) {
	ct = 20;
      }
      int cn = ((Number) getVarValue( "CN", srcIndex )).intValue();
      if( cn < 0 ) {
	cn = 0;
      }
      Number value = (Number) getVarValue( "SR", srcIndex );
      if( cn > 0 ) {
	StringBuilder fmt = new StringBuilder();
	fmt.append( '%' );
	fmt.append( cn + 1 );
	fmt.append( '.' );
	fmt.append( cn );
	fmt.append( 'f' );
	text = String.format(
			Locale.ROOT,
			fmt.toString(),
			value.doubleValue() );
      } else {
	if( value instanceof Integer ) {
	  text = value.toString();
	} else {
	  text = String.valueOf( Math.round( value.doubleValue() ) );
	}
      }
    }
    if( text != null ) {
      if( text.length() > ct ) {
	text = null;			// not displayable
      }
    }
    if( text != null ) {
      int len = text.length();
      if( len < ct ) {
	StringBuilder buf = new StringBuilder();
	for( int i = len; i < ct; i++ ) {
	  buf.append( '\u0020' );
	}
	buf.append( text );
	text = buf.toString();
      }
    } else {
      StringBuilder buf = new StringBuilder( ct );
      while( ct > 0 ) {
	buf.append( '*' );
	--ct;
      }
      text = buf.toString();
    }
    synchronized( this.simpleVars ) {
      this.simpleVars.put( "SR$", text );
    }
  }


  /*
   * Convert a string to uppercase,
   * According to BASICODE standard all characters in the range
   * 96 (60h) to 128 (80h) will be changed.
   *
   * To ensure that the resulting string has the same length
   * the convertion is done character by character
   */
  private void doGOSUB330( int srcIndex )
  {
    StringBuilder buf = new StringBuilder( getStringVarSR( srcIndex ) );
    int len = buf.length();
    for( int i = 0; i < len; i++ ) {
      char ch = Character.toUpperCase( buf.charAt( i ) );
      if( (ch >= '\u0060') && (ch <= '\u007E') ) {
	ch -= '\u0020';
      } else if( ch >= '\u0080' ) {
	ch = Character.toLowerCase( ch );
      }
      buf.setCharAt( i, ch );
    }
    synchronized( this.simpleVars ) {
      this.simpleVars.put( "SR$", buf.toString() );
    }
  }


  // Printer output
  private void doGOSUB350( int srcIndex )
  {
    String text = getStringVarSR( srcIndex );
    if( text != null ) {
      getPrinter().append( text );
      this.mainFrm.fireCurPrintDocChanged();
    }
  }


  // Printer carriage return
  private void doGOSUB360()
  {
    getPrinter().append( '\n' );
    this.mainFrm.fireCurPrintDocChanged();
  }


  // Sound
  private void doGOSUB400( int srcIndex ) throws InterruptedException
  {
    int    duration = 0;
    Object varSD    = this.simpleVars.get( "SD" );
    if( varSD != null ) {
      duration = ((Number) varSD).intValue();
    }
    if( duration > 0 ) {
      if( duration > 255 ) {
	duration = 255;
      }
      boolean audible = false;
      boolean done    = false;
      Object  varSV   = this.simpleVars.get( "SV" );
      if( varSV != null ) {
	if( ((Number) varSV).intValue() > 0 ) {
	  audible = true;
	}
      }
      if( this.audioEnabled && audible ) {
	Object varSP = this.simpleVars.get( "SP" );
	if( varSP != null ) {
	  playNote(
		((Number) varSP).intValue(),
		duration,
		srcIndex );
	  done = true;
	}
      }
      if( !done
	  && (this.audioEnabled || getWaitDurationWhenSoundOffSetting()) )
      {
	waitTenthOfSeconds( duration );
      }
    }
  }


  // Pause
  private void doGOSUB450( int srcIndex ) throws InterruptedException
  {
    int  sd    = ((Number) getVarValue( "SD", srcIndex )).intValue();
    long mWait = sd * 100;
    long mBeg  = System.currentTimeMillis();
    setKeyboardVars(
	this.bcScreen.readCharFromKeyboard( this.debugFld, mWait ) );
    long mRemain = mBeg + mWait - System.currentTimeMillis();
    if( mRemain < 0 ) {
      mRemain = 0;
    }
    synchronized( this.simpleVars ) {
      this.simpleVars.put( "SD", (int) (mRemain / 100) );
    }
  }


  // open file
  private void doGOSUB500( int srcIndex ) throws InterruptedException
  {
    synchronized( this.simpleVars ) {
      this.simpleVars.put( "IN", -1 );
    }
    int fileNum = getFileNum( srcIndex );
    if( this.dataStreams.containsKey( fileNum ) ) {
      this.prgParser.putError(
			Main.getText( "bc.file_already_open" ),
			srcIndex );
    } else {
      String fileName = null;
      Object nameVar  = this.simpleVars.get( "NF$" );
      if( nameVar != null ) {
	fileName = nameVar.toString().trim();
      }
      int    fileMode = 0;
      Object modeVar  = this.simpleVars.get( "NF" );
      if( modeVar != null ) {
	if( modeVar instanceof Number ) {
	  fileMode = ((Number) modeVar).intValue();
	}
      }
      boolean forSave = (fileMode & 0x01) != 0;
      File    file    = UIUtil.openDataFileDlg(
					this.mainFrm,
					forSave,
					recentDirFile,
					fileName );
      if( file != null ) {
	if( this.dataStreams.isEmpty() ) {
	  this.dataFileEOR = getEndOfRecord();
	}
	try {
	  Closeable stream = null;
	  if( forSave ) {
	    stream = new BufferedOutputStream(
					new FileOutputStream( file ) );
	  } else {
	    stream = new PushbackInputStream(
					new FileInputStream( file ) );
	  }
	  if( stream != null ) {
	    this.dataStreams.put( fileNum, stream );
	    synchronized( this.simpleVars ) {
	      this.simpleVars.put( "IN", 0 );	// success
	    }
	    recentDirFile = file.getParentFile();
	  }
	}
	catch( IOException ex ) {
	  this.prgParser.putError( ex.getMessage(), srcIndex );
	}
      }
    }
  }


  // read file
  private void doGOSUB540( int srcIndex )
  {
    synchronized( this.simpleVars ) {
      this.simpleVars.put( "IN", -1 );
      this.simpleVars.put( "IN$", "" );
    }
    Closeable closeable = this.dataStreams.get( getFileNum( srcIndex ) );
    if( closeable != null ) {
      if( closeable instanceof PushbackInputStream ) {
	PushbackInputStream in = (PushbackInputStream) closeable;
	try {
	  int b = in.read();
	  if( b >= 0 ) {
	    StringBuilder buf = new StringBuilder();
	    while( (b >= 0) && (b != '\r') && (b != '\n') ) {
	      buf.append( (char) b );
	      b = in.read();
	    }
	    if( b == '\r' ) {
	      b = in.read();
	      if( (b >= 0) && (b != '\n') ) {
		in.unread( b );
	      }
	    }
	    synchronized( this.simpleVars ) {
	      this.simpleVars.put( "IN$", buf.toString() );
	    }

	    // check end of file
	    b = in.read();
	    if( b >= 0 ) {
	      in.unread( b );
	      synchronized( this.simpleVars ) {
		this.simpleVars.put( "IN", 0 );		// success
	      }
	    } else {
	      synchronized( this.simpleVars ) {
		this.simpleVars.put( "IN", 1 );		// end of file
	      }
	    }
	  } else {
	    synchronized( this.simpleVars ) {
	      this.simpleVars.put( "IN", 1 );		// end of file
	    }
	  }
	}
	catch( IOException ex ) {
	  this.prgParser.putError( ex.getMessage(), srcIndex );
	}
      } else {
	this.prgParser.putError(
			Main.getText( "bc.file_access_mode_mismatch" ),
			srcIndex );
      }
    } else {
      this.prgParser.putError(
			Main.getText( "bc.file_not_open" ),
			srcIndex );
    }
  }


  // write file
  private void doGOSUB560( int srcIndex )
  {
    synchronized( this.simpleVars ) {
      this.simpleVars.put( "IN", -1 );
    }
    Closeable closeable = this.dataStreams.get( getFileNum( srcIndex ) );
    if( closeable != null ) {
      if( closeable instanceof OutputStream ) {
	OutputStream out = (OutputStream) closeable;
	try {
	  String text = getStringVarSR( srcIndex );
	  if( text != null ) {
	    int len = text.length();
	    for( int i = 0; i < len; i++ ) {
	      out.write( text.charAt( i ) );
	    }
	    switch( this.dataFileEOR ) {
	      case CR:
		out.write( '\r' );
		break;
	      case CRNL:
		out.write( '\r' );
		out.write( '\n' );
		break;
	      case NL:
		out.write( '\n' );
		break;
	    }
	    synchronized( this.simpleVars ) {
	      this.simpleVars.put( "IN", 0 );	// success
	    }
	  }
	}
	catch( IOException ex ) {
	  this.prgParser.putError( ex.getMessage(), srcIndex );
	}
      } else {
	this.prgParser.putError(
			Main.getText( "bc.file_access_mode_mismatch" ),
			srcIndex );
      }
    } else {
      this.prgParser.putError(
			Main.getText( "bc.file_not_open" ),
			srcIndex );
    }
  }


  // close file
  private void doGOSUB580( int srcIndex )
  {
    synchronized( this.simpleVars ) {
      this.simpleVars.put( "IN", -1 );
    }
    int       fileNum   = getFileNum( srcIndex );
    Closeable closeable = this.dataStreams.get( fileNum );
    if( closeable != null ) {
      try {
	closeable.close();
	synchronized( this.simpleVars ) {
	  this.simpleVars.put( "IN", 0 );
	}
      }
      catch( IOException ex ) {
	this.prgParser.putError( ex.getMessage(), srcIndex );
      }
      finally {
	this.dataStreams.remove( fileNum );
      }
    }
  }


  // Set graphics mode and Clear screen
  private void doGOSUB600( int srcIndex )
  {
    this.bcScreen.setFgColorNum( getVarCCasInt( 0, srcIndex ) );
    this.bcScreen.setBgColorNum( getVarCCasInt( 1, srcIndex ) );
    this.bcScreen.setGraphicsMode( true );
    this.bcScreen.clear();
    synchronized( this.simpleVars ) {
      this.simpleVars.put( "HG", this.bcScreen.getScreenWidth() );
      this.simpleVars.put( "VG", this.bcScreen.getScreenHeight() );
    }
  }


  // Set pixel in graphics mode
  private void doGOSUB620( int srcIndex )
  {
    this.bcScreen.setPixel(
			getVarHOasDouble( srcIndex ),
			getVarVEasDouble( srcIndex ),
			getGraphicsFgColorNum( srcIndex ) );
  }


  // Draw line in graphics mode
  private void doGOSUB630( int srcIndex )
  {
    this.bcScreen.drawLineTo(
			getVarHOasDouble( srcIndex ),
			getVarVEasDouble( srcIndex ),
			getGraphicsFgColorNum( srcIndex ) );
  }


  // Draw text in graphics mode
  private void doGOSUB650( int srcIndex )
  {
    this.bcScreen.drawText(
			getVarHOasDouble( srcIndex ),
			getVarVEasDouble( srcIndex ),
			getGraphicsFgColorNum( srcIndex ),
			getStringVarSR( srcIndex ) );
  }


  // Init program and jump to line 1010
  private void doGOTO20( int srcIndex ) throws GotoException
  {
    // clear variables
    this.dimVars.clear();
    this.simpleVars.clear();

    // declare variables
    declareVarCC();
    synchronized( this.simpleVars ) {
      this.simpleVars.put( "HO", this.bcScreen.getCharCols() - 1 );
      this.simpleVars.put( "VE", this.bcScreen.getCharRows() - 1 );
      this.simpleVars.put( "HG", this.bcScreen.getScreenWidth() );
      this.simpleVars.put( "VG", this.bcScreen.getScreenHeight() );
      this.simpleVars.put( "SV", 35 );		// BASICODE version
    }

    // clear screen
    doGOSUB100( srcIndex );

    // go to line 1010
    throw new GotoException( -1, 1010 );
  }


  private void doGOTO950() throws EndException
  {
    if( !Main.getBooleanProperty(
			PROP_GOTO950_WITHOUT_CLS,
			DEFAULT_GOTO950_WITHOUT_CLS ) )
    {
      this.bcScreen.clear();
      this.bcScreen.setDisabled();
    }
    doEnd();
  }


  /*
   * Formats s number:
   *  - First character is space or minus sign
   *  - The actual number is a maximum of 11 characters long
   *    including the decimal point
   *  - Last character is space
   */
  private static String formatNumber( Number value, boolean withSpaces )
  {
    String text = value.toString();
    if( text != null ) {
      int len = text.length();
      if( len > 0 ) {
	StringBuilder buf   = new StringBuilder();
	int           width = 11;
	int           idx   = 0;
	char          ch    = text.charAt( idx );
	if( ch == '+' ) {			// should never happen
	  idx++;
	  if( withSpaces ) {
	    buf.append( '\u0020' );
	    width++;
	  }
	} else if( ch == '-' ) {
	  idx++;
	  buf.append( ch );
	  width++;
	} else if( withSpaces ) {
	  buf.append( '\u0020' );
	  width++;
	}
	int pt  = -1;
	int exp = -1;
	while( idx < len ) {
	  ch = text.charAt( idx++ );
	  if( ch == '.' ) {
	    pt = buf.length();
	    buf.append( ch );
	  } else if( (ch == 'E') || (ch == 'e') ) {
	    exp = buf.length();
	    buf.append( 'E' );
	  } else {
	    buf.append( ch );
	  }
	}
	if( pt >= 0 ) {
	  ch  = '\u0000';
	  idx = (exp > 0 ? exp : len);
	  while( (idx > pt) && (buf.length() > width) ) {
	    --idx;
	    ch = buf.charAt( idx );
	    buf.deleteCharAt( idx );
	  }
	  if( (ch >= '5') && (idx > 0) ) {
	    // round up previous digit
	    --idx;
	    ch = buf.charAt( idx );
	    if( (ch >= '0') && (ch < '9') ) {
	      buf.setCharAt( idx, (char) (ch + 1) );
	    }
	  }
	}
	if( withSpaces ) {
	  buf.append( '\u0020' );
	}
	text = buf.toString();
      }
    }
    return text;
  }


  private DimVar getDimVar(
			String varName,
			int    srcIndex ) throws ParseException
  {
    DimVar dimVar = this.dimVars.get( varName );
    if( dimVar == null ) {
      throw new ParseException(
			ProgramParser.createMsg(
					"bc.field_not_found",
					varName ),
			srcIndex );
    }
    return dimVar;
  }


  private int getFileNum( int srcIndex )
  {
    return ((Number) getVarValue( "NF", srcIndex )).intValue();
  }


  private int getGraphicsFgColorNum( int srcIndex )
  {
    int colorNum = this.bcScreen.getBgColorNum();
    if( ((Number) getVarValue( "CN", srcIndex )).intValue() == 0 ) {
      colorNum = getVarCCasInt( 0, srcIndex );
    }
    return colorNum;
  }


  private StringBuffer getPrinter()
  {
    if( this.printer == null ) {
      this.printer = new StringBuffer();
      this.mainFrm.fireAddPrintDoc( this.printer );
    }
    return this.printer;
  }


  private String getStringVarSR( int srcIndex )
  {
    return getVarValue( "SR$", srcIndex ).toString();
  }


  private int getVarCCasInt( int fieldIdx, int srcIdx )
  {
    int    value = BCScreen.DEFAULT_COLOR_AT_START_FG;
    DimVar ccVar = this.dimVars.get( "CC" );
    if( ccVar != null ) {
      value = ((Number) ccVar.getValue( fieldIdx )).intValue();
      if( (value < 0) || (value > 7) ) {
	this.prgParser.putSpecViolatedWarning(
		MessageFormat.format(
			Main.getText( "bc.var_x_with_invalid_colornum_y" ),
			String.format( "CC(%1d)", fieldIdx ),
			String.valueOf( value ) ),
		srcIdx );
      }
    } else {
      value = BCScreen.getFgColorNumSetting();
    }
    return value;
  }


  private double getVarHOasDouble( int srcIndex )
  {
    return ((Number) getVarValue( "HO", srcIndex )).doubleValue();
  }


  private int getVarHOasInt( int srcIndex )
  {
    return ((Number) getVarValue( "HO", srcIndex )).intValue();
  }


  private double getVarVEasDouble( int srcIndex )
  {
    return ((Number) getVarValue( "VE", srcIndex )).doubleValue();
  }


  private int getVarVEasInt( int srcIndex )
  {
    return ((Number) getVarValue( "VE", srcIndex )).intValue();
  }


  private void playNote(
			int pitch,
			int duration,
			int srcIndex ) throws InterruptedException
  {
    if( this.midiAvailable
	&& (pitch >= 0) && (pitch <= 127)
	&& (duration >= 1) && (duration <= 255) )
    {
      try {
	if( this.synthesizer == null ) {
	  this.synthesizer = MidiSystem.getSynthesizer();
	  this.synthesizer.open();
	}
	if( this.midiChannel == null ) {

	  // find a non-null channel
	  MidiChannel[] channels = this.synthesizer.getChannels();
	  if( channels != null ) {
	    for( MidiChannel channel : channels ) {
	      if( channel != null ) {
		this.midiChannel = channel;
		break;
	      }
	    }
	  }
	}
	if( this.midiChannel != null ) {
	  this.midiChannel.noteOn( pitch, 100 );
	  waitTenthOfSeconds( duration );
	  this.midiChannel.noteOff( pitch );
	}
      }
      catch( MidiUnavailableException ex ) {
	this.prgParser.putError( ex.getMessage(), srcIndex );
	this.midiAvailable = false;
      }
    }
  }


  private void setKeyboardVars( int keyCode )
  {
    String strValue = "";
    int    intValue = 0;
    if( keyCode < 0 ) {
      intValue = keyCode;
    } else if( (keyCode > 0) && (keyCode < 0x10000) ) {
      char ch  = (char) keyCode;
      strValue = String.valueOf( ch );
      if( (keyCode >= 0x60) && (keyCode < 0x7F) ) {
	keyCode -= 0x20;	// incl. special character in that range
      } else if( keyCode >= 0x80 ) {
	keyCode = Character.toUpperCase( ch );
      }
      if( (keyCode == 13) || ((keyCode >= 0x1B) && (keyCode < 0x80)) ) {
	intValue = keyCode;
      } else {
	intValue = -keyCode;
      }
    }
    synchronized( this.simpleVars ) {
      this.simpleVars.put( "IN", intValue);
      this.simpleVars.put( "IN$", strValue);
    }
  }


  private void waitTenthOfSeconds( int tenthOfSecond )
					throws InterruptedException
  {
    synchronized( this.waitMonitor ) {
      try {
	this.waitMonitor.wait( tenthOfSecond * 100 );
      }
      catch( IllegalArgumentException ex ) {}
      catch( IllegalMonitorStateException ex ) {}
    }
  }
}
