/*
 * (c) 2025 Jens Mueller
 *
 * JBASICODE
 *
 * BASICODE Screen Device
 */

package jbasicode.bc;

import java.awt.Color;
import java.awt.Font;
import java.awt.EventQueue;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.image.BufferedImage;
import java.util.Arrays;
import jbasicode.Main;
import jbasicode.ui.DebugFld;
import jbasicode.ui.MainFrm;
import jbasicode.ui.ScreenFld;


public class BCScreen
{
  public static final String VALUE_ASPECT_RATIO_4_3   = "4:3";
  public static final String VALUE_ASPECT_RATIO_16_10 = "16:10";
  public static final String DEFAULT_ASPECT_RATIO = VALUE_ASPECT_RATIO_4_3;

  public static final int DEFAULT_CHAR_COLS         = 40;
  public static final int DEFAULT_CHAR_ROWS         = 25;
  public static final int DEFAULT_FONT_SIZE         = 13;
  public static final int DEFAULT_SCREEN_WIDTH      = 320;
  public static final int DEFAULT_SCREEN_HEIGHT     = 240;
  public static final int DEFAULT_MARGIN            = 10;
  public static final int DEFAULT_COLOR_AT_START_FG = 7;
  public static final int DEFAULT_COLOR_AT_START_BG = 0;

  public static final boolean DEFAULT_CURSOR_ALWAYS_VISIBLE = false;
  public static final boolean DEFAULT_INPUT_REVERSE_CASE    = false;

  public static final String PROP_ASPECT_RATIO      = "screen.aspect_ratio";
  public static final String PROP_COLOR_AT_START_BG = "start.color.bg";
  public static final String PROP_COLOR_AT_START_FG = "start.color.fg";
  public static final String PROP_CURSOR_ALWAYS_VISIBLE
						= "cursor.always_visible";
  public static final String PROP_INPUT_REVERSE_CASE
						= "bc.input.reverse_case";

  private static final Color[] colors = {
					Color.BLACK,
					Color.BLUE,
					Color.RED,
					Color.MAGENTA,
					Color.GREEN,
					Color.CYAN,
					Color.YELLOW,
					Color.WHITE,
					// same colors with half brightness
					Color.BLACK,
					new Color( 0xFF00007F ),
					new Color( 0xFF007F00 ),
					new Color( 0xFF007F7F ),
					new Color( 0xFF7F0000 ),
					new Color( 0xFF7F007F ),
					new Color( 0xFF7F7F00 ),
					new Color( 0xFF7F7F7F ) };

  private MainFrm                mainFrm;
  private volatile ScreenFld     screenFld;
  private volatile boolean       bcBreakKeyEnabled;
  private volatile boolean       bcBreakKeyPressed;
  private volatile boolean       cursorAlwaysVisible;
  private volatile boolean       graphicsMode;
  private volatile boolean       inputCharMode;
  private boolean                inputReverseCase;
  private volatile Object        inputMonitor;
  private StringBuilder          inputLineBuf;
  private int                    inputCharBuf;
  private int                    curKeyChar;
  private int                    cursorCol;
  private int                    cursorRow;
  private double                 cursorX;
  private double                 cursorY;
  private int                    charCols;
  private int                    charRows;
  private int                    wChar;
  private int                    hChar;
  private int                    margin;
  private int                    height;
  private int                    width;
  private int                    bgColorNum;
  private int                    fgColorNum;
  private int[]                  bgColorBuf;
  private int[]                  fgColorBuf;
  private char[]                 charBuf;
  private Font                   font;
  private volatile BufferedImage image;
  private String                 textWaitForInput;
  private String                 textWaitForKey;
  private Object                 threadLock;


  public BCScreen( MainFrm mainFrm )
  {
    this.mainFrm             = mainFrm;
    this.screenFld           = screenFld;
    this.bcBreakKeyEnabled   = true;
    this.bcBreakKeyPressed   = false;
    this.cursorAlwaysVisible = false;
    this.graphicsMode        = false;
    this.inputCharMode       = false;
    this.inputCharBuf        = 0;
    this.inputLineBuf        = new StringBuilder();
    this.inputMonitor        = null;
    this.curKeyChar          = 0;
    this.cursorCol           = 0;
    this.cursorRow           = 0;
    this.cursorX             = 0.0;
    this.cursorY             = 0.0;
    this.charCols            = 0;
    this.charRows            = 0;
    this.charBuf             = new char[ 0 ];
    this.bgColorBuf          = new int[ 0 ];
    this.fgColorBuf          = new int[ 0 ];
    this.wChar               = 0;
    this.hChar               = 0;
    this.margin              = 0;
    this.height              = 0;
    this.width               = 0;
    this.bgColorNum          = getBgColorNumSetting();
    this.fgColorNum          = getFgColorNumSetting();
    this.font                = null;
    this.image               = null;
    this.textWaitForInput    = Main.getText( "status.bc.wait_for_input" );
    this.textWaitForKey      = Main.getText( "status.bc.wait_for_key" );
    this.threadLock          = new Object();
    settingsChanged();
    clear();
  }


  public void clear()
  {
    synchronized( this.threadLock ) {
      BufferedImage image = this.image;
      if( image != null ) {
	Graphics g = this.image.getGraphics();
	try {
	  g.setColor( getColor( this.bgColorNum ) );
	  g.fillRect( 0, 0, image.getWidth(), image.getHeight() );
	}
	finally {
	  g.dispose();
	}
      }
      Arrays.fill( this.charBuf, '\u0020' );
      Arrays.fill( this.bgColorBuf, this.bgColorNum );
      Arrays.fill( this.fgColorBuf, this.fgColorNum );
      this.cursorRow = 0;
      this.cursorCol = 0;
      this.cursorX   = 0.0;
      this.cursorY   = 0.0;
    }
    setDirty();
  }


  public void drawLineTo(
		double x,
		double y,
		int    colorNum )
  {
    BufferedImage image = this.image;
    if( image != null ) {
      int imgW  = image.getWidth();
      int imgH  = image.getHeight();
      int imgX1 = (int) Math.round( x * (double) imgW );
      int imgY1 = (int) Math.round( y * (double) imgH );
      int imgX2 = (int) Math.round( this.cursorX * (double) imgW );
      int imgY2 = (int) Math.round( this.cursorY * (double) imgH );
      Graphics g = image.getGraphics();
      try {
	g.setColor( getColor( colorNum ) );
	g.drawLine( imgX1, imgY1, imgX2, imgY2 );
      }
      finally {
	g.dispose();
	this.cursorX = x;
	this.cursorY = y;
      }
      setDirty();
    }
  }


  public void drawText( double x, double y, int colorNum, String text )
  {
    BufferedImage image = this.image;
    if( (image != null) && (text != null) ) {
      int imgW = image.getWidth();
      int imgH = image.getHeight();
      int imgX = (int) Math.round( x * (double) imgW );
      int imgY = (int) Math.round( y * (double) imgH );
      Graphics g = image.getGraphics();
      try {
	Font font = this.font;
	if( font != null ) {
	  g.setFont( font );
	}
	g.setColor( getColor( colorNum ) );
	int len = text.length();
	for( int i = 0; i < len; i++ ) {
	  g.drawString(
		text.substring( i, i + 1 ),
		imgX,
		imgY + this.hChar - 6 );
	  imgX += this.wChar;
	}
      }
      finally {
	g.dispose();
	this.cursorX = x;
	this.cursorY = y;
      }
      setDirty();
    }
  }


  public int getBgColorNum()
  {
    return this.bgColorNum;
  }


  public static int getBgColorNumSetting()
  {
    return Main.getIntProperty(
			PROP_COLOR_AT_START_BG,
			DEFAULT_COLOR_AT_START_BG );
  }


  public int getCharAt( int row, int col )
  {
    int rv = 0;
    synchronized( this.threadLock ) {
      if( (row >= 0) && (row < this.charRows)
	  && (col >= 0) && (col < this.charCols) )
      {
	int idx = (row * this.charCols) + col;
	if( idx < this.charBuf.length ) {
	  rv = this.charBuf[ idx ];
	}
      }
    }
    return rv;
  }


  public int getCharCols()
  {
    return this.charCols;
  }


  public int getCharRows()
  {
    return this.charRows;
  }


  public int getCurKeyChar()
  {
    return this.curKeyChar;
  }


  public int getCursorCol()
  {
    return this.cursorCol;
  }


  public int getCursorRow()
  {
    return this.cursorRow;
  }


  public static int getFgColorNumSetting()
  {
    return Main.getIntProperty(
			PROP_COLOR_AT_START_FG,
			DEFAULT_COLOR_AT_START_FG );
  }


  public int getHeight()
  {
    return this.height;
  }


  public static boolean getInputReverseCase()
  {
    return Main.getBooleanProperty(
			PROP_INPUT_REVERSE_CASE,
			DEFAULT_INPUT_REVERSE_CASE );
  }


  public int getScreenHeight()
  {
    BufferedImage image = this.image;
    return image != null ? image.getHeight() : 0;
  }


  public int getScreenWidth()
  {
    BufferedImage image = this.image;
    return image != null ? image.getWidth() : 0;
  }


  public int getWidth()
  {
    return this.width;
  }


  public boolean isGraphicsMode()
  {
    return this.graphicsMode;
  }


  public void keyTyped( char ch )
  {
    if( ch == '\n' ) {
      ch = '\r';
    }
    if( ch >= '\u0020' ) {
      this.curKeyChar = ch;
    }
    if( (ch == '\r') || (ch == '\u0008') || (ch >= '\u0020') ) {
      if( this.inputReverseCase ) {
	if( Character.isUpperCase( ch ) ) {
	  ch = Character.toLowerCase( ch );
	} else if( Character.isLowerCase( ch ) ) {
	  ch = Character.toUpperCase( ch );
	}
      }
      Object inputMonitor = this.inputMonitor;
      if( inputMonitor != null ) {
	boolean endInput = false;
	if( this.inputCharMode ) {
	  this.inputCharBuf = ch;
	  endInput = true;
	} else {
	  if( ch == '\r' ) {
	    newLine();
	    setDirty();
	    endInput = true;
	  } else if( (ch == '\u0008') || (ch == '\u007F') ) {
	    if( this.inputLineBuf.length() > 0 ) {
	      this.inputLineBuf.setLength( this.inputLineBuf.length() - 1 );
	      synchronized( this.threadLock ) {
		int idx = getCursorIdx();
		if( idx >= 0 ) {
		  this.charBuf[ idx ]    = '\u0020';
		  this.bgColorBuf[ idx ] = this.bgColorNum;
		  this.fgColorBuf[ idx ] = this.fgColorNum;
		}
		printChar( '\u0008', this.bgColorNum, this.fgColorNum );
	      }
	    }
	  }
	  else if( (ch >= '\u0020')
		   && (this.inputLineBuf.length() < 255) )
	  {
	    this.inputLineBuf.append( ch );
	    printChar( ch, this.bgColorNum, this.fgColorNum );
	    setDirty();
	  }
	}
	if( endInput ) {
	  synchronized( inputMonitor ) {
	    try {
	      inputMonitor.notifyAll();
	    }
	    catch( IllegalMonitorStateException ex ) {}
	  }
	}
      }
    }
  }


  public void paint( Graphics graphics, int width, int height )
  {
    Graphics g = graphics.create();
    try {
      g.setColor( getColor( this.bgColorNum ) );
      g.fillRect( 0, 0, width, height );
      if( this.margin > 0 ) {
	g.translate( this.margin, this.margin );
	width -= (2 * this.margin);
	height -= (2 * this.margin);
      }

      // translate graphics area
      int wContent = 0;
      int hContent = 0;
      if( this.graphicsMode ) {
	BufferedImage image = this.image;
	if( image != null ) {
	  wContent = image.getWidth();
	  hContent = image.getHeight();
	}
      } else {
	wContent = this.charCols * this.wChar;
	hContent = this.charRows * this.hChar;
      }
      Graphics2D g2d    = null;
      Double     scaleF = null;
      if( g instanceof Graphics2D ) {
	if( (wContent > 0) && (hContent > 0) ) {
	  double fw = (double) width / (double) wContent;
	  double fh = (double) height / (double) hContent;
	  double f  = Math.min( fw, fh );
	  if( f != 1.0 ) {
	    g2d      = (Graphics2D) g;
	    scaleF   = Double.valueOf( f );
	    wContent = (int) Math.round( wContent * f );
	    hContent = (int) Math.round( hContent * f );
	  }
	}
      }
      int x = (width - wContent) / 2;
      if( x < 0 ) {
	x = 0;
      }
      int y = (height - hContent) / 2;
      if( y < 0 ) {
	y = 0;
      }
      if( (x > 0) || (y > 0) ) {
	g.translate( x, y );
      }

      // scale
      if( (g2d != null) && (scaleF != null) ) {
	g2d.scale( scaleF, scaleF );
      }

      // paint content
      if( this.graphicsMode ) {
	if( image != null ) {
	  g.drawImage( image, 0, 0, this.mainFrm );
	}
      } else {
	if( (this.charRows > 0) && (this.charCols > 0)
	    && (this.wChar > 0) && (this.hChar > 0) )
	{
	  // draw characters
	  x = 0;
	  y = 0;
	  Font font = this.font;
	  if( font != null ) {
	    g.setFont( font );
	    for( int row = 0; row < this.charRows; row++ ) {
	      y = row * this.hChar;
	      for( int col = 0; col < this.charCols; col++ ) {
		x       = col * this.wChar;
		int idx = (row * this.charCols) + col;
		if((idx >= 0) && (idx < this.charBuf.length) ) {
		  g.setColor( getColor( this.bgColorBuf[ idx ] ) );
		  g.fillRect( x, y, this.wChar, this.hChar );
		  char ch = this.charBuf[ idx ];
		  if( font.canDisplay( ch ) ) {
		    g.setColor( getColor( this.fgColorBuf[ idx ] ) );
		    g.drawChars(
			      this.charBuf,
			      idx,
			      1,
			      x,
			      y + this.hChar - 3 );
		  }
		}
	      }
	    }
	  }

	  // draw cursor
	  int cursorCol = -1;
	  int cursorRow = -1;
	  synchronized( this.threadLock ) {
	    cursorCol = this.cursorCol;
	    cursorRow = this.cursorRow;
	  }
	  if( (this.cursorAlwaysVisible
		|| ((this.inputMonitor != null) && !this.inputCharMode))
	      && (cursorCol >= 0) && (cursorCol < this.charCols)
	      && (cursorRow >= 0) && (cursorRow < this.charRows) )
	  {
	    x = cursorCol * this.wChar;
	    y = cursorRow * this.hChar;
	    g.setColor( getColor( this.fgColorNum ) );
	    g.fillRect( x, y, this.wChar - 1, this.hChar - 1 );
	  }
	}
      }
    }
    finally {
      g.dispose();
    }
  }


  public void print( CharSequence text )
  {
    print( text, this.bgColorNum, this.fgColorNum );
  }


  public void print( CharSequence text, int bgColorNum, int fgColorNum )
  {
    if( !this.graphicsMode && (text != null) ) {
      int len = text.length();
      if( len > 0 ) {
	for( int i = 0; i < len; i++ ) {
	  printChar( text.charAt( i ), bgColorNum, fgColorNum );
	}
	setDirty();
      }
    }
  }


  public void println()
  {
    if( !this.graphicsMode ) {
      newLine();
      setDirty();
    }
  }


  /*
   * Reads a character from keybord,
   * This method is thread safe and blocks the calling thread
   * until any key is pressed.
   */
  public int readCharFromKeyboard(
			DebugFld debugFld,
			long     timeoutMillis ) throws InterruptedException
  {
    try {
      this.mainFrm.fireSetStatusText( this.textWaitForKey );
      readFromKeyboard( debugFld, true, timeoutMillis );
    }
    finally {
      this.mainFrm.fireShowDefaultStatusText();
    }
    return this.inputCharBuf;
  }


  /*
   * Reads characters from keybord,
   * This method is thread safe and blocks the calling thread
   * untill the enter key is pressed.
   */
  public String readLineFromKeyboard( DebugFld debugFld )
						throws InterruptedException
  {
    try {
      this.mainFrm.fireSetStatusText( this.textWaitForInput );
      readFromKeyboard( debugFld, false, -1 );
    }
    finally {
      this.mainFrm.fireShowDefaultStatusText();
    }
    return this.inputLineBuf.toString();
  }


  public void setBgColorNum( int colorNum )
  {
    this.bgColorNum = colorNum;
  }


  public void setBCBreakKeyEnabled( boolean state )
  {
    this.bcBreakKeyEnabled = state;
    this.bcBreakKeyPressed = false;
  }


  public void setCurKeyChar( int ch )
  {
    this.curKeyChar = ch;
    if( this.bcBreakKeyEnabled && ((ch == 27) || (ch == -27)) ) {
      this.bcBreakKeyPressed = true;
    }
    if( this.inputCharMode && (ch != 0) ) {
      Object inputMonitor = this.inputMonitor;
      if( inputMonitor != null ) {
	this.inputCharBuf = ch;
	synchronized( inputMonitor ) {
	  try {
	    inputMonitor.notifyAll();
	  }
	  catch( IllegalMonitorStateException ex ) {}
	}
      }
    }
  }


  public void setCursorPos( int row, int col )
  {
    synchronized( this.threadLock ) {
      if( !this.graphicsMode
	  && (row >= 0) && (row < this.charRows)
	  && (col >= 0) && (col < this.charCols) )
      {
	this.cursorRow = row;
	this.cursorCol = col;
	setDirty();
      }
    }
  }


  public void setDisabled()
  {
    ScreenFld screenFld = this.screenFld;
    if( screenFld != null ) {
      screenFld.fireSetDisabled();
    }
  }


  public void setFgColorNum( int colorNum )
  {
    this.fgColorNum = colorNum;
  }


  public void setGraphicsMode( boolean state )
  {
    if( state != this.graphicsMode ) {
      this.graphicsMode = state;
      setDirty();
    }
  }


  public void setPixel( double x, double y, int colorNum )
  {
    BufferedImage image = this.image;
    if( image != null ) {
      int imgW = image.getWidth();
      int imgH = image.getHeight();
      int imgX = (int) Math.round( x * (double) imgW );
      int imgY = (int) Math.round( y * (double) imgH );
      if( (imgX >= 0) && (imgX < imgW)
	  && (imgY >=0) && (imgY < imgH) )
      {
	image.setRGB( imgX, imgY, getColor( colorNum ).getRGB() );
	setDirty();
      }
      this.cursorX = x;
      this.cursorY = y;
    }
  }


  public void setScreenFld( ScreenFld screenFld )
  {
    this.screenFld = screenFld;
    if( screenFld != null ) {
      screenFld.repaint();
    }
  }


  public void settingsChanged()
  {
    this.cursorAlwaysVisible = Main.getBooleanProperty(
					PROP_CURSOR_ALWAYS_VISIBLE,
					DEFAULT_CURSOR_ALWAYS_VISIBLE );
    this.inputReverseCase = getInputReverseCase();
  }


  public void updScreenSize()
  {
    int    screenHeight = DEFAULT_SCREEN_HEIGHT;
    String aspectRatio  = Main.getProperty( PROP_ASPECT_RATIO );
    if( aspectRatio != null ) {
      if( aspectRatio.equals( VALUE_ASPECT_RATIO_4_3 ) ) {
	screenHeight = DEFAULT_SCREEN_WIDTH * 3 / 4;
      } else if( aspectRatio.equals( VALUE_ASPECT_RATIO_16_10 ) ) {
	screenHeight = DEFAULT_SCREEN_WIDTH * 10 / 16;
      }
    }
    setScreenSize(
		DEFAULT_CHAR_ROWS,
		DEFAULT_CHAR_COLS,
		DEFAULT_FONT_SIZE,
		DEFAULT_SCREEN_WIDTH,
		screenHeight,
		DEFAULT_MARGIN );
  }


	/* --- Private Methods --- */

  private static Color getColor( int colorNum )
  {
    colorNum &= 0x0F;
    return (colorNum >= 0) && (colorNum < colors.length) ?
					colors[ colorNum ] : Color.BLACK;
  }


  private int getCursorIdx()
  {
    int idx = -1;
    synchronized( this.threadLock ) {
      if( (this.cursorRow >= 0) && (this.cursorRow < this.charRows)
	  && (this.cursorCol >= 0) && (this.cursorCol < this.charCols) )
      {
	idx = (this.cursorRow * this.charCols) + this.cursorCol;
	if( idx >= this.charBuf.length ) {
	  idx = -1;
	}
      }
    }
    return idx;
  }


  private void newLine()
  {
    synchronized( this.threadLock ) {
      if( (this.cursorRow + 1) < this.charRows ) {
	setCursorPos( this.cursorRow + 1, 0 );
      } else {
	scrollUp();
	setCursorPos( this.charRows - 1, 0 );
      }
    }
  }


  private void printChar( char ch, int bgColorNum, int fgColorNum )
  {
    switch( ch ) {
      case '\u0008':
	synchronized( this.threadLock ) {
	  int idx = getCursorIdx();
	  if( idx > 0 ) {
	    this.charBuf[ idx ]    = '\u0020';
	    this.bgColorBuf[ idx ] = bgColorNum;
	    this.fgColorBuf[ idx ] = fgColorNum;
	    --idx;
	    int row = this.cursorRow;
	    int col = this.cursorCol;
	    if( col > 0 ) {
	      --col;
	    } else if( row > 0 ) {
	      --row;
	      col = charCols - 1;
	    }
	    setCursorPos( row, col );
	  }
	}
	break;
      case '\f':
	clear();
	break;
      case '\r':
	newLine();
	break;
      case '\u007F':
	synchronized( this.threadLock ) {
	  int col = this.cursorCol;
	  int idx = getCursorIdx();
	  if( (col > 0) && (idx > 0) ) {
	    --col;
	    --idx;
	    this.charBuf[ idx ]    = '\u0020';
	    this.bgColorBuf[ idx ] = bgColorNum;
	    this.fgColorBuf[ idx ] = fgColorNum;
	    setCursorPos( this.cursorRow, col );
	  }
	}
	break;
      default:
	synchronized( this.threadLock ) {
	  int idx = getCursorIdx();
	  if( idx >= 0 ) {
	    this.charBuf[ idx ]    = ch;
	    this.bgColorBuf[ idx ] = bgColorNum;
	    this.fgColorBuf[ idx ] = fgColorNum;
	  }
	  if( (this.cursorCol + 1) < this.charCols ) {
	    setCursorPos( this.cursorRow, this.cursorCol + 1 );
	  } else if( (this.cursorRow + 1) < this.charRows ) {
	    setCursorPos( this.cursorRow + 1, 0 );
	  } else {
	    scrollUp();
	    setCursorPos( this.charRows - 1, 0 );
	  }
	}
    }
  }


  /*
   * Reads one or more characters from the keybord,
   * This method is thread safe and blocks the calling thread
   * until any key or the enter key is pressed,
   * depending on the parameter incharMode.
   * A negative timeout values means no timeout is used.
   *
   * The read characters can be found in inputCharBuf or inputLineBuf.
   */
  private void readFromKeyboard(
			DebugFld debugFld,
			boolean  incharMode,
			long     timeoutMillis ) throws InterruptedException
  {
    if( !incharMode && this.graphicsMode ) {
      setGraphicsMode( false );
    }
    if( debugFld != null ) {
      debugFld.fireSetCurDebugDataVisible( true, false );
    }
    ScreenFld screenFld = this.screenFld;
    if( screenFld != null ) {
      EventQueue.invokeLater( ()->screenFld.requestFocus() );
    }
    Thread.sleep( 20 );		// should prevent premature notifyAll()
    this.curKeyChar        = 0;
    this.bcBreakKeyPressed = false;
    Object inputMonitor    = new Object();
    synchronized( inputMonitor ) {
      this.inputLineBuf.setLength( 0 );
      this.inputCharBuf  = 0;
      this.inputCharMode = incharMode;
      this.inputMonitor  = inputMonitor;
      try {
	if( timeoutMillis > 0 ) {
	  inputMonitor.wait( timeoutMillis );
	} else {
	  inputMonitor.wait();
	}
      }
      catch( IllegalMonitorStateException ex ) {}
      finally {
	this.curKeyChar   = 0;
	this.inputMonitor = null;
	if( debugFld != null ) {
	  debugFld.fireSetCurDebugDataVisible( false, false );
	}
      }
    }
    if( this.bcBreakKeyPressed ) {
      throw new InterruptedException();
    }
  }


  private void scrollUp()
  {
    synchronized( this.threadLock ) {
      int dist = this.charCols;
      if( dist > 0 ) {
	System.arraycopy(
		this.charBuf,
		dist,
		this.charBuf,
		0,
		this.charBuf.length - dist );
	System.arraycopy(
		this.bgColorBuf,
		dist,
		this.bgColorBuf,
		0,
		this.bgColorBuf.length - dist );
	System.arraycopy(
		this.fgColorBuf,
		dist,
		this.fgColorBuf,
		0,
		this.fgColorBuf.length - dist );
	Arrays.fill(
		this.charBuf,
		this.charBuf.length - dist,
		this.charBuf.length,
		'\u0020' );
	Arrays.fill(
		this.bgColorBuf,
		this.bgColorBuf.length - dist,
		this.bgColorBuf.length,
		this.bgColorNum );
	Arrays.fill(
		this.fgColorBuf,
		this.fgColorBuf.length - dist,
		this.fgColorBuf.length,
		this.fgColorNum );
      }
    }
    setDirty();
  }


  private void setDirty()
  {
    ScreenFld screenFld = this.screenFld;
    if( screenFld != null ) {
      screenFld.repaint();
    }
  }


  private void setScreenSize(
			int charRows,
			int charCols,
			int fontSize,
			int wScreen,
			int hScreen,
			int margin )
  {
    this.margin = margin;

    if( this.image != null ) {
      if( (this.image.getWidth() != wScreen)
	  || (this.image.getHeight() != hScreen) )
      {
	this.image.flush();
	this.image = null;
      }
    }
    if( this.image == null ) {
      this.image = new BufferedImage(
				wScreen,
				hScreen,
				BufferedImage.TYPE_INT_RGB );
    }

    if( (charRows != this.charRows) || (charCols != this.charCols) ) {
      this.charRows = charRows;
      this.charCols = charCols;
      synchronized( this.threadLock ) {
	this.charBuf    = new char[ this.charRows * this.charCols ];
	this.bgColorBuf = new int[ this.charBuf.length ];
	this.fgColorBuf = new int[ this.charBuf.length ];
      }
      clear();
    }

    this.font  = new Font( Font.MONOSPACED, Font.BOLD, fontSize );
    Graphics g = this.mainFrm.getGraphics();
    if( g != null ) {
      try {
	g.setFont( this.font );
	int         wChar = 0;
	FontMetrics fm    = g.getFontMetrics();
	if( fm != null ) {
	  wChar = fm.charWidth( 'W' );
	}
	this.wChar = wChar + 1;
	this.hChar = this.font.getSize() + 1;
      }
      finally {
	g.dispose();
      }
    }
    this.width  = Math.max( (this.charCols * this.wChar), wScreen );
    this.height = Math.max( (this.charRows * this.hChar), hScreen );
    this.width  += (2 * this.margin);
    this.height += (2 * this.margin);
  }
}
