/*
 * (c) 2025 Jens Mueller
 *
 * JBASICODE
 *
 * Data of a variable declared with DIM instruction
 */

package jbasicode.bc;

import java.text.ParseException;
import java.util.Arrays;
import java.util.List;
import jbasicode.Main;
import jbasicode.ui.VarView;


public class DimVar
{
  private String   name;
  private int      dim1;
  private int      dim2;
  private Object[] values;


  public DimVar( String name, int dim )
  {
    this.name   = name;
    this.dim1   = dim;
    this.dim2   = -1;
    this.values = new Object[ dim + 1 ];
    initValues();
  }


  public DimVar( String name, int dim1, int dim2 )
  {
    this.name   = name;
    this.dim1   = dim1;
    this.dim2   = dim2;
    this.values = new Object[ (dim1 + 1) * (dim2 + 1) ];
    initValues();
  }


  public int calcStringSpace()
  {
    int stringSpace = 0;
    for( Object o : values ) {
      if( o instanceof String ) {
	int len = ((String) o).length();
	if( len > 0 ) {
	  stringSpace = stringSpace + len + 1;
	}
      }
    }
    return stringSpace;
  }


  public VarView createVarView()
  {
    StringBuilder declBuf = new StringBuilder();
    declBuf.append( this.name );
    declBuf.append( '(' );
    declBuf.append( this.dim1 );
    if( this.dim2 >= 0 ) {
      declBuf.append( ',' );
      declBuf.append( this.dim2 );
    }
    declBuf.append( ')' );

    StringBuilder valueBuf = new StringBuilder( 128 );
    for( int i = 0; i <= this.dim1; i++ ) {
      if( this.dim2 >= 0 ) {
	valueBuf.append( '[' );
	for( int k = 0; k <= this.dim2; k++ ) {
	  if( k > 0 ) {
	    valueBuf.append( ", " );
	  }
	  valueBuf.append( getElemValueView( (i * this.dim2) + k) );
	}
	valueBuf.append( ']' );
      } else {
	if( i > 0 ) {
	  valueBuf.append( ", " );
	}
	valueBuf.append( getElemValueView( i ) );
      }
    }

    return new VarView( declBuf.toString(), valueBuf.toString() );
  }


  public Object getValue( int idx )
  {
    return (idx >= 0) && (idx < this.values.length) ?
					this.values[ idx ] : null;
  }


  public Object getValue(
			BCRuntime    bcRuntime,
			List<Number> indices,
			int          srcIndex ) throws ParseException
  {
    int    idx   = getValueIndex( indices, srcIndex );
    Object value = this.values[ idx ];
    if( value == null ) {
      // should never be happen
      value = ExprParser.getDefaultVarValue( this.name );
      this.values[ idx ] = value;
    }
    return value;
  }


  public void setValue( int idx, Object value )
  {
    if( (idx >= 0) && (idx < this.values.length) )
      this.values[ idx ] = value;
  }


  public void setValue(
		List<Number>  indices,
		Object        value,
		int           srcIndex ) throws ParseException
  {
    this.values[ getValueIndex( indices, srcIndex ) ] = value;
  }


	/* --- private Methods --- */

  private String getElemValueView( int idx )
  {
    String rv = "";
    if( (idx >= 0) && (idx < this.values.length) ) {
      rv = VarView.toValueString( this.values[ idx ] );
    }
    return rv;
  }


  private int getValueIndex(
			List<Number> indices,
			int          srcIndex ) throws ParseException
  {
    int valueIdx = -1;
    int nIndices = indices.size();
    if( (nIndices == 1) && (this.dim2 < 0) ) {
      valueIdx = indices.get( 0 ).intValue();
      if( (valueIdx < 0) || (valueIdx > this.dim1) ) {
	throw new ParseException(
			ProgramParser.createMsg(
				"bc.index_out_of_range",
				valueIdx ),
			srcIndex );
      }
    } else if( (nIndices == 2) && (this.dim2 >= 0) ) {
      int idx1 = indices.get( 0 ).intValue();
      if( (idx1 < 0) || (idx1 > this.dim1) ) {
	throw new ParseException(
			ProgramParser.createMsg(
				"bc.index_out_of_range",
				idx1 ),
			srcIndex );
      }
      int idx2 = indices.get( 1 ).intValue();
      if( (idx2 < 0) || (idx2 > this.dim2) ) {
	throw new ParseException(
			ProgramParser.createMsg(
				"bc.index_out_of_range",
				idx1 ),
			srcIndex );
      }
      valueIdx= (idx1 * (this.dim2 + 1)) + idx2;
    }
    if( valueIdx < 0 ) {
      throw new ParseException(
			Main.getText( "bc.dimension_mismatch" ),
			srcIndex );
    }
    return valueIdx;
  }


  private void initValues()
  {
    Arrays.fill( this.values, ExprParser.getDefaultVarValue( this.name ) );
  }
}
