/*
 * (c) 2025 Jens Mueller
 *
 * JBASICODE
 *
 * BASICODE Expresson Parser
 */

package jbasicode.bc;

import java.text.CharacterIterator;
import java.text.NumberFormat;
import java.text.ParseException;
import java.text.ParsePosition;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import jbasicode.Main;


public class ExprParser
{
  private ProgramParser prgParser;
  private BCRuntime     bcRuntime;


  public ExprParser( ProgramParser prgParser, BCRuntime bcRuntime )
  {
    this.prgParser = prgParser;
    this.bcRuntime = bcRuntime;
  }


  public static Object getDefaultVarValue( String varName )
  {
    return varName.endsWith( "$" ) ? "" : 0;
  }


  public Object parseExpr( CharacterIterator iter ) throws ParseException
  {
    Object value = parseCondExpr( iter );
    for(;;) {
      boolean opOR  = false;
      boolean opAND = this.prgParser.checkAndParseToken( iter, "AND" );
      if( !opAND ) {
	opOR = this.prgParser.checkAndParseToken( iter, "OR" );
      }
      if( !opAND && !opOR ) {
	break;
      }
      if( !(value instanceof Number) ) {
	throw new ParseException(
			Main.getText( "bc.type_mismatch" ),
			iter.getIndex() );
      }
      Number vNum1  = (Number) value;
      Object value2 = parseCondExpr( iter );
      if( !(value2 instanceof Number) ) {
	throw new ParseException(
			Main.getText( "bc.type_mismatch" ),
			iter.getIndex() );
      }
      Number vNum2 = (Number) value2;
      if( (vNum1 instanceof Integer) && (vNum2 instanceof Integer) ) {
	if( opAND ) {
	  value = vNum1.intValue() & vNum2.intValue();
	} else if( opOR ) {
	  value = vNum1.intValue() | vNum2.intValue();
	}
      } else {
	boolean state1 = (vNum1.doubleValue() != 0.0);
	boolean state2 = (vNum2.doubleValue() != 0.0);
	if( opAND ) {
	  value = (state1 && state2 ? -1 : 0);
	} else if( opOR ) {
	  value = (state1 || state2 ? -1 : 0);
	}
      }
    }
    return value;
  }


  public static void parseLeftParenthesis( CharacterIterator iter )
						throws ParseException
  {
    if( ProgramParser.skipBlanks( iter ) != '(' ) {
      throw new ParseException(
			ProgramParser.createMsg( "bc.token_expected", '(' ),
			iter.getIndex() );
    }
    iter.next();
  }


  public Number parseNumericExpr( CharacterIterator iter )
						throws ParseException
  {
    Number rv    = null;
    Object value = parseExpr( iter );
    if( value != null ) {
      if( value instanceof Number ) {
	rv = (Number) value;
      }
    }
    if( rv == null ) {
      throw new ParseException(
			Main.getText( "bc.numeric_expr_excepted" ),
			iter.getIndex() );
    }
    return rv;
  }


  public static void parseRightParenthesis( CharacterIterator iter )
						throws ParseException
  {
    if( ProgramParser.skipBlanks( iter ) != ')' ) {
      throw new ParseException(
			ProgramParser.createMsg( "bc.token_expected", ')' ),
			iter.getIndex() );
    }
    iter.next();
  }


  public String parseStringExpr( CharacterIterator iter )
						throws ParseException
  {
    String rv    = null;
    Object value = parseExpr( iter );
    if( value != null ) {
      if( value instanceof String ) {
	rv = (String) value;
      }
    }
    if( rv == null ) {
      throw new ParseException(
			Main.getText( "bc.string_expr_excepted" ),
			iter.getIndex() );
    }
    return rv;
  }


  public static Number readNumber(
			CharacterIterator iter,
			boolean           ignoreCase,
			boolean           allowNeg ) throws ParseException
  {
    Number        value  = null;
    StringBuilder buf    = new StringBuilder();
    long          lValue = 0;
    boolean       isInt  = false;
    boolean       isNeg  = false;
    char          ch     = ProgramParser.skipBlanks( iter );
    int           index  = iter.getIndex();
    if( allowNeg && (ch == '-') ) {
      buf.append( ch );
      isNeg = true;
      ch    = iter.next();
    }
    while( ProgramParser.isDigit( ch ) ) {
      buf.append( ch );
      lValue = (lValue * 10) + (ch - '0');
      if( lValue > (long) Integer.MAX_VALUE ) {
	isInt = false;
      } else {
	isInt  = true;
      }
      ch = iter.next();
    }
    if( ch == '.' ) {
      buf.append( ch );
      isInt = false;
      ch    = iter.next();
    }
    while( ProgramParser.isDigit( ch ) ) {
      buf.append( ch );
      ch = iter.next();
    }
    if( (ch == 'E') || (ignoreCase && (ch == 'e')) ) {
      buf.append( Character.toUpperCase( ch ) );
      isInt = false;
      ch    = iter.next();
      if( (ch == '+') || (ch == '-') ) {
	buf.append( ch );
	ch = iter.next();
      }
      while( ProgramParser.isDigit( ch ) ) {
	buf.append( ch );
	ch = iter.next();
      }
    }
    try {
      if( isInt ) {
	if( isNeg ) {
	  lValue = -lValue;
	}
	value = Integer.valueOf( (int) lValue );
      } else {
	value = Double.valueOf( buf.toString() );
      }
    }
    catch( NumberFormatException ex ) {}
    if( value == null ) {
      iter.setIndex( index );
    }
    return value;
  }


	/* --- Private Methods --- */

  private Object parseCondExpr( CharacterIterator iter )
						throws ParseException
  {
    Object value = parseAddExpr( iter );
    char   ch    = ProgramParser.skipBlanks( iter );
    while( (ch == '<') || (ch == '=') || (ch == '>') ) {
      int    opIdx = iter.getIndex();
      String op    = String.valueOf( ch );
      ch = iter.next();
      if( (ch == '<') || (ch == '=') || (ch == '>') ) {
	op += ch;
	iter.next();
      }
      if( op.equals( "<<" )
	  || op.equals( "=<" )
	  || op.equals( "==" )
	  || op.equals( "=>" )
	  || op.equals( "><" )
	  || op.equals( ">>" ) )
      {
	throw new ParseException(
			ProgramParser.createMsg( "bc.invalid_operator", op ),
			opIdx );
      }
      Object value1 = value;
      Object value2 = parseAddExpr( iter );
      value         = 0;
      if( (value1 instanceof Number) && (value2 instanceof Number) ) {
	if( (value1 instanceof Integer) && (value2 instanceof Integer) ) {
	  switch( op ) {
	    case "<":
	      if( ((Integer) value1).intValue()
				< ((Integer) value2).intValue() )
	      {
		value = -1;
	      }
	      break;
	    case "<=":
	      if( ((Integer) value1).intValue()
				<= ((Integer) value2).intValue() )
	      {
		value = -1;
	      }
	      break;
	    case "<>":
	      if( ((Integer) value1).intValue()
				!= ((Integer) value2).intValue() )
	      {
		value = -1;
	      }
	      break;
	    case "=":
	      if( ((Integer) value1).intValue()
				== ((Integer) value2).intValue() )
	      {
		value = -1;
	      }
	      break;
	    case ">=":
	      if( ((Integer) value1).intValue()
				>= ((Integer) value2).intValue() )
	      {
		value = -1;
	      }
	      break;
	    case ">":
	      if( ((Integer) value1).intValue()
				> ((Integer) value2).intValue() )
	      {
		value = -1;
	      }
	      break;
	    default:
	      // should never be reached
	      throw new ParseException(
			Main.getText( "bc.invalid_operator" ),
			opIdx );
	  }
	} else {
	  switch( op ) {
	    case "<":
	      if( ((Number) value1).doubleValue()
				< ((Number) value2).doubleValue() )
	      {
		value = -1;
	      }
	      break;
	    case "<=":
	      if( ((Number) value1).doubleValue()
				<= ((Number) value2).doubleValue() )
	      {
		value = -1;
	      }
	      break;
	    case "<>":
	      if( ((Number) value1).doubleValue()
				!= ((Number) value2).doubleValue() )
	      {
		value = -1;
	      }
	      break;
	    case "=":
	      if( ((Number) value1).doubleValue()
				== ((Number) value2).doubleValue() )
	      {
		value = -1;
	      }
	      break;
	    case ">=":
	      if( ((Number) value1).doubleValue()
				>= ((Number) value2).doubleValue() )
	      {
		value = -1;
	      }
	      break;
	    case ">":
	      if( ((Number) value1).doubleValue()
				> ((Number) value2).doubleValue() )
	      {
		value = -1;
	      }
	      break;
	    default:
	      // should never be reached
	      throw new ParseException(
			Main.getText( "bc.invalid_operator" ),
			opIdx );
	  }
	}
      } else if( (value1 instanceof String) && (value2 instanceof String) ) {
	int result = ((String) value1).compareTo( (String) value2 );
	switch( op ) {
	  case "<":
	    if( result < 0 ) {
	      value = -1;
	    }
	    break;
	  case "<=":
	    if( result <= 0 ) {
	      value = -1;
	    }
	    break;
	  case "<>":
	    if( result != 0 ) {
	      value = -1;
	    }
	    break;
	  case "=":
	    if( result == 0 ) {
	      value = -1;
	    }
	    break;
	  case ">=":
	    if( result >= 0 ) {
	      value = -1;
	    }
	    break;
	  case ">":
	    if( result > 0 ) {
	      value = -1;
	    }
	    break;
	  default:
	    // should never be reached
	    throw new ParseException(
			Main.getText( "bc.invalid_operator" ),
			opIdx );
	}
      } else {
	throw new ParseException(
			Main.getText( "bc.type_mismatch" ),
			opIdx );
      }
      ch = ProgramParser.skipBlanks( iter );
    }
    return value;
  }


  private Object parseAddExpr( CharacterIterator iter ) throws ParseException
  {
    Object value = parseMulExpr( iter );
    char   ch    = ProgramParser.skipBlanks( iter );
    while( (ch == '+') || (ch == '-') ) {
      iter.next();
      Object value2 = parseMulExpr( iter );
      if( (value instanceof Number) && (value2 instanceof Number) ) {
	if( (value instanceof Integer) && (value2 instanceof Integer) ) {
	  if( ch == '+' ) {
	    value = ((Integer) value).intValue()
					+ ((Integer) value2).intValue();
	  } else {
	    value = ((Integer) value).intValue()
					- ((Integer) value2).intValue();
	  }
	} else {
	  if( ch == '+' ) {
	    value = ((Number) value).doubleValue()
					+ ((Number) value2).doubleValue();
	  } else {
	    value = ((Number) value).doubleValue()
					- ((Number) value2).doubleValue();
	  }
	}
      } else if( (value instanceof String) && (value2 instanceof String) ) {
	if( ch == '-' ) {
	  throw new ParseException(
			Main.getText( "bc.invalid_operator" ),
			iter.getIndex() );
	}
	value = ((String) value) + ((String) value2);
      } else {
	throw new ParseException(
			Main.getText( "bc.type_mismatch" ),
			iter.getIndex() );
      }
      ch = ProgramParser.skipBlanks( iter );
    }
    return value;
  }


  private Object parseMulExpr( CharacterIterator iter ) throws ParseException
  {
    Object value = parseExpExpr( iter );
    char   ch    = ProgramParser.skipBlanks( iter );
    while( (ch == '*') || (ch == '/') ) {
      iter.next();
      Object value2 = parseExpExpr( iter );
      if( (value instanceof Number) && (value2 instanceof Number) ) {
	boolean done = false;
	if( (value instanceof Integer) && (value2 instanceof Integer) ) {
	  if( ch == '*' ) {
	    value = ((Integer) value).intValue()
					* ((Integer) value2).intValue();
	    done  = true;
	  } else {
	    int v1 = ((Integer) value).intValue();
	    int v2 = ((Integer) value2).intValue();
	    if( this.bcRuntime != null ) {
	      if( v2 == 0 ) {
		throw new ParseException(
				Main.getText( "bc.division_by_zero" ),
				iter.getIndex() );
	      }
	      int result = v1 / v2;
	      if( (result * v2) == v1 ) {
		value = result;
		done  = true;
	      }
	    } else {
	      // result not relevant
	      done = true;
	    }
	  }
	}
	if( !done ) {
	  if( ch == '*' ) {
	    value = ((Number) value).doubleValue()
					* ((Number) value2).doubleValue();
	  } else {
	    double v2 = ((Number) value2).doubleValue();
	    if( this.bcRuntime != null ) {
	      if( v2 == 0.0 ) {
		throw new ParseException(
			Main.getText( "bc.division_by_zero" ),
			iter.getIndex() );
	      }
	      value = ((Number) value).doubleValue() / v2;
	    }
	  }
	}
      } else {
	throw new ParseException(
			Main.getText( "bc.type_mismatch" ),
			iter.getIndex() );
      }
      ch = ProgramParser.skipBlanks( iter );
    }
    return value;
  }


  private Object parseExpExpr( CharacterIterator iter )
						throws ParseException
  {
    Object value = parseUnaryExpr( iter );
    char   ch    = ProgramParser.skipBlanks( iter );
    while( ch == '^' ) {
      iter.next();
      Object value2 = parseUnaryExpr( iter );
      if( (value instanceof Number) && (value2 instanceof Number) ) {
	if( this.bcRuntime != null ) {
	  boolean done = false;
	  if( (value instanceof Integer) && (value2 instanceof Integer) ) {
	    long n = ((Number) value2).intValue();
	    if( n == 0 ) {
	      value = 1;
	      done  = true;
	    } else if( n > 0 ) {
	      long v = ((Number) value).longValue();
	      if( v == -1 ) {
		if( (n & 1) != 0 ) {
		  value = -1;
		} else {
		  value = 1;
		}
		done = true;
	      } else if( v == 1 ) {
		value = 1;
		done  = true;
	      } else {
		--n;
		while( n > 0 ) {
		  v *= ((Number) value).longValue();
		  if( Math.abs( v ) > (long) Integer.MAX_VALUE ) {
		    break;
		  }
		  --n;
		}
		if( n == 0 ) {
		  value = (int) v;
		  done  = true;
		}
	      }
	    }
	  }
	  if( !done ) {
	    value = Math.pow(
			  ((Number) value).doubleValue(),
			  ((Number) value2).doubleValue() );
	  }
	}
      } else {
	throw new ParseException(
			Main.getText( "bc.type_mismatch" ),
			iter.getIndex() );
      }
      ch = ProgramParser.skipBlanks( iter );
    }
    return value;
  }


  private Object parseUnaryExpr( CharacterIterator iter )
						throws ParseException
  {
    boolean opNeg    = false;
    boolean opNOT    = false;
    boolean enclosed = false;
    if( ProgramParser.skipBlanks( iter ) == '-' ) {
      opNeg = true;
      iter.next();
    } else if( this.prgParser.checkAndParseToken( iter, "NOT" ) ) {
      opNOT = true;
      enclosed = this.prgParser.checkAndParseToken( iter, "(" );
      if( !enclosed ) {
	this.prgParser.putSpecViolatedWarning(
			ProgramParser.createMsg( "bc.token_expected", '(' ),
			iter.getIndex() );
      }
    }
    Object value = enclosed ? parseExpr( iter ) : parsePrimExpr( iter );
    if( opNeg ) {
      if( value instanceof Number ) {
	if( value instanceof Integer ) {
	  value = -((Integer) value).intValue();
	} else {
	  value = -((Number) value).doubleValue();
	}
      } else {
	throw new ParseException(
			Main.getText( "bc.type_mismatch" ),
			iter.getIndex() );
      }
    } else if( opNOT ) {
      if( value instanceof Number ) {
	if( value instanceof Integer ) {
	  value = ~((Integer) value).intValue();
	} else {
	  if( ((Number) value).doubleValue() == 0.0 ) {
	    value = -1;
	  } else {
	    value = 0;
	  }
	}
      } else {
	throw new ParseException(
			Main.getText( "bc.type_mismatch" ),
			iter.getIndex() );
      }
    }
    if( enclosed ) {
      parseRightParenthesis( iter );
    }
    return value;
  }


  private Object parsePrimExpr( CharacterIterator iter )
						throws ParseException
  {
    Object value = null;
    char   ch    = ProgramParser.skipBlanks( iter );
    if( ch == '(' ) {
      iter.next();
      value = parseExpr( iter );
      parseRightParenthesis( iter );
    } else if( ch == '\"' ) {
      value = this.prgParser.readStringLiteral( iter );
    } else if( (ch == '.') || ProgramParser.isDigit( ch ) ) {
      value = readNumber( iter, this.prgParser.getIgnoreCase(), false );
      if( value == null ) {
	throw new ParseException(
			Main.getText( "bc.invalid_number" ),
			iter.getIndex() );
      }
    } else if( ProgramParser.isLetter( ch ) ) {
      value = parseFunctionOrVar( iter );
    }
    if( value == null ) {
      throw new ParseException(
			Main.getText( "bc.expr_excepted" ),
			iter.getIndex() );
    }
    return value;
  }


  private Object parseFunctionOrVar( CharacterIterator iter )
						throws ParseException
  {
    Object value      = null;
    String identifier = this.prgParser.readIdentifier( iter );
    if( identifier.startsWith( "FN" ) ) {
      if( identifier.length() > 2 ) {
	identifier = identifier.substring( 2 );
      } else {
	identifier = this.prgParser.readIdentifier( iter );
	if( identifier == null ) {
	  throw new ParseException(
			Main.getText( "bc.funcname_expected" ),
			iter.next() );
	}
      }
      identifier = this.prgParser.checkIdentifier(
						identifier,
						iter.getIndex() );
      if( this.bcRuntime != null ) {
	value = this.bcRuntime.callUserFunction( identifier, this, iter );
      } else {
	if( this.prgParser.checkAndParseToken( iter, "(" ) ) {
	  parseExpr( iter );
	  parseRightParenthesis( iter );
	}
	value = 0;
      }
    } else {
      switch( identifier ) {
	case "ABS":
	  value = parseABS( iter );
	  break;
	case "ASC":
	  value = parseASC( iter );
	  break;
	case "ATN":
	  value = parseATN( iter );
	  break;
	case "CHR$":
	  value = parseStringCHR( iter );
	  break;
	case "COS":
	  value = parseCOS( iter );
	  break;
	case "EXP":
	  value = parseEXP( iter );
	  break;
	case "INT":
	  value = parseINT( iter );
	  break;
	case "LEFT$":
	  value = parseStringLEFT( iter );
	  break;
	case "LEN":
	  value = parseLEN( iter );
	  break;
	case "LOG":
	  value = parseLOG( iter );
	  break;
	case "MID$":
	  value = parseStringMID( iter );
	  break;
	case "RIGHT$":
	  value = parseStringRIGHT( iter );
	  break;
	case "SGN":
	  value = parseSGN( iter );
	  break;
	case "SIN":
	  value = parseSIN( iter );
	  break;
	case "SQR":
	  value = parseSQR( iter );
	  break;
	case "TAN":
	  value = parseTAN( iter );
	  break;
	case "VAL":
	  value = parseVAL( iter );
	  break;
	default:
	  value = parseVar(
			iter,
			this.prgParser.checkIdentifier(
						identifier,
						iter.getIndex() ) );
      }
    }
    return value;
  }


  private Number parseABS( CharacterIterator iter )
						throws ParseException
  {
    parseLeftParenthesis( iter );
    Number value = parseNumericExpr( iter );
    parseRightParenthesis( iter );
    if( value instanceof Integer ) {
      value = Integer.valueOf( Math.abs( ((Integer) value).intValue() ) );
    } else {
      value = Double.valueOf( Math.abs( ((Number) value).doubleValue() ) );
    }
    return value;
  }


  private int parseASC( CharacterIterator iter )
						throws ParseException
  {
    int rv = 0;
    parseLeftParenthesis( iter );
    String text = parseStringExpr( iter );
    if( this.bcRuntime != null ) {
      if( text.isEmpty() ) {
	throw new ParseException(
			Main.getText( "bc.string_empty" ),
			iter.getIndex() );
      }
      rv = text.charAt( 0 );
    }
    parseRightParenthesis( iter );
    return rv;
  }


  private double parseATN( CharacterIterator iter ) throws ParseException
  {
    double rv = 0.0;
    parseLeftParenthesis( iter );
    Number value = parseNumericExpr( iter );
    if( this.bcRuntime != null ) {
      rv = Math.atan( ((Number) value).doubleValue() );
    }
    parseRightParenthesis( iter );
    return rv;
  }


  private double parseCOS( CharacterIterator iter ) throws ParseException
  {
    double rv = 0.0;
    parseLeftParenthesis( iter );
    Number value = parseNumericExpr( iter );
    if( this.bcRuntime != null ) {
      rv = Math.cos( ((Number) value).doubleValue() );
    }
    parseRightParenthesis( iter );
    return rv;
  }


  private double parseEXP( CharacterIterator iter ) throws ParseException
  {
    double rv = 0.0;
    parseLeftParenthesis( iter );
    Number value = parseNumericExpr( iter );
    if( this.bcRuntime != null ) {
      rv = Math.exp( ((Number) value).doubleValue() );
    }
    parseRightParenthesis( iter );
    return rv;
  }


  private int parseINT( CharacterIterator iter )
						throws ParseException
  {
    parseLeftParenthesis( iter );
    Number value = parseNumericExpr( iter );
    parseRightParenthesis( iter );
    return value instanceof Integer ?
	value.intValue()
	: ((int) Math.round( Math.floor( value.doubleValue() ) ));
  }


  private int parseLEN( CharacterIterator iter ) throws ParseException
  {
    parseLeftParenthesis( iter );
    String text = parseStringExpr( iter );
    parseRightParenthesis( iter );
    return text.length();
  }


  private double parseLOG( CharacterIterator iter )
						throws ParseException
  {
    double rv = 0.0;
    parseLeftParenthesis( iter );
    Number value = parseNumericExpr( iter );
    if( this.bcRuntime != null ) {
      rv = Math.log( ((Number) value).doubleValue() );
    }
    parseRightParenthesis( iter );
    return rv;
  }


  private int parseSGN( CharacterIterator iter ) throws ParseException
  {
    int rv = 0;
    parseLeftParenthesis( iter );
    Number value = parseNumericExpr( iter );
    parseRightParenthesis( iter );
    if( value instanceof Integer ) {
      int v = ((Integer) value).intValue();
      if( v < 0 ) {
	rv = -1;
      } else if( v > 0 ) {
	rv = 1;
      }
    } else {
      double v = ((Number) value).doubleValue();
      if( v < 0.0 ) {
	rv = -1;
      } else if( v > 0.0 ) {
	rv = 1;
      }
    }
    return rv;
  }


  private double parseSIN( CharacterIterator iter )
						throws ParseException
  {
    double rv = 0.0;
    parseLeftParenthesis( iter );
    Number value = parseNumericExpr( iter );
    if( this.bcRuntime != null ) {
      rv = Math.sin( ((Number) value).doubleValue() );
    }
    parseRightParenthesis( iter );
    return rv;
  }


  private double parseSQR( CharacterIterator iter )
						throws ParseException
  {
    double rv = 0.0;
    parseLeftParenthesis( iter );
    Number value = parseNumericExpr( iter );
    if( this.bcRuntime != null ) {
      rv = Math.sqrt( ((Number) value).doubleValue() );
    }
    parseRightParenthesis( iter );
    return rv;
  }


  private double parseTAN( CharacterIterator iter )
						throws ParseException
  {
    double rv = 0.0;
    parseLeftParenthesis( iter );
    Number value = parseNumericExpr( iter );
    if( this.bcRuntime != null ) {
      rv = Math.tan( ((Number) value).doubleValue() );
    }
    parseRightParenthesis( iter );
    return rv;
  }


  private Number parseVAL( CharacterIterator iter ) throws ParseException
  {
    parseLeftParenthesis( iter );
    String text = parseStringExpr( iter ).trim();
    parseRightParenthesis( iter );
    Number rv = Integer.valueOf( 0 );
    if( this.bcRuntime != null ) {
      NumberFormat numFmt = NumberFormat.getInstance( Locale.ROOT );
      numFmt.setGroupingUsed( false );
      numFmt.setParseIntegerOnly( false );
      Number value = numFmt.parse( text, new ParsePosition( 0 ) );
      if( value != null ) {
	if( value instanceof Integer ) {
	  rv = (Integer) value;
	} else if( value instanceof Long ) {
	  if( (value.longValue() >= Integer.MIN_VALUE)
	      && (value.longValue() <= Integer.MAX_VALUE) )
	  {
	    rv = Integer.valueOf( value.intValue() );
	  } else {
	    rv = Double.valueOf( value.doubleValue() );
	  }
	} else if( value instanceof Double ) {
	  rv = (Double) value;
	} else {
	  rv = Double.valueOf( value.doubleValue() );
	}
      }
    }
    return rv;
  }


  private String parseStringCHR( CharacterIterator iter )
						throws ParseException
  {
    parseLeftParenthesis( iter );
    int value = parseNumericExpr( iter ).intValue();
    int index = iter.getIndex();
    parseRightParenthesis( iter );
    if( (this.bcRuntime != null)
	&& ((value < 0) || (value > 0xFFFF)
	    || !Character.isDefined( (char) value )) )
    {
      throw new ParseException(
			ProgramParser.createMsg(
				"bc.illegal_arg_value",
				value ),
			index );
    }
    return String.valueOf( (char) value );
  }


  private String parseStringLEFT( CharacterIterator iter )
						throws ParseException
  {
    parseLeftParenthesis( iter );
    String text = parseStringExpr( iter );
    parseComma( iter );
    Number count = parseNumericExpr( iter );
    int    index = iter.getIndex();
    parseRightParenthesis( iter );
    if( this.bcRuntime != null ) {
      int n = count.intValue();
      if( n < 0 ) {
	throw new ParseException(
			ProgramParser.createMsg(
				"bc.illegal_arg_value",
				count ),
			index );
      }
      if( n < text.length() ) {
	text = text.substring( 0, n );
      }
    }
    return text;
  }


  private String parseStringMID( CharacterIterator iter )
						throws ParseException
  {
    parseLeftParenthesis( iter );
    String text = parseStringExpr( iter );
    parseComma( iter );
    Number posArg = parseNumericExpr( iter );
    int    pos    = posArg.intValue() - 1;
    if( (this.bcRuntime != null) && (pos < 0) ) {
      throw new ParseException(
			ProgramParser.createMsg(
				"bc.illegal_arg_value",
				posArg ),
			iter.getIndex() );
    }
    if( this.prgParser.checkAndParseToken( iter, "," ) ) {
      Number countArg = parseNumericExpr( iter );
      if( this.bcRuntime != null ) {
	int count = countArg.intValue();
	if( count < 0 ) {
	  throw new ParseException(
			ProgramParser.createMsg(
				"bc.illegal_arg_value",
				countArg ),
			iter.getIndex() );
	}
	int len = text.length();
	if( (count > 0) && (pos < len) ) {
	  if( (pos + count) < len ) {
	    text = text.substring( pos, pos + count );
	  } else {
	    text = text.substring( pos );
	  }
	} else {
	  text = "";
	}
      } 
    } else {
      if( this.bcRuntime != null ) {
	int len = text.length();
	if( pos < len ) {
	  text = text.substring( pos );
	} else {
	  text = "";
	}
      }
    }
    parseRightParenthesis( iter );
    return text;
  }


  private String parseStringRIGHT( CharacterIterator iter )
						throws ParseException
  {
    parseLeftParenthesis( iter );
    String text = parseStringExpr( iter );
    parseComma( iter );
    Number count = parseNumericExpr( iter );
    int    index = iter.getIndex();
    parseRightParenthesis( iter );
    if( this.bcRuntime != null ) {
      int n = count.intValue();
      if( n < 0 ) {
	throw new ParseException(
			ProgramParser.createMsg(
				"bc.illegal_arg_value",
				count ),
			index );
      }
      int len = text.length();
      if( n < len ) {
	text = text.substring( len - n );
      }
    }
    return text;
  }


  private Object parseVar(
			CharacterIterator iter,
			String            varName ) throws ParseException
  {
    Object value = null;
    int    index = iter.getIndex();
    if( ProgramParser.skipBlanks( iter ) == '(' ) {
      iter.next();
      List<Number> indices = new ArrayList<>();
      for(;;) {
	indices.add( parseNumericExpr( iter ) );
	if( ProgramParser.skipBlanks( iter ) != ',' ) {
	  break;
	}
	iter.next();
      }
      parseRightParenthesis( iter );
      if( this.bcRuntime != null ) {
	value = this.bcRuntime.getVarValue(
					varName,
					indices,
					iter.getIndex() );
      }
    } else {
      if( this.bcRuntime != null ) {
	value = this.bcRuntime.getVarValue( varName, iter.getIndex() );
      }
    }
    return value != null ? value : getDefaultVarValue( varName );
  }


  private static void parseComma( CharacterIterator iter )
						throws ParseException
  {
    if( ProgramParser.skipBlanks( iter ) != ',' ) {
      throw new ParseException(
			ProgramParser.createMsg( "bc.token_expected", ',' ),
			iter.getIndex() );
    }
    iter.next();
  }
}
