/*
 * (c) 2025 Jens Mueller
 *
 * JBASICODE
 *
 * BASICODE Program Parser
 */

package jbasicode.bc;

import java.text.CharacterIterator;
import java.text.MessageFormat;
import java.text.ParseException;
import java.text.StringCharacterIterator;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import jbasicode.Main;
import jbasicode.ui.MainFrm;


public class ProgramParser
{
  public static final boolean DEFAULT_IGNORE_CASE        = false;
  public static final boolean DEFAULT_WARN_NON_ASCII     = true;
  public static final boolean DEFAULT_WARN_SPEC_VIOLATED = true;

  public static final String PROP_IGNORE_CASE        = "bc.ignore_case";
  public static final String PROP_WARN_NON_ASCII     = "warn.non_ascii";
  public static final String PROP_WARN_SPEC_VIOLATED = "warn.spec_violated";

  private static String[] sortedReservedWords = {
	"ABS", "ASC", "ATN", "CHR$", "COS",
	"DATA", "DEF", "DIM", "EXP", "FOR", "GOSUB", "GOTO",
	"IF", "INPUT", "INT", "LEFT$", "LEN", "LET", "LOG",
	"MID$", "NEXT", "ON", "PRINT",
	"READ", "REM", "RESTORE", "RETURN", "RIGHT$",
	"SGN", "SIN", "SQR", "STEP", "TAB", "TAN", "THEN", "TO", "VAL" };

  private static String[] sortedIllegalIdentifiers = {
	"AS", "AT", "DI", "DI$", "DO", "DO$", "DS", "DS$",
	"EI", "EI$", "EL", "ER", "FN", "GO", "GO$", "GR", "HC",
	"LN", "MA", "MP", "PI", "SQ", "SQ$", "ST", "ST$",
	"TI", "TI$", "TO$" };

  private static int[] sortedSysLineNums = {
	20, 100, 110, 120, 150, 200, 210, 220, 250, 260, 270, 280,
	300, 310, 330, 350, 360, 400, 450, 500, 540, 560, 580,
	600, 620, 630, 650, 950 };

  private MainFrm              mainFrm;
  private BCRuntime            bcRuntime;
  private boolean              aborted;
  private boolean              abortFired;
  private boolean              ignoreCase;
  private boolean              ignoreColon;
  private boolean              errState;
  private boolean              warnNonAsciiEnabled;
  private boolean              warnSpecViolatedEnabled;
  private boolean              warnEnabled;
  private boolean              warnState;
  private int                  curLineIdx;
  private int                  dataReadIdx;
  private List<SourceLine>     srcLines;
  private List<Object>         dataValues;
  private Map<Integer,Integer> bcLineNum2LineIdx;
  private Map<Integer,Integer> bcLineNum2DataIdx;
  private Set<String>          shownMessages;


  public ProgramParser( MainFrm mainFrm, BCRuntime bcRuntime )
  {
    this.mainFrm                 = mainFrm;
    this.bcRuntime               = bcRuntime;
    this.aborted                 = false;
    this.abortFired              = false;
    this.ignoreCase              = getIgnoreCaseSetting();
    this.ignoreColon             = false;
    this.errState                = false;
    this.warnNonAsciiEnabled     = getWarnNonAsciiSetting();
    this.warnSpecViolatedEnabled = getWarnSpecViolatedSetting();
    this.warnState               = false;
    this.warnEnabled             = true;
    this.curLineIdx              = 0;
    this.dataReadIdx             = 0;
    this.dataValues              = new ArrayList<>();
    this.srcLines                = new ArrayList<>();
    this.bcLineNum2LineIdx       = new HashMap<>();
    this.bcLineNum2DataIdx       = new HashMap<>();
    this.shownMessages           = new TreeSet<>();
  }


  public boolean checkAndParseToken(
				CharacterIterator iter,
				String            token )
						throws ParseException
  {
    boolean rv  = true;
    char    ch  = toSyntaxCase( skipBlanks( iter ) );
    int     pos = iter.getIndex();
    int     len = token.length();
    for( int i = 0; i < len; i++ ) {
      if( ch != token.charAt( i ) ) {
	rv = false;
	iter.setIndex( pos );
	break;
      }
      ch = toSyntaxCase( iter.next() );
    }
    return rv;
  }


  public String checkIdentifier(
			String identifier,
			int    srcIndex ) throws ParseException
  {
    String upperIdentifier = identifier.toUpperCase();
    if( upperIdentifier.startsWith( "O" )
	|| (!this.ignoreCase && !identifier.equals( upperIdentifier )) )
    {
      putSpecViolatedWarning(
			createMsg( "bc.illegal_identifier", identifier ),
			srcIndex );
    }
    identifier = upperIdentifier;
    ensureNoReservedKeyword( identifier, srcIndex );
    if( Arrays.binarySearch( sortedIllegalIdentifiers, identifier ) >= 0 ) {
      putSpecViolatedWarning(
			createMsg( "bc.illegal_identifier", identifier ),
			srcIndex );
    }
    int len = identifier.length();
    if( (len > 3) || ((len > 2) && !identifier.endsWith( "$" )) ) {
      putSpecViolatedWarning(
		createMsg( "bc.identifier_longer_than_2", identifier ),
		srcIndex );
    }
    return identifier;
  }


  public static String createMsg(
				String    textKey,
				Object... args )
  {
    /*
     * Convert arguments to strings so that MessageFormat does not do this.
     * MessageFormat uses grouping, which would interfere here.
     */
    for( int i = 0; i < args.length; i++ ) {
      args[ i ] = args[ i ].toString();
    }
    return MessageFormat.format( Main.getText( textKey ), args );
  }


  public String createNextDataReadValuesView()
  {
    String rv   = "";
    int    size = this.dataValues.size();
    int    pos  = this.dataReadIdx;
    if( pos < size ) {
      StringBuilder buf = new StringBuilder( 256 );
      while( pos < size ) {
	int len = buf.length();
	if( len > 0 ) {
	  buf.append( ", " );
	}
	if( len > 200 ) {
	  buf.append( "..." );
	  break;
	}
	Object value = this.dataValues.get( pos++ );
	if( value instanceof String ) {
	  buf.append( '\"' );
	  buf.append( (String) value );
	  buf.append( '\"' );
	} else {
	  buf.append( value.toString() );
	}
      }
      rv = buf.toString();
    }
    return rv;
  }


  public static void ensureNoReservedKeyword(
				String text,
				int    index ) throws ParseException
  {
    if( Arrays.binarySearch( sortedReservedWords, text ) >= 0 ) {
      throw new ParseException(
			createMsg( "bc.reserved_keyword", text ),
			index );
    }
  }


  public void fireAbort()
  {
    this.abortFired = true;
  }


  public boolean getIgnoreCase()
  {
    return this.ignoreCase;
  }


  public static boolean getIgnoreCaseSetting()
  {
    return Main.getBooleanProperty(
				PROP_IGNORE_CASE,
				DEFAULT_IGNORE_CASE );
  }


  public Integer getLineIndexByBCLineNumber( int bcLineNum )
  {
    return bcLineNum2LineIdx.get( bcLineNum );
  }


  public boolean getWarningsEnabled()
  {
    return this.warnEnabled;
  }


  public boolean getWarnNonAsciiEnabled()
  {
    return this.warnNonAsciiEnabled;
  }


  public static boolean getWarnNonAsciiSetting()
  {
    return Main.getBooleanProperty(
				PROP_WARN_NON_ASCII,
				DEFAULT_WARN_NON_ASCII );
  }


  public boolean getWarnSpecViolatedEnabled()
  {
    return this.warnSpecViolatedEnabled;
  }


  public static boolean getWarnSpecViolatedSetting()
  {
    return Main.getBooleanProperty(
				PROP_WARN_SPEC_VIOLATED,
				DEFAULT_WARN_SPEC_VIOLATED );
  }


  public boolean hasBeenAborted()
  {
    return this.aborted;
  }


  public boolean hasErrors()
  {
    return this.errState;
  }


  public boolean hasWarnings()
  {
    return this.warnState;
  }


  public static boolean isDigit( char ch )
  {
    return (ch >= '0') && (ch <= '9');
  }


  public static boolean isLetter( char ch )
  {
    return ((ch >= 'A') && (ch <= 'Z')) || ((ch >= 'a') && (ch <= 'z'));
  }


  public void parse( String program ) throws InterruptedException
  {
    program = trimTrailingBlankLines( program );
    try {
      this.mainFrm.fireClearLog();
      this.abortFired  = false;
      this.errState    = false;
      this.warnState   = false;
      this.dataReadIdx = 0;
      this.dataValues.clear();
      this.srcLines.clear();
      this.bcLineNum2LineIdx.clear();
      this.bcLineNum2DataIdx.clear();
      this.shownMessages.clear();

      if( program != null ) {

	// parse line numbers and build list and map containing the lines
	int len           = program.length();
	int recentLineNum = -1;
	int pos           = 0;
	this.curLineIdx   = 0;
	while( pos < len ) {
	  checkAbort();
	  int eol = program.indexOf( '\n', pos );
	  if( eol < pos ) {
	    scanLine( this.curLineIdx + 1, program.substring( pos ) );
	    break;
	  }
	  scanLine( this.curLineIdx + 1, program.substring( pos, eol ) );
	  pos = eol + 1;
	  if( this.curLineIdx < this.srcLines.size() ) {
	    Integer bcLineNum = this.srcLines.get(
					this.curLineIdx ).getBCLineNumber();
	    if( bcLineNum != null ) {
	      if( bcLineNum.intValue() <= recentLineNum ) {
		throw new ParseException(
				createMsg(
					"bc.line_num_not_gt_prev",
					bcLineNum ),
				-1 );
	      }
	      recentLineNum = bcLineNum.intValue();
	    }
	    this.curLineIdx++;
	  }
	}

	// parse and execute lines
	this.curLineIdx = 0;
	int charIdx     = 0;
	int nLines      = this.srcLines.size();
	while( this.curLineIdx < nLines ) {
	  checkAbort();
	  try {
	    try {
	      parseLine( this.srcLines.get( this.curLineIdx ), charIdx );
	      this.curLineIdx++;
	    }
	    catch( GotoException ex ) {
	      int bcLineNum  = ex.getBCLineNumber();
	      int newLineNum = -1;
	      Integer idx = this.bcLineNum2LineIdx.get( bcLineNum );
	      if( idx == null ) {
		for( int i = 0; i < nLines; i++ ) {
		  Integer tmpLineNum = this.srcLines.get( i )
							.getBCLineNumber();
		  if( tmpLineNum != null ) {
		    if( tmpLineNum.intValue() >= bcLineNum ) {
		      newLineNum = tmpLineNum.intValue();
		      idx = i;
		      break;
		    }
		  }
		}
	      }
	      if( idx == null ) {
		throw new ParseException(
				createMsg(
					"bc.line_not_found",
					bcLineNum ),
				ex.getIndex() );
	      }
	      if( (newLineNum > 0) && (newLineNum != bcLineNum) ) {
		putSpecViolatedWarning(
			createMsg(
				"bc.goto_line_instaed_of",
				newLineNum,
				bcLineNum ),
			ex.getIndex() );
	      }
	      this.curLineIdx = idx.intValue();
	    }
	    catch( ParseException ex ) {
	      // catch exception here only when checking the syntax
	      if( this.bcRuntime != null ) {
		throw ex;
	      }
	      putError( ex );
	      this.curLineIdx++;
	    }
	    finally {
	      charIdx = 0;
	    }
	  }
	  catch( JumpException ex ) {
	    this.curLineIdx = ex.getLineIndex();
	    charIdx         = ex.getCharIndex();
	  }
	}
      }
    }
    catch( EndException ex ) {}
    catch( ParseException ex ) {
      this.aborted = true;
      putError( ex );
    }
  }


  public void putError( String msg, int index )
  {
    this.errState = true;
    putLineNumMsgToLog( "bc.error_in_line", msg, index );
  }


  public void putSpecViolatedWarning( String msg, int srcIndex )
  {
    if( this.warnSpecViolatedEnabled )
      putWarning( msg, srcIndex );
  }


  public void putWarning( String msg, int index )
  {
    if( this.warnEnabled ) {
      this.warnState = true;
      putLineNumMsgToLog( "bc.warning_in_line", msg, index );
    }
  }


  public String readIdentifier( CharacterIterator iter )
  {
    String rv = null;
    char   ch = toSyntaxCase( skipBlanks( iter ) );
    if( isLetter( ch ) ) {
      StringBuilder buf = new StringBuilder();
      buf.append( toSyntaxCase( ch ) );
      ch = iter.next();
      while( isLetter( ch ) || isDigit( ch )) {
	buf.append( toSyntaxCase( ch ) );
	ch = iter.next();
      }
      if( ch == '$' ) {
	buf.append( ch );
	iter.next();
      }
      rv = buf.toString();
    }
    return rv;
  }


  public String readStringLiteral( CharacterIterator iter )
						throws ParseException
  {
    String text = null;
    char   ch    = ProgramParser.skipBlanks( iter );
    int    index = iter.getIndex();
    if( ch == '\"' ) {
      StringBuilder buf = new StringBuilder();
      ch                = iter.next();
      while( (ch != CharacterIterator.DONE) && (ch !='\"') ) {
	if( (ch < '\u0020') || (ch > '\u007E') ) {
	  if( this.warnNonAsciiEnabled ) {
	    putWarning(
		createMsg( "bc.non_ascii_char", ch ),
		iter.getIndex() );
	  }
	}
	buf.append( ch );
	ch = iter.next();
      }
      if( ch != '\"' ) {
	throw new ParseException(
			Main.getText( "bc.unclosed_string" ),
			index );
      }
      iter.next();
      text = buf.toString();
    }
    return text;
  }


  public void setWarningsEnabled( boolean state )
  {
    this.warnEnabled = state;
  }


  public static char skipBlanks( CharacterIterator iter )
  {
    char ch = iter.current();
    while( (ch != CharacterIterator.DONE)
	   && ((ch == '\u0020') || Character.isWhitespace( ch )) )
    {
      ch = iter.next();
    }
    return ch;
  }


  private char toSyntaxCase( char ch )
  {
    return this.ignoreCase ? Character.toUpperCase( ch ) : ch;
  }


	/* --- Private Methods --- */

  private void checkAbort() throws InterruptedException
  {
    if( this.abortFired )
      throw new InterruptedException();
  }


  private void checkEndOfInstr( CharacterIterator iter )
						throws ParseException
  {
    char ch = skipBlanks( iter );
    if( (ch != ':') && (ch != CharacterIterator.DONE) ) {
      throw new ParseException(
			createMsg( "bc.unexpected_char", ch ),
			iter.getIndex() );
    }
  }


  private BCStackItem createStackItem( CharacterIterator iter )
  {
    int oldIdx = iter.getIndex();
    if( skipBlanks( iter ) == ':' ) {
      iter.next();
    }
    int charIdx = iter.getIndex();	// next instruction or line end
    iter.setIndex( oldIdx );
    return new BCStackItem(
		this.curLineIdx,
		this.srcLines.get( this.curLineIdx ).getBCLineNumber(),
		charIdx );
  }


  private void parseLine(
			SourceLine srcLine,
			int        charIdx ) throws
					EndException,
					GotoException,
					InterruptedException,
					JumpException,
					ParseException
  {
    if( srcLine != null ) {
      boolean mandatoryInstr = true;
      String  lineText       = srcLine.getLineText();
      int     firstInstrIdx  = srcLine.getFirstInstructionIndex();
      int     endPos         = lineText.length();
      if( charIdx > firstInstrIdx ) {
	mandatoryInstr = false;
      } else {
	charIdx = firstInstrIdx;
      }
      if( charIdx > endPos ) {
	charIdx = endPos;
      }
      CharacterIterator iter = new StringCharacterIterator(
						lineText,
						firstInstrIdx,
						endPos,
						charIdx );
      try {
	if( mandatoryInstr
	    || (skipBlanks( iter ) != CharacterIterator.DONE) )
	{
	  this.ignoreColon = false;
	  parseInstruction( srcLine, iter );
	  while( this.ignoreColon || checkAndParseToken( iter, ":" ) ) {
	    checkAbort();
	    this.ignoreColon = false;
	    parseInstruction( srcLine, iter );
	  }
	  char ch = skipBlanks( iter );
	  if( ch != CharacterIterator.DONE ) {
	    throw new ParseException(
			createMsg( "bc.unexpected_char", ch ),
			iter.getIndex() );
	  }
	}
      }
      catch( ArithmeticException ex ) {
	StringBuilder buf = new StringBuilder();
	buf.append( Main.getText( "bc.arithmetic_error" ) );
	String msg = ex.getMessage();
	if( msg != null ) {
	  if( !msg.isEmpty() ) {
	    buf.append( ": " );
	    buf.append( msg );
	  }
	}
	throw new ParseException( buf.toString(), iter.getIndex() );
      }
    }
  }


  private Object parseExpr( CharacterIterator iter ) throws ParseException
  {
    return (new ExprParser( this, this.bcRuntime )).parseExpr( iter );
  }


  private Number parseNumericExpr( CharacterIterator iter )
						throws ParseException
  {
    return (new ExprParser( this, this.bcRuntime )).parseNumericExpr( iter );
  }


  private String parseStringExpr( CharacterIterator iter )
						throws ParseException
  {
    return (new ExprParser( this, this.bcRuntime )).parseStringExpr( iter );
  }


  private void parseInstruction(
			SourceLine        srcLine,
			CharacterIterator iter ) throws
							EndException,
							GotoException,
							InterruptedException,
							JumpException,
							ParseException
  {
    BCRuntime bcRuntime = this.bcRuntime;
    if( bcRuntime != null ) {
      bcRuntime.checkDebugAction( this.curLineIdx, srcLine, iter.getIndex() );
      bcRuntime.checkSpeedBrake();
    }
    char ch = toSyntaxCase( skipBlanks( iter ) );
    if( (ch != CharacterIterator.DONE) && (ch != ':') ) {
      if( !isLetter( ch ) ) {
	throw new ParseException(
			Main.getText( "bc.instruction_expected" ),
			iter.getIndex() );
      }
      StringBuilder buf = new StringBuilder();
      boolean found = false;
      for(;;) {
	buf.append( ch );
	ch = toSyntaxCase( iter.next() );
	switch( buf.toString() ) {
	  case "DATA":
	    parseDATA( iter );
	    found = true;
	    break;
	  case "DEF":
	    parseDEF( iter );
	    found = true;
	    break;
	  case "DIM":
	    parseDIM( iter );
	    found = true;
	    break;
	  case "END":
	    parseEND( iter );
	    found = true;
	    break;
	  case "FOR":
	    parseFOR( iter );
	    found = true;
	    break;
	  case "GOSUB":
	    parseGosubGoto( iter, true );
	    found = true;
	    break;
	  case "GOTO":
	    parseGosubGoto( iter, false );
	    found = true;
	    break;
	  case "IF":
	    parseIF( iter );
	    found = true;
	    break;
	  case "INPUT":
	    parseINPUT( iter );
	    found = true;
	    break;
	  case "LET":
	    parseLET( iter );
	    found = true;
	    break;
	  case "NEXT":
	    parseNEXT( iter );
	    found = true;
	    break;
	  case "ON":
	    parseON( iter );
	    found = true;
	    break;
	  case "PRINT":
	    parsePRINT( iter );
	    found = true;
	    break;
	  case "READ":
	    parseREAD( iter );
	    found = true;
	    break;
	  case "REM":
	    parseREM( iter );
	    found = true;
	    break;
	  case "RESTORE":
	    parseRESTORE( iter );
	    found = true;
	    break;
	  case "RETURN":
	    parseRETURN( iter );
	    found = true;
	    break;
	  case "TAB":
	    parseTAB( iter );
	    found = true;
	    break;
	}
	if( found ) {
	  break;
	}
	if( !isLetter( ch ) ) {
	  break;
	}
      }
      if( !found ) {
	if( buf.length() == 0 ) {
	  throw new ParseException(
			Main.getText( "bc.instr_or_assigmnent_expected" ),
			iter.getIndex() );
	}
	// in case of a variable name with digits
	while( isLetter( ch ) || isDigit( ch ) ) {
	  buf.append( ch );
	  ch = toSyntaxCase( iter.next() );
	}
	// in case of a string variable
	if( iter.current() == '$' ) {
	  buf.append( '$' );
	  iter.next();
	}
	int          varNameIndex = iter.getIndex();
	List<Number> indices      = parseOptIndexList( iter );
	if( skipBlanks( iter ) != '=' ) {
	  throw new ParseException(
			createMsg(
				"bc.unknown_instruction",
				buf.toString() ),
			varNameIndex );
	}
	iter.next();
	parseAssignment(
		checkIdentifier( buf.toString(), iter.getIndex() ),
		indices,
		iter );
      }
    }
  }


  private Object parseAssignment(
			String            varName,
			List<Number>      indices,
			CharacterIterator iter ) throws ParseException
  {
    Object value = null;
    if( varName.endsWith( "$" ) ) {
      value = parseStringExpr( iter );
    } else {
      value = parseNumericExpr( iter );
    }
    if( this.bcRuntime != null ) {
      this.bcRuntime.setVarValue(
				varName,
				indices,
				value,
				iter.getIndex() );
    }
    return value;
  }


  private void parseDATA( CharacterIterator iter ) throws ParseException
  {
    /*
     * DATA instructions have already been processed during line scanning.
     * Here the instruction will be skipped.
     */
    char ch = skipBlanks( iter );
    while( (ch != CharacterIterator.DONE) && (ch != ':') ) {
      if( readStringLiteral( iter ) != null ) {
	ch = iter.current();
      } else {
	ch = iter.next();
      }
    }
  }


  private void parseDEF( CharacterIterator iter ) throws ParseException
  {
    parseToken( iter, "FN" );
    String funcName = readIdentifier( iter );
    if( funcName == null ) {
      throw new ParseException(
			Main.getText( "bc.funcname_expected" ),
			iter.getIndex() );
    }
    if( funcName.endsWith( "$" ) ) {
      throw new ParseException(
			Main.getText( "bc.illegal_type" ),
			iter.getIndex() );
    }
    funcName       = checkIdentifier( funcName, iter.getIndex() );
    String argName = null;
    if( checkAndParseToken( iter, "(" ) ) {
      argName = readIdentifier( iter );
      parseToken( iter, ")" );
    }
    parseToken( iter, "=" );
    skipBlanks( iter );
    int       begIdx = iter.getIndex();
    BCRuntime rt     = this.bcRuntime;
    try {
      this.bcRuntime = null;	// expression must not be calculated here
      parseNumericExpr( iter );
    }
    finally {
      this.bcRuntime = rt;
    }
    int endIdx  = iter.getIndex();
    int funcLen = endIdx - begIdx;
    if( funcLen < 1 ) {
      throw new ParseException(
			Main.getText( "bc.numeric_expr_excepted" ),
			iter.getIndex() );
    }
    StringBuilder buf = new StringBuilder( funcLen );
    iter.setIndex( begIdx );
    while( iter.getIndex() < endIdx ) {
      buf.append( toSyntaxCase( iter.current() ) );
      iter.next();
    }
    int index = iter.getIndex();
    checkEndOfInstr( iter );
    if( this.bcRuntime != null ) {
      this.bcRuntime.declareUserFunction(
					funcName,
					argName,
					buf.toString(),
					index );
    }
  }


  private void parseDIM( CharacterIterator iter ) throws ParseException
  {
    for(;;) {
      String varName = parseVarName( iter );
      if( skipBlanks( iter ) != '(' ) {
	throw new ParseException(
			createMsg( "bc.token_expected", '(' ),
			iter.getIndex() );
      }
      List<Number> dimensions = parseOptIndexList( iter );
      if( this.bcRuntime != null ) {
	this.bcRuntime.declareDimVar( varName, dimensions, iter.getIndex() );
      }
      if( skipBlanks( iter ) != ',' ) {
	break;
      }
      iter.next();
    }
  }


  private void parseEND( CharacterIterator iter ) throws
							EndException,
							ParseException
  {
    putSpecViolatedWarning(
		createMsg( "bc.should_be_replaced_by", "END", "GOTO 950" ),
		iter.getIndex() );
    checkEndOfInstr( iter );
    if( this.bcRuntime != null ) {
      this.bcRuntime.doEnd();
    }
  }


  private void parseFOR( CharacterIterator iter ) throws ParseException
  {
    String varName = parseVarName( iter );
    if( varName.endsWith( "$" ) ) {
      throw new ParseException(
			Main.getText( "bc.type_mismatch" ),
			iter.getIndex() );
    }
    List<Number> indices = parseOptIndexList( iter );
    parseToken( iter, "=" );
    Number begValue = (Number) parseAssignment( varName, indices, iter );
    parseToken( iter, "TO" );
    Number endValue = parseNumericExpr( iter );
    Number stepSize = null;
    if( checkAndParseToken( iter, "STEP" ) ) {
      stepSize = parseNumericExpr( iter );
    }
    checkEndOfInstr( iter );
    if( this.bcRuntime != null ) {
      if( stepSize == null ) {
	if( begValue.doubleValue() < endValue.doubleValue() ) {
	  stepSize = Integer.valueOf( 1 );
	} else {
	  stepSize = Integer.valueOf( -1 );
	}
      }
      this.bcRuntime.doFor(
		new ForStackItem(
			varName,
			indices,
			endValue,
			stepSize,
			createStackItem( iter ) ) );
    }
  }


  private void parseGosubGoto(
			CharacterIterator iter,
			boolean           gosub ) throws
							EndException,
							GotoException,
							InterruptedException,
							JumpException,
							ParseException
  {
    Integer bcLineNum = readInt( iter );
    if( bcLineNum == null ) {
      throw new ParseException(
			Main.getText( "bc.line_num_expected" ),
			iter.getIndex() );
    }
    if( bcLineNum.intValue() < 1000 ) {
      if( Arrays.binarySearch(
			sortedSysLineNums,
			bcLineNum.intValue() ) < 0 )
      {
	throw new ParseException(
			createMsg( "bc.unknown_system_line_num", bcLineNum ),
			iter.getIndex() );
      }
    }
    int index = iter.getIndex();
    checkEndOfInstr( iter );
    if( this.bcRuntime != null ) {
      this.bcRuntime.doGosubGoto(
	bcLineNum.intValue(),
	gosub ? new GosubStackItem(
			bcLineNum.intValue(),
			createStackItem( iter ) ) : null,
	iter.getIndex() );
    }
  }


  private void parseIF( CharacterIterator iter ) throws
						EndException,
						GotoException,
						InterruptedException,
						JumpException,
						ParseException
  {
    boolean status    = true;
    Number  condition = parseNumericExpr( iter );
    if( this.bcRuntime != null ) {
      if( condition.doubleValue() == 0.0 ) {
	status = false;
      }
    }
    if( !checkAndParseToken( iter, "THEN" ) ) {
      putSpecViolatedWarning(
		createMsg( "bc.token_expected", "THEN" ), iter.getIndex() );
    }
    if( status ) {
      if( isDigit( skipBlanks( iter ) ) ) {
	parseGosubGoto( iter, false );
      }
      this.ignoreColon = true;
    } else {
      iter.setIndex( iter.getEndIndex() );
    }
  }


  private void parseINPUT( CharacterIterator iter )
		throws EndException, InterruptedException, ParseException
  {
    String text = readStringLiteral( iter );
    if( text != null ) {
      parseToken( iter, ";" );
    }
    String varName = parseVarName( iter );
    List<Number> indices = parseOptIndexList( iter );
    checkEndOfInstr( iter );
    if( this.bcRuntime != null ) {
      this.bcRuntime.doInput( text, varName, indices, iter.getIndex() );
    }
  }


  private void parseLET( CharacterIterator iter ) throws ParseException
  {
    String varName = parseVarName( iter );
    List<Number> indices = parseOptIndexList( iter );
    parseToken( iter, "=" );
    parseAssignment( varName, indices, iter );
    checkEndOfInstr( iter );
  }


  private void parseNEXT( CharacterIterator iter )
				throws JumpException, ParseException
  {
    String varName = null;
    if( isLetter( skipBlanks( iter ) ) ) {
      varName = parseVarName( iter );
      if( varName.endsWith( "$" ) ) {
	throw new ParseException(
			Main.getText( "bc.type_mismatch" ),
			iter.getIndex() );
      }
    } else {
      putSpecViolatedWarning(
		Main.getText( "bc.varname_expected" ),
		iter.getIndex() );
    }
    List<Number> indices = parseOptIndexList( iter );
    if( this.bcRuntime != null ) {
      this.bcRuntime.doNext( varName, indices, iter.getIndex() );
    }
  }


  private void parseON( CharacterIterator iter ) throws
						EndException,
						GotoException,
						InterruptedException,
						JumpException,
						ParseException
  {
    Number  pos   = parseNumericExpr( iter );
    boolean gosub = false;
    if( checkAndParseToken( iter, "GOSUB" ) ) {
      gosub = true;
    } else if( !checkAndParseToken( iter, "GOTO" ) ) {
      throw new ParseException(
			Main.getText( "bc.gosub_or_goto_expected" ),
			iter.getIndex() );
    }
    List<Integer> bcLineNums = new ArrayList<>();
    List<Integer> srcIndexes = new ArrayList<>();
    do {
      Integer bcLineNum = readInt( iter );
      if( bcLineNum == null ) {
	throw new ParseException(
			Main.getText( "bc.line_num_expected" ),
			iter.getIndex() );
      }
      srcIndexes.add( iter.getIndex() );
      bcLineNums.add( bcLineNum );
    } while( checkAndParseToken( iter, "," ) );
    checkEndOfInstr( iter );
    if( this.bcRuntime != null ) {
      int idx = pos.intValue() - 1;
      if( (idx >= 0) && (idx < bcLineNums.size()) ) {
	int srcIndex = iter.getIndex();
	if( (idx >= 0) && (idx < srcIndexes.size()) ) {
	  srcIndex = srcIndexes.get( idx );
	}
	int bcLineNum = bcLineNums.get( idx );
	this.bcRuntime.doGosubGoto(
		bcLineNum,
		gosub ? new GosubStackItem(
					bcLineNum,
					createStackItem( iter ) ) : null,
		srcIndex );
      }
    }
  }


  private void parsePRINT( CharacterIterator iter ) throws ParseException
  {
    List<Object> items  = new ArrayList<>();
    boolean      newFld = true;
    char         ch     = skipBlanks( iter );
    while( (ch != CharacterIterator.DONE) && (ch != ':') ) {
      if( !newFld ) {
	throw new ParseException(
			createMsg( "bc.unexpected_char", ch ),
			iter.getIndex() );
      }
      newFld      = false;
      Object item = null;
      if( checkAndParseToken( iter, "TAB" ) ) {
	item = BCRuntime.createPrintTabItem( parseNumericExpr( iter ) );
      } else {
	item = parseExpr( iter );
      }
      if( item == null ) {
	item = "";
      }
      items.add( item );
      ch = skipBlanks( iter );
      if( ch == ',' ) {
	items.add( BCRuntime.PRINT_TAB_ITEM);
	iter.next();
	newFld = true;
	ch     = skipBlanks( iter );
      } else if( ch == ';' ) {
	items.add( BCRuntime.PRINT_CONCAT_ITEM);
	iter.next();
	newFld = true;
	ch     = skipBlanks( iter );
      }
    }
    checkEndOfInstr( iter );
    if( this.bcRuntime != null ) {
      this.bcRuntime.print( items );
    }
  }


  private void parseREAD( CharacterIterator iter ) throws ParseException
  {
    do {
      String       varName = parseVarName( iter );
      List<Number> indices = parseOptIndexList( iter );
      if( this.dataReadIdx >= this.dataValues.size() ) {
	throw new ParseException(
			Main.getText( "bc.out_of_data" ),
			iter.getIndex() );
      }
      if( this.bcRuntime != null ) {
	this.bcRuntime.setVarValue(
			varName,
			indices,
			this.dataValues.get( this.dataReadIdx++ ),
			iter.getIndex() );
      }
    } while( checkAndParseToken( iter, "," ) );
    checkEndOfInstr( iter );
  }


  private void parseREM( CharacterIterator iter )
  {
    iter.setIndex( iter.getEndIndex() );
  }


  private void parseRESTORE( CharacterIterator iter ) throws ParseException
  {
    Integer bcLineNum = readInt( iter );
    if( bcLineNum != null ) {
      Integer idx = this.bcLineNum2DataIdx.get( bcLineNum );
      if( idx == null ) {
	throw new ParseException(
			Main.getText( "bc.line_without_data" ),
			iter.getIndex() );
      }
      this.dataReadIdx = idx.intValue();
      putSpecViolatedWarning( Main.getText( "bc.restore_with_line_num" ) );
    } else {
      this.dataReadIdx = 0;
    }
    checkEndOfInstr( iter );
  }


  private void parseRETURN( CharacterIterator iter )
				throws JumpException, ParseException
  {
    int index = iter.getIndex();
    checkEndOfInstr( iter );
    if( this.bcRuntime != null ) {
      this.bcRuntime.doReturn( index );
    }
  }


  private void parseTAB( CharacterIterator iter ) throws ParseException
  {
    Object item = BCRuntime.createPrintTabItem( parseNumericExpr( iter ) );
    checkEndOfInstr( iter );
    if( this.bcRuntime != null ) {
      this.bcRuntime.print( Collections.singletonList( item ) );
    }
  }


  private List<Number> parseOptIndexList( CharacterIterator iter )
						throws ParseException
  {
    List<Number> indices = null;
    if( skipBlanks( iter ) == '(' ) {
      iter.next();
      indices = new ArrayList<>();
      for(;;) {
	if( indices.size() == 2 ) {
	  throw new ParseException(
			Main.getText( "bc.too_many_indices" ),
			iter.getIndex() );
	}
	indices.add( parseNumericExpr( iter ) );
        if( skipBlanks( iter ) != ',' ) {
	  break;
	}
	iter.next();
      }
      parseToken( iter, ")" );
    }
    return indices;
  }


  private void parseToken(
			CharacterIterator iter,
			String            token ) throws ParseException
  {
    char ch  = toSyntaxCase( skipBlanks( iter ) );
    int  idx = iter.getIndex();
    int  len = token.length();
    for( int i = 0; i < len; i++ ) {
      if( ch != token.charAt( i ) ) {
	iter.setIndex( idx );
	throw new ParseException(
			createMsg( "bc.token_expected", token ),
			idx );
      }
      ch = toSyntaxCase( iter.next() );
    }
  }


  private String parseVarName( CharacterIterator iter )
						throws ParseException
  {
    String varName = readIdentifier( iter );
    if( varName == null ) {
      throw new ParseException(
			Main.getText( "bc.varname_expected" ),
			iter.getIndex() );
    }
    return checkIdentifier( varName, iter.getIndex() );
  }


  private void putError( ParseException ex )
  {
    this.errState = true;
    putLineNumMsgToLog(
		"bc.error_in_line",
		ex.getMessage(),
		ex.getErrorOffset() );
  }


  private void putLineNumMsgToLog(
			String lineNumTextKey,
			String msg,
			int    index )
  {
    SourceLine srcLine = null;
    if( this.curLineIdx < this.srcLines.size() ) {
      srcLine = this.srcLines.get( this.curLineIdx );
    }
    StringBuilder buf = new StringBuilder( 256 );
    buf.append( createMsg( lineNumTextKey, this.curLineIdx + 1 ) );
    buf.append( ": " );
    buf.append( msg );
    buf.append( '\n' );
    String msgKey = buf.toString();
    if( this.shownMessages.add( msgKey ) ) {
      if( srcLine != null ) {
	String lineText = srcLine.getLineText();
	if( lineText != null ) {
	  if( (index >= 0)
	      && (index <= srcLine.getLineText().length()) )
	  {
	    buf.append( lineText );
	    buf.append( '\n' );
	    for( int i = 0; i < index; i++ ) {
	      buf.append( '\u0020' );
	    }
	    buf.append( "^\n" );
	  }
	}
      }
      this.mainFrm.fireAppendToLog( buf.toString() );
    }
  }


  private void putSpecViolatedWarning( String msg )
  {
    putSpecViolatedWarning( msg, -1 );
  }


  private Integer readInt( CharacterIterator iter ) throws ParseException
  {
    Integer rv = null;
    char    ch = skipBlanks( iter );
    if( isDigit( ch ) ) {
      long v = ch - '0';
      ch     = iter.next();
      while( isDigit( ch ) ) {
	v = (v * 10) + (ch - '0');
	if( v > (long) Integer.MAX_VALUE ) {
	  throw new ParseException(
			Main.getText( "bc.numeric_overflow" ),
			iter.getIndex() );
	}
	ch = iter.next();
      }
      rv = Integer.valueOf( (int) v );
    }
    return rv;
  }


  private void scanLine( int lineNum, String line )
				throws InterruptedException, ParseException
  {
    if( line != null ) {
      CharacterIterator iter          = new StringCharacterIterator( line );
      boolean           startsWithNum = isDigit( iter.current() );
      Integer bcLineNum = readInt( iter );
      if( bcLineNum != null ) {
	if( !startsWithNum ) {
	  putSpecViolatedWarning(
			Main.getText( "bc.blanks_before_line_num" ) );
	}
	if( bcLineNum.intValue() < 1000 ) {
	  throw new ParseException(
			createMsg(
				"bc.illegal_line_num",
				bcLineNum.intValue() ),
			iter.getIndex() );
	}
	if( bcLineNum.intValue() > 0x7FFF ) {
	  putSpecViolatedWarning(
		createMsg(
			"bc.line_num_gt_32767",
			bcLineNum.intValue() ) );
	}
      } else {
	putSpecViolatedWarning( Main.getText( "bc.line_num_expected" ) );
      }
      skipBlanks( iter );
      SourceLine srcLine = new SourceLine(
					lineNum,
					line,
					bcLineNum,
					iter.getIndex() );
      if( bcLineNum != null ) {
	this.bcLineNum2LineIdx.put( bcLineNum, this.srcLines.size() );
      }
      this.srcLines.add( srcLine );

      // process DATA instructions
      char ch = skipBlanks( iter );
      while( ch != CharacterIterator.DONE ) {
	checkAbort();
	if( checkAndParseToken( iter, "DATA" ) ) {
	  if( bcLineNum != null ) {
	    this.bcLineNum2DataIdx.put( bcLineNum, this.dataValues.size() );
	    bcLineNum = null;	// suppress future putting to map
	  }
	  do {
	    Object value = ExprParser.readNumber(
						iter,
						this.ignoreCase,
						true );
	    if( value == null ) {
	      value = readStringLiteral( iter );
	    }
	    if( value != null ) {
	      this.dataValues.add( value );
	    } else {
	      putSpecViolatedWarning(
			Main.getText( "bc.literal_expected" ),
			iter.getIndex() );
	    }
	  } while( checkAndParseToken( iter, "," ) );
	}
	ch = skipBlanks( iter );
	while( (ch != CharacterIterator.DONE) && (ch != ':') ) {
	  if( ch == '\"' ) {
	    readStringLiteral( iter );
	    ch = iter.current();
	  } else {
	    ch = iter.next();
	  }
	}
	if( ch == ':' ) {
	  ch = iter.next();
	}
      }
      if( iter.getIndex() >= 60 ) {
	putSpecViolatedWarning(
		Main.getText( "bc.line_longer_than_60" ),
		iter.getIndex() );
      }
    }
  }


  private static String trimTrailingBlankLines( String text )
  {
    if( text != null ) {
      int len = text.length();
      int idx = len - 1;
      while( idx >= 0 ) {
	if( text.charAt( idx ) != '\n' ) {
	  break;
	}
	--idx;
      }
      idx++;
      if( (idx - 1) < len ) {
	text = text.substring( 0, idx );
      }
    } else {
      text = "";
    }
    return text;
  }
}
