/*
 * (c) 2025 Jens Mueller
 *
 * JBASICODE
 *
 * Debugger Component
 */

package jbasicode.ui;

import java.awt.BorderLayout;
import java.awt.EventQueue;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.text.CharacterIterator;
import java.text.MessageFormat;
import java.text.StringCharacterIterator;
import java.util.Set;
import java.util.TreeSet;
import javax.swing.JButton;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTable;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.text.BadLocationException;
import jbasicode.Main;
import jbasicode.bc.BCRuntime;
import jbasicode.bc.ProgramParser;
import jbasicode.bc.SourceLine;


public class DebugFld extends JPanel implements
					ActionListener,
					MouseListener
{
  public enum DebugAction { STOP, RUN, STEP_OVER, STEP_INTO, STEP_UP };

  private volatile BCRuntime bcRuntime;
  private MainFrm             mainFrm;
  private DebugAction         debugAction;
  private Set<Integer>        bpLineIndexes;
  private SourceLine          srcLine;
  private boolean             notified;
  private int                 curInstrBegIdx;
  private int                 curInstrEndIdx;
  private int                 curLineIdx;
  private int                 curLineStartOffs;
  private int                 triggerLineIdx;
  private int                 instrBegIdx;
  private int                 stackDepth;
  private String              textPrgRunning;
  private String              textPrgStopped;
  private Object              waitMonitor;
  private JPopupMenu          popupMenu;
  private JMenuItem           popupBpSet;
  private JMenuItem           popupBpRemove;
  private JMenuItem           popupBpRemoveAll;
  private JSplitPane          splitPane;
  private JButton             btnMarkCurInstr;
  private JTextField          fldCurInstr;
  private JTextField          fldNextReadValues;
  private JTextField          fldStackItems;
  private JTextArea           prgArea;
  private JScrollPane         prgSP;
  private DebugPrgRowHeader   prgRowHeader;
  private JTable              varTable;
  private VarTableModel       varTableModel;


  public DebugFld( MainFrm mainFrm )
  {
    this.mainFrm          = mainFrm;
    this.bcRuntime        = null;
    this.srcLine          = null;
    this.curInstrBegIdx   = -1;
    this.curInstrEndIdx   = -1;
    this.curLineIdx       = -1;
    this.curLineStartOffs = -1;
    this.triggerLineIdx   = -1;
    this.instrBegIdx      = 0;
    this.stackDepth       = 0;
    this.notified         = false;
    this.bpLineIndexes    = new TreeSet<>();
    this.debugAction      = DebugAction.RUN;
    this.textPrgRunning   = Main.getText( "status.bc.prg_running" );
    this.textPrgStopped   = Main.getText( "status.bc.prg_stopped" );
    this.waitMonitor      = new Object();
    setLayout( new BorderLayout() );

    this.prgArea = new JTextArea();
    this.prgArea.setEditable( false );
    this.prgArea.addCaretListener( mainFrm );

    Font font = this.prgArea.getFont();
    this.prgArea.setFont(
		new Font(
			Font.MONOSPACED,
			Font.PLAIN,
			font != null ? font.getSize() : 13 ) );

    this.prgSP = new JScrollPane( this.prgArea );

    this.prgRowHeader = new DebugPrgRowHeader(
					this.prgArea,
					this.bpLineIndexes );
    this.prgSP.setRowHeaderView( this.prgRowHeader );

    JPanel panelInstrInfo = new JPanel( new GridBagLayout() );

    GridBagConstraints gbcInstrInfo = new GridBagConstraints(
						0, 0,
						1, 1,
						0.0, 0.0,
						GridBagConstraints.EAST,
						GridBagConstraints.NONE,
						new Insets( 5, 5, 0, 0 ),
						0, 0 );

    panelInstrInfo.add(
		new JLabel( Main.getText( "label.cur_instr" ) ),
		gbcInstrInfo );
    gbcInstrInfo.insets.top = 0;
    gbcInstrInfo.gridy++;
    panelInstrInfo.add(
		new JLabel( Main.getText( "label.instr_stack" ) ),
		gbcInstrInfo );
    gbcInstrInfo.insets.bottom = 5;
    gbcInstrInfo.gridy++;
    panelInstrInfo.add(
		new JLabel( Main.getText( "label.instr_read_values" ) ),
		gbcInstrInfo );

    this.fldCurInstr = new JTextField();
    this.fldCurInstr.setEditable( false );
    gbcInstrInfo.anchor        = GridBagConstraints.WEST;
    gbcInstrInfo.fill          = GridBagConstraints.HORIZONTAL;
    gbcInstrInfo.insets.top    = 5;
    gbcInstrInfo.insets.bottom = 0;
    gbcInstrInfo.weightx       = 1.0;
    gbcInstrInfo.gridy         = 0;
    gbcInstrInfo.gridx++;
    panelInstrInfo.add( this.fldCurInstr, gbcInstrInfo );

    this.btnMarkCurInstr = new JButton( Main.getText( "action.mark" ) );
    this.btnMarkCurInstr.setEnabled( false );
    gbcInstrInfo.fill    = GridBagConstraints.NONE;
    gbcInstrInfo.weightx = 0.0;
    gbcInstrInfo.gridx++;
    panelInstrInfo.add( this.btnMarkCurInstr, gbcInstrInfo );

    this.fldStackItems = new JTextField();
    this.fldStackItems.setEditable( false );
    gbcInstrInfo.fill      = GridBagConstraints.HORIZONTAL;
    gbcInstrInfo.weightx   = 1.0;
    gbcInstrInfo.gridwidth = GridBagConstraints.REMAINDER;
    --gbcInstrInfo.gridx;
    gbcInstrInfo.gridy++;
    panelInstrInfo.add( this.fldStackItems, gbcInstrInfo );

    this.fldNextReadValues = new JTextField();
    this.fldNextReadValues.setEditable( false );
    gbcInstrInfo.insets.bottom = 5;
    gbcInstrInfo.gridy++;
    panelInstrInfo.add( this.fldNextReadValues, gbcInstrInfo );

    JPanel panelNorth = new JPanel( new BorderLayout( 0, 0 ) );
    panelNorth.add( this.prgSP, BorderLayout.CENTER );
    panelNorth.add( panelInstrInfo, BorderLayout.SOUTH );

    this.varTableModel = new VarTableModel();
    this.varTable      = new JTable( this.varTableModel );
    this.varTable.setAutoResizeMode( JTable.AUTO_RESIZE_LAST_COLUMN );
    this.varTable.setDragEnabled( false );
    this.varTable.setFillsViewportHeight( true );
    this.varTable.setColumnSelectionAllowed( false );
    this.varTable.setRowSelectionAllowed( true );
    this.varTable.setSelectionMode( ListSelectionModel.SINGLE_SELECTION );
    this.varTable.setShowHorizontalLines( false );
    this.varTable.setShowVerticalLines( false );
    this.varTable.getColumnModel().getColumn( 0 ).setPreferredWidth( 100 );

    this.splitPane = new JSplitPane(
				JSplitPane.VERTICAL_SPLIT,
				panelNorth,
				new JScrollPane( this.varTable ) );
    add( this.splitPane, BorderLayout.CENTER );

    this.popupMenu = new JPopupMenu();

    this.popupBpSet = new JMenuItem( Main.getText( "action.debug.bp.set" ) );
    this.popupMenu.add( this.popupBpSet );

    this.popupBpRemove = new JMenuItem(
				Main.getText( "action.debug.bp.remove" ) );
    this.popupMenu.add( this.popupBpRemove );
    this.popupMenu.addSeparator();

    this.popupBpRemoveAll = new JMenuItem(
			Main.getText( "action.debug.bp.remove_all" ) );
    this.popupMenu.add( this.popupBpRemoveAll );
  }


  public void checkDebugAction(
			int        curLineIdx,
			SourceLine srcLine,
			int        instrBegIdx,
			int        stackDepth ) throws InterruptedException
  {
    this.curLineIdx  = curLineIdx;
    this.srcLine     = srcLine;
    this.instrBegIdx = instrBegIdx;
    if( (this.debugAction == DebugAction.STOP)
	|| (this.debugAction == DebugAction.STEP_INTO)
	|| ((this.debugAction == DebugAction.STEP_OVER)
		&& (stackDepth <= this.stackDepth))
	|| ((this.debugAction == DebugAction.STEP_UP)
		&& (stackDepth < this.stackDepth))
	|| (this.bpLineIndexes.contains( curLineIdx )
		&& (instrBegIdx == srcLine.getFirstInstructionIndex())) )
    {
      this.mainFrm.fireShowStatusText( this.textPrgStopped );
      updFields( true );
      synchronized( this.waitMonitor ) {
	this.waitMonitor.wait();
      }
      this.stackDepth = stackDepth;
      updFields( false );
      this.mainFrm.fireShowStatusText( this.textPrgRunning );
    }
  }


  public void doAddBreakpoint()
  {
    BCRuntime bcRuntime = this.bcRuntime;
    if( bcRuntime != null ) {
      Integer lineIdx = null;
      while( lineIdx == null ) {
	String text = JOptionPane.showInputDialog(
			this.mainFrm,
			Main.getText( "label.bc.line_num" ),
			Main.getText( "title.bp.add" ),
			JOptionPane.QUESTION_MESSAGE );
	if( text == null ) {
	  break;
	}
	try {
	  int lineNum = Integer.parseInt( text.trim() );
	  lineIdx     = bcRuntime.getProgramParser()
				.getLineIndexByBCLineNumber( lineNum );
	  if( lineIdx != null ) {
	    break;
	  }
	  UIUtil.showErrorMsg(
			this.mainFrm,
			MessageFormat.format(
				Main.getText( "bc.line_not_found" ),
				String.valueOf( lineNum ) ) );
	}
	catch( NumberFormatException ex ) {
	  UIUtil.showErrorMsg(
			this.mainFrm,
			Main.getText( "bc.invalid_number" ) );
	}
      }
      if( lineIdx != null ) {
	this.bpLineIndexes.add( lineIdx );
	this.prgRowHeader.repaint();
	this.mainFrm.fireBreakpointsChanged();
      }
    }
  }


  public void fireSetCurDebugDataVisible( boolean state, boolean markInstr )
  {
    int        curLineIdx  = this.curLineIdx;
    int        instrBegIdx = this.instrBegIdx;
    SourceLine srcLine     = this.srcLine;
    if( (curLineIdx >= 0) && (instrBegIdx >= 0) && (srcLine != null) ) {
      EventQueue.invokeLater(
		()->setCurDebugDataVisible(
				state,
				curLineIdx,
				srcLine.getLineText(),
				instrBegIdx,
				markInstr ) );
    }
  }


  public JTextArea getTextArea()
  {
    return this.prgArea;
  }


  public boolean hasBreakpoints()
  {
    return !this.bpLineIndexes.isEmpty();
  }


  public void removeAllBreakpoints()
  {
    this.bpLineIndexes.clear();
    repaint();
    this.mainFrm.fireBreakpointsChanged();
  }


  public void setData( String prgText, BCRuntime bcRuntime )
  {
    this.prgArea.setText( prgText != null ? prgText : "" );
    this.btnMarkCurInstr.setEnabled( false );
    this.fldCurInstr.setText( "" );
    this.fldStackItems.setText( "" );
    this.fldNextReadValues.setText( "" );
    this.curInstrBegIdx   = -1;
    this.curInstrEndIdx   = -1;
    this.curLineIdx       = -1;
    this.curLineStartOffs = -1;
    this.prgRowHeader.setCurLineIndex( -1 );

    // clear old runtime object
    if( this.bcRuntime != null ) {
      this.bcRuntime.setDebugFld( null );
      this.varTableModel.setRows( null );
    }

    // set new runtime object
    this.bcRuntime = bcRuntime;
    if( bcRuntime != null ) {
      boolean prgStopped = false;
      if( bcRuntime.getState() == Thread.State.NEW ) {
	this.debugAction   = DebugAction.STOP;
	prgStopped = false;
	bcRuntime.start();
      } else {
	this.debugAction = DebugAction.RUN;
      }
      bcRuntime.setDebugFld( this );
      this.mainFrm.fireDebugStatusChanged( prgStopped );
    }
  }


  public void setDebugAction( DebugAction debugAction )
  {
    synchronized( this.waitMonitor ) {
      this.debugAction = debugAction;
      if( debugAction == DebugAction.STOP ) {
	BCRuntime bcRuntime = this.bcRuntime;
	if( bcRuntime != null ) {
	  bcRuntime.debugStopFired();
	}
      } else {
	try {
	  this.waitMonitor.notifyAll();
	}
	catch( IllegalMonitorStateException ex ) {}
      }
    }
  }


  @Override
  public void addNotify()
  {
    super.addNotify();
    if( !this.notified ) {
      this.notified = true;
      this.btnMarkCurInstr.addActionListener( this );
      this.popupBpSet.addActionListener( this );
      this.popupBpRemove.addActionListener( this );
      this.popupBpRemoveAll.addActionListener( this );
      this.prgArea.addMouseListener( this );
    }
  }


  @Override
  public void removeNotify()
  {
    if( this.notified ) {
      this.notified = false;
      this.btnMarkCurInstr.removeActionListener( this );
      this.popupBpSet.removeActionListener( this );
      this.popupBpRemove.removeActionListener( this );
      this.popupBpRemoveAll.removeActionListener( this );
      this.prgArea.removeMouseListener( this );
    }
    super.removeNotify();
  }


  @Override
  public void setVisible( boolean state )
  {
    super.setVisible( state );
    if( state ) {
      this.splitPane.setDividerLocation( 0.67 );
    }
  }


	/* --- ActionListener --- */

  @Override
  public void actionPerformed( ActionEvent e )
  {
    Object src = e.getSource();
    if( src == this.btnMarkCurInstr ) {
      doMarkCurInstr();
    } else if( src == this.popupBpSet ) {
      doSetBreakpoint();
    } else if( src == this.popupBpRemove ) {
      doRemoveBreakpoint();
    } else if( src == this.popupBpRemoveAll ) {
      doRemoveAllBreakpoints();
    }
  }


	/* --- MouseListener --- */

  @Override
  public void mouseClicked( MouseEvent e )
  {
    if( e.isPopupTrigger() )
      showPopupMenu( e );
  }


  @Override
  public void mouseEntered( MouseEvent e )
  {
    // empty
  }


  @Override
  public void mouseExited( MouseEvent e )
  {
    // empty
  }


  @Override
  public void mousePressed( MouseEvent e )
  {
    if( e.isPopupTrigger() )
      showPopupMenu( e );
  }


  @Override
  public void mouseReleased( MouseEvent e )
  {
    if( e.isPopupTrigger() )
      showPopupMenu( e );
  }


	/* --- Private methods --- */

  private void doMarkCurInstr()
  {
    if( (this.curLineStartOffs >= 0)
	&& (this.curInstrBegIdx >= 0)
	&& (this.curInstrEndIdx > this.curInstrBegIdx) )
    {
      EventQueue.invokeLater(
		()->selectPrgText(
			this.curLineStartOffs + this.curInstrBegIdx,
			this.curLineStartOffs + this.curInstrEndIdx ) );
    }
  }


  private void doRemoveBreakpoint()
  {
    if( this.triggerLineIdx >= 0 ) {
      this.bpLineIndexes.remove( this.triggerLineIdx );
      this.prgRowHeader.repaint();
    }
  }


  private void doRemoveAllBreakpoints()
  {
    this.bpLineIndexes.clear();
    this.prgRowHeader.repaint();
  }


  private void doSetBreakpoint()
  {
    if( this.triggerLineIdx >= 0 ) {
      this.bpLineIndexes.add( this.triggerLineIdx );
      this.prgRowHeader.repaint();
    }
  }


  private void selectPrgText( int startIdx, int endIdx )
  {
    try {
      this.prgArea.setCaretPosition( startIdx );
      this.prgArea.moveCaretPosition( endIdx );
      this.prgArea.requestFocus();
    }
    catch( IllegalArgumentException ex ) {}
  }


  private void setCurDebugDataVisible(
				boolean state,
				int     lineIdx,
				String  lineText,
				int     instrBegIdx,
				boolean markInstr )
  {
    this.btnMarkCurInstr.setEnabled( false );
    this.fldCurInstr.setText( "" );
    this.fldStackItems.setText( "" );
    this.fldNextReadValues.setText( "" );
    if( state && (lineIdx >= 0) ) {
      this.prgRowHeader.setCurLineIndex( lineIdx );
      try {
	// find end of instruction and extract and select instruction
	StringBuilder     buf  = new StringBuilder();
	CharacterIterator iter = new StringCharacterIterator(
							lineText,
							instrBegIdx );
	char ch = ProgramParser.skipBlanks( iter );
	while( (ch != CharacterIterator.DONE)
	       && (ch != ':') && (ch != '\n') )
	{
	  buf.append( ch );
	  if( ch == '\"' ) {
	    ch = iter.next();
	    while( (ch != CharacterIterator.DONE)
		   && (ch != '\"') && (ch != '\n') )
	    {
	      buf.append( ch );
	      ch = iter.next();
	    }
	    if( ch == '\"' ) {
	      ch = iter.next();
	      buf.append( ch );
	    }
	  } else {
	    ch = iter.next();
	  }
	}
	this.fldCurInstr.setText( buf.toString() );
	try {
	  this.fldCurInstr.setCaretPosition( 0 );
	}
	catch( IllegalArgumentException ex ) {}
	BCRuntime bcRuntime = this.bcRuntime;
	if( bcRuntime != null ) {
	  String text = bcRuntime.createStackItemsView();
	  this.fldStackItems.setText( text );
	  this.fldStackItems.setToolTipText( text.isEmpty() ? null : text );
	  try {
	    this.fldStackItems.setCaretPosition( 0 );
	  }
	  catch( IllegalArgumentException ex ) {}
	  text = bcRuntime.getProgramParser().createNextDataReadValuesView();
	  this.fldNextReadValues.setText( text );
	  this.fldNextReadValues.setToolTipText(
				text.isEmpty() ? null : text );
	  try {
	    this.fldNextReadValues.setCaretPosition( 0 );
	  }
	  catch( IllegalArgumentException ex ) {}
	}
	this.curInstrBegIdx   = instrBegIdx;
	this.curInstrEndIdx   = iter.getIndex();
	this.curLineStartOffs = this.prgArea.getLineStartOffset( lineIdx );
	if( markInstr ) {
	  this.btnMarkCurInstr.setEnabled( true );
	  doMarkCurInstr();
	}
      }
      catch( BadLocationException ex ) {}
      catch( IllegalArgumentException ex ) {}
      this.varTableModel.setRows( bcRuntime.createVarViewList() );
    } else {
      try {
	this.prgArea.moveCaretPosition( this.prgArea.getCaretPosition() );
      }
      catch( IllegalArgumentException ex ) {}
      this.varTableModel.setRows( null );
    }
  }


  private void showPopupMenu( MouseEvent e )
  {
    this.triggerLineIdx = -1;
    try {
      int modelIdx = UIUtil.viewToModel( this.prgArea, e.getPoint() );
      if( modelIdx >= 0 ) {
	this.triggerLineIdx = this.prgArea.getLineOfOffset( modelIdx );
      }
    }
    catch( BadLocationException ex ) {}

    boolean isBP = this.bpLineIndexes.contains( this.triggerLineIdx );
    this.popupBpSet.setEnabled( !isBP );
    this.popupBpRemove.setEnabled( isBP );
    this.popupBpRemoveAll.setEnabled( !this.bpLineIndexes.isEmpty() );

    e.consume();
    this.popupMenu.show( e.getComponent(), e.getX(), e.getY() );
    e.consume();
  }


  private void updFields( boolean pause )
  {
    fireSetCurDebugDataVisible( pause, true );
    this.mainFrm.fireDebugStatusChanged( pause );
  }
}
