/*
 * (c) 2025 Jens Mueller
 *
 * JBASICODE
 *
 * RowHeader Component to show breakpoints and mark current basic line
 */

package jbasicode.ui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.geom.Rectangle2D;
import java.net.URL;
import java.util.Collection;
import javax.swing.JComponent;
import javax.swing.JTextArea;
import javax.swing.text.BadLocationException;


public class DebugPrgRowHeader extends JComponent
{
  private static final int MARGIN = 5;
  private static final int MARKER = 10;

  private static Image stopImg = null;

  private JTextArea           textArea;
  private Collection<Integer> bpLineIndexes;
  private int                 curLineIdx;


  public DebugPrgRowHeader(
		JTextArea           textArea,
		Collection<Integer> bpLineIndexes )
  {
    this.textArea      = textArea;
    this.bpLineIndexes = bpLineIndexes;
    this.curLineIdx    = -1;
    if( stopImg == null ) {
      URL url = getClass().getResource( "/images/debug/bp_stop.png" );
      if( url != null ) {
	stopImg = getToolkit().createImage( url );
      }
    }
  }


  public void setCurLineIndex( int lineIdx )
  {
    this.curLineIdx = lineIdx;
    repaint();
  }


  @Override
  public Dimension getPreferredSize()
  {
    Dimension rv = null;
    if( isPreferredSizeSet() ) {
      rv = super.getPreferredSize();
    } else {
      int       w  = 0;
      int       h  = 0;
      if( stopImg != null ) {
	w = stopImg.getWidth( this );
	if( w < 0 ) {
	  w = 20;
	}
	w += (MARGIN + MARKER);
      }
      Dimension prefTextSize = this.textArea.getPreferredSize();
      if( prefTextSize != null ) {
	h = prefTextSize.height;
      }
      rv = new Dimension( w, h );
    }
   return rv;
  }


  @Override
  public void paintComponent( Graphics g )
  {
    super.paintComponent( g );
    int wComp = getWidth();
    if( wComp > 0 ) {
      g = g.create();
      try {
	g.setPaintMode();

	// mark current line
	if( this.curLineIdx >= 0 ) {
	  try {
	    Rectangle2D r = UIUtil.modelToView(
					this.textArea,
					this.textArea.getLineStartOffset(
							this.curLineIdx ) );
	    if( r != null ) {
	      int yRow = (int) Math.round( r.getY() );
	      int hRow = (int) Math.round( r.getHeight() );

	      int[] xPoints = new int[] {
					wComp - MARKER,
					wComp - 1,
					wComp - MARKER };
	      int[] yPoints = new int[] {
					yRow,
					yRow + (hRow / 2),
					yRow + hRow };
	      g.setColor( Color.RED );
	      g.fillPolygon(
			xPoints,
			yPoints,
			Math.min( xPoints.length, yPoints.length ) );
	    }
	  }
	  catch( BadLocationException ex ) {}
	}

	// draw breakpoints
	if( stopImg != null ) {
	  int hImg = stopImg.getWidth( this );
	  if( hImg > 0 ) {
	    for( Integer lineIdx : this.bpLineIndexes ) {
	      try {
		Rectangle2D r = UIUtil.modelToView(
					this.textArea,
					this.textArea.getLineStartOffset(
						lineIdx.intValue() ) );
		if( r != null ) {
		  int yRow = (int) Math.round( r.getY() );
		  int hRow = (int) Math.round( r.getHeight() );

		  yRow += ((hRow - hImg) / 2);
		  g.drawImage( stopImg, MARGIN, yRow + 5, this );
		}
	      }
	      catch( BadLocationException ex ) {}
	    }
	  }
	}
      }
      finally {
	g.dispose();
      }
    }
  }
}
