/*
 * (c) 2025 Jens Mueller
 *
 * JBASICODE
 *
 * Progran File Downloader
 */

package jbasicode.ui;

import java.awt.EventQueue;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DropTargetDropEvent;
import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.io.UnsupportedEncodingException;
import java.net.HttpURLConnection;
import java.net.URI;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLDecoder;
import java.text.MessageFormat;
import javax.swing.ProgressMonitorInputStream;
import jbasicode.Main;
import jbasicode.etc.ExtByteArrayOutputStream;
import jbasicode.etc.IOUtil;


public class Downloader extends Thread
{
  private TextEditFld textEditFld;
  private URL         url;
  private String      fileName;


  public static boolean checkAndStart(
				TextEditFld         textEditFld,
				File                file,
				DropTargetDropEvent e )
  {
    boolean done   = false;
    boolean delete = false;
    if( e != null ) {
      int dndAction = e.getDropAction();
      if( ((dndAction == DnDConstants.ACTION_COPY_OR_MOVE)
		|| (dndAction == DnDConstants.ACTION_MOVE))
	  && ((file.lastModified() + 1000L) > System.currentTimeMillis()) )
      {
	/*
	 * If the drop action allows deleting the source and
	 * the url file has just been created,
	 * then delete url file.
	 */
	delete = true;
      }
    }

    // read file and extract url
    URL url = null;
    try {
      String fName = file.getName();
      if( fName != null ) {
	if( fName.toLowerCase().endsWith( ".url" ) ) {
	  try( BufferedReader in = new BufferedReader(
					new FileReader( file ) ) )
	  {
	    boolean header = false;
	    String  line   = in.readLine();
	    while( line != null ) {
	      if( line.trim().equals( "[InternetShortcut]" ) ) {
		header = true;
	      } else {
		if( header && (line.length() > 4)
		    && line.startsWith( "URL=" ) )
		{
		  url = (new URI( line.substring( 4 ) )).toURL();
		  break;
		}
	      }
	      line = in.readLine();
	    }
	  }
	}
      }
    }
    catch( Exception ex ) {}

    // start download
    if( url != null ) {
      String fileName = url.getFile();
      if( fileName != null ) {
	try {
	  fileName = URLDecoder.decode( fileName, "UTF-8" );
	}
	catch( UnsupportedEncodingException ex ) {}
	int idx  = fileName.lastIndexOf( '/' );
	if( idx >= 0 ) {
	  fileName = fileName.substring( idx + 1 );
	}
	if( fileName.isEmpty() ) {
	  fileName = null;
	}
      }
      (new Downloader( textEditFld, url, fileName )).start();
      if( delete ) {
	file.delete();
      }
      done = true;
    }
    return done;
  }


	/* --- ueberschriebene Methoden --- */

  @Override
  public void run()
  {
    ExtByteArrayOutputStream fileBuf = null;
    IOException              fileEx  = null;
    URLConnection            urlCon  = null;
    InputStream              in      = null;
    try {
      urlCon = this.url.openConnection();
      urlCon.setReadTimeout( 20000 );
      long contentLen = urlCon.getContentLengthLong();
      if( contentLen > 0 ) {
	TextEditFld.checkFileSize( contentLen );
      }
      String info = this.fileName;
      if( info == null ) {
	info = this.url.toExternalForm();
      }
      in = new BufferedInputStream(
			new ProgressMonitorInputStream(
				this.textEditFld,
				MessageFormat.format(
					Main.getText( "title.download" ),
					info ),
				urlCon.getInputStream() ) );
      if( contentLen > 0 ) {
	fileBuf = new ExtByteArrayOutputStream( (int) contentLen );
      } else {
	fileBuf = new ExtByteArrayOutputStream( 0x4000 );
      }
      int n = 0;
      int b = in.read();
      while( b >= 0 ) {
	TextEditFld.checkFileSize( n );
	fileBuf.write( b );
	n++;
	b = in.read();
      }
      if( urlCon instanceof HttpURLConnection ) {
	((HttpURLConnection) urlCon).disconnect();
      }
    }
    catch( InterruptedIOException ex ) {}
    catch( IOException ex ) {
      fileEx = ex;
    }
    finally {
      IOUtil.closeSilently( in );
    }
    if( (fileBuf != null) || (fileEx != null) ) {
      final ExtByteArrayOutputStream fileBuf1 = fileBuf;
      final IOException              fileEx1  = fileEx; 
      EventQueue.invokeLater(
			new Runnable()
			{
			  @Override
			  public void run()
			  {
			    downloadFinished( fileBuf1, fileEx1 );
			  }
			} );
    }
  }


	/* --- Konstruktor --- */

  private Downloader(
		TextEditFld textEditFld,
		URL         url,
		String      fileName )
  {
    super( Main.APPNAME + " Downloader" );
    this.textEditFld = textEditFld;
    this.url         = url;
    this.fileName    = fileName;
  }


	/* --- private Methoden --- */

  private void downloadFinished(
			ExtByteArrayOutputStream fileBuf,
			IOException              fileEx )
  {
    if( fileEx != null ) {
      UIUtil.fireShowErrorMsg( this.textEditFld, null, fileEx );
    }
    else if( fileBuf != null ) {
      try( BufferedInputStream in = new BufferedInputStream(
					fileBuf.createInputStream() ) )
      {
	this.textEditFld.openFile( in, null, this.fileName );
      }
      catch( IOException ex ) {
	UIUtil.fireShowErrorMsg( this.textEditFld, null, ex );
      }
    }
  }
}
