/*
 * (c) 2025 Jens Mueller
 *
 * JBASICODE
 *
 * Dialog for searching text
 */

package jbasicode.ui;

import java.awt.EventQueue;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import jbasicode.Main;


public class FindAndReplaceDlg extends JDialog implements ActionListener
{
  public enum Action { FIND_NEXT, REPLACE_ALL, CANCEL };

  private Action     action;
  private String     searchText;
  private String     replaceText;
  private JTextField fldSearchText;
  private JTextField fldReplaceText;
  private JCheckBox  cbCaseSensitive;
  private JButton    btnSearch;
  private JButton    btnReplaceAll;
  private JButton    btnCancel;


  public static FindAndReplaceDlg createFindDlg(
					Window owner,
					String searchText )
  {
    return new FindAndReplaceDlg( owner, searchText, null, false, null );
  }


  public static FindAndReplaceDlg createFindAndReplaceDlg(
					Window  owner,
					String  searchText,
					boolean caseSensitive,
					String  replaceText )
  {
    return new FindAndReplaceDlg(
			owner,
			searchText,
			caseSensitive,
			true,
			replaceText );
  }


  public Action getAction()
  {
    return this.action;
  }


  public boolean getCaseSensitive()
  {
    return this.cbCaseSensitive != null ?
			this.cbCaseSensitive.isSelected()
			: false;
  }


  public String getReplaceText()
  {
    return this.replaceText != null ? this.replaceText : "";
  }


  public String getSearchText()
  {
    return this.searchText != null ? this.searchText : "";
  }


	/* --- Constructor --- */

  private FindAndReplaceDlg(
			Window  owner,
			String  searchText,
			Boolean caseSensitive,
			boolean withReplace,
			String  replaceText )
  {
    super( owner, JDialog.ModalityType.DOCUMENT_MODAL );
    setTitle( Main.getText( "title.find_text" ) );
    setDefaultCloseOperation( JDialog.DISPOSE_ON_CLOSE );
    this.action      = Action.CANCEL;
    this.searchText  = null;
    this.replaceText = null;


    // Layout
    setLayout( new GridBagLayout() );

    GridBagConstraints gbc = new GridBagConstraints(
					0, 0,
					1, 1,
					0.0, 0.0,
					GridBagConstraints.EAST,
					GridBagConstraints.NONE,
					new Insets( 5, 5, 5, 5 ),
					0, 0 );


    // Text input area
    add( new JLabel( Main.getText( "label.search_for" ) ), gbc );

    this.fldSearchText = new JTextField();
    if( searchText != null ) {
      this.fldSearchText.setText( searchText );
    }
    gbc.anchor  = GridBagConstraints.WEST;
    gbc.fill    = GridBagConstraints.HORIZONTAL;
    gbc.weightx = 1.0;
    gbc.gridx++;
    add( this.fldSearchText, gbc );

    if( withReplace ) {
      gbc.anchor  = GridBagConstraints.EAST;
      gbc.fill    = GridBagConstraints.NONE;
      gbc.weightx = 0.0;
      gbc.gridx   = 0;
      gbc.gridy++;
      add( new JLabel( Main.getText( "label.replace_with" ) ), gbc );

      this.fldReplaceText = new JTextField();
      if( replaceText != null ) {
	this.fldReplaceText.setText( replaceText );
      }
      gbc.anchor  = GridBagConstraints.WEST;
      gbc.fill    = GridBagConstraints.HORIZONTAL;
      gbc.weightx = 1.0;
      gbc.gridx++;
      add( this.fldReplaceText, gbc );
    } else {
      this.fldReplaceText = null;
    }


    // Upper/lower case
    gbc.fill    = GridBagConstraints.NONE;
    gbc.weightx = 0.0;
    if( caseSensitive != null ) {
      gbc.gridy++;
      this.cbCaseSensitive = new JCheckBox(
				Main.getText( "option.case_sensitive" ),
				caseSensitive );
      add( this.cbCaseSensitive, gbc );
    } else {
      this.cbCaseSensitive = null;
    }


    // Actions
    JPanel panelBtn = new JPanel();
    panelBtn.setLayout( new GridLayout( 1, withReplace ? 3 : 2, 5, 5 ) );

    this.btnSearch = new JButton( Main.getText( "action.edit.search" ) );
    panelBtn.add( this.btnSearch );

    if( withReplace ) {
      this.btnReplaceAll = new JButton(
				Main.getText( "action.edit.replace_all" ) );
      panelBtn.add( this.btnReplaceAll );
    } else {
      this.btnReplaceAll = null;
    }

    this.btnCancel = new JButton( Main.getText( "action.cancel" ) );
    panelBtn.add( this.btnCancel );

    gbc.anchor        = GridBagConstraints.CENTER;
    gbc.insets.left   = 10;
    gbc.insets.right  = 10;
    gbc.insets.bottom = 10;
    gbc.gridwidth     = GridBagConstraints.REMAINDER;
    gbc.gridx         = 0;
    gbc.gridy++;
    add( panelBtn, gbc );


    // Listeners
    this.fldSearchText.addActionListener( this );
    if( this.fldReplaceText != null ) {
      this.fldReplaceText.addActionListener( this );
    }
    this.btnSearch.addActionListener( this );
    if( this.btnReplaceAll != null ) {
      this.btnReplaceAll.addActionListener( this );
    }
    this.btnCancel.addActionListener( this );


    // Window bounds
    pack();
    UIUtil.setParentCentered( this );
    setResizable( false );

    // Focus
    EventQueue.invokeLater( ()->this.fldSearchText.requestFocus() );
  }


	/* --- ActionListener --- */

  @Override
  public void actionPerformed( ActionEvent e )
  {
    Object  src = e.getSource();
    if( (src == this.fldReplaceText) || (src == this.btnSearch) ) {
      doFind( Action.FIND_NEXT );
    }
    else if( src == this.fldSearchText ) {
      if( this.fldReplaceText != null ) {
	this.fldReplaceText.requestFocus();
      } else {
	doFind( Action.FIND_NEXT );
      }
    }
    else if( src == this.btnReplaceAll ) {
      doFind( Action.REPLACE_ALL );
    }
    else if( src == this.btnCancel ) {
      doClose();
    }
  }


	/* --- Private Methods --- */

  private void doFind( Action action )
  {
    this.searchText = this.fldSearchText.getText();
    if( this.fldReplaceText != null  ) {
      this.replaceText = this.fldReplaceText.getText();
    }
    if( this.searchText != null ) {
      if( !this.searchText.isEmpty() ) {
	this.action = action;
	doClose();
      }
    }
  }


  private void doClose()
  {
    setVisible( false );
    dispose();
  }
}
