/*
 * (c) 2025 Jens Mueller
 *
 * JBASICODE
 *
 * Window for displaying help
 */

package jbasicode.ui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.io.IOException;
import java.net.URL;
import java.util.EmptyStackException;
import java.util.Properties;
import java.util.Stack;
import javax.swing.JButton;
import javax.swing.JEditorPane;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JScrollPane;
import javax.swing.JToolBar;
import javax.swing.JViewport;
import javax.swing.Timer;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;
import javax.swing.event.MenuEvent;
import javax.swing.event.MenuListener;
import javax.swing.text.Document;
import jbasicode.Main;


public class HelpFrm extends JFrame implements
					ActionListener,
					HyperlinkListener,
					MenuListener
{
  public static final String PROP_PREFIX = Main.PROP_PREFIX + "help.";

  private static final String HOME_PAGE    = "/help/home.htm";
  private static final String LICENSE_PAGE = "/help/license.htm";


  private static class PageStackEntry
  {
    private URL    url;
    private Double viewPos;	// range 0.0 to 1.0

    private PageStackEntry( URL url )
    {
      this.url     = url;
      this.viewPos = null;
    }
  };


  private boolean               notified;
  private URL                   urlHome;
  private URL                   urlLicense;
  private Stack<PageStackEntry> pageStack;
  private String                pageResource;
  private Double                scrollToPos;
  private Timer                 scrollTimer;
  private JButton               btnBack;
  private JButton               btnHome;
  private JMenu                 mnuEdit;
  private JMenuItem             mnuClose;
  private JMenuItem             mnuCopy;
  private JMenuItem             mnuSelectAll;
  private JEditorPane           editorPane;
  private JScrollPane           scrollPane;


  public HelpFrm()
  {
    this.notified    = false;
    this.pageStack   = new Stack<>();
    this.scrollToPos = null;
    this.scrollTimer = new Timer( 500, this );
    this.scrollTimer.setRepeats( false );
    this.urlHome = getClass().getResource(
				Main.getText( "resource.help.home" ) );
    this.urlLicense = getClass().getResource(
				Main.getText( "resource.help.license" ) );
    UIUtil.setIconImagesAt( this );
    setDefaultCloseOperation( JFrame.DISPOSE_ON_CLOSE );
    setTitle( Main.APPNAME + "\u0020"
				+ Main.getText( "title.window.help" ) );


    // Menu
    JMenu mnuFile = new JMenu( Main.getText( "menu.file" ) );

    this.mnuClose = new JMenuItem( Main.getText( "action.close" ) );
    mnuFile.add( this.mnuClose );

    this.mnuEdit = new JMenu( Main.getText( "menu.edit" ) );

    this.mnuCopy = new JMenuItem( Main.getText( "action.edit.copy" ) );
    UIUtil.setShortcutAt( this.mnuCopy, KeyEvent.VK_C, false );
    this.mnuEdit.add( this.mnuCopy );
    this.mnuEdit.addSeparator();

    this.mnuSelectAll = new JMenuItem(
				Main.getText( "action.edit.select_all" ) );
    UIUtil.setShortcutAt( this.mnuSelectAll, KeyEvent.VK_A, false );
    mnuEdit.add( this.mnuSelectAll );

    JMenuBar mnuBar = new JMenuBar();
    mnuBar.add( mnuFile );
    mnuBar.add( this.mnuEdit );
    setJMenuBar( mnuBar );


    // Window content
    setLayout( new BorderLayout() );

    JToolBar toolBar = new JToolBar();
    toolBar.setFloatable( false );
    toolBar.setRollover( true );
    add( toolBar, BorderLayout.NORTH );

    this.btnBack = UIUtil.createImageButton(
				this,
				"/images/nav/back.png",
				Main.getText( "action.nav.back" ) );
    toolBar.add( this.btnBack );

    this.btnHome = UIUtil.createImageButton(
				this,
				"/images/nav/home.png",
				Main.getText( "action.nav.home" ) );
    toolBar.add( this.btnHome );

    this.editorPane = new JEditorPane();
    this.editorPane.setEditable( false );
    this.editorPane.setPreferredSize( new Dimension( 500, 400 ) );

    this.scrollPane = new JScrollPane( editorPane );
    add( this.scrollPane, BorderLayout.CENTER );


    // Window bounds
    if( !UIUtil.restoreWindowBounds(
			this,
			Main.getProperties(),
			PROP_PREFIX ) )
    {
      pack();
      setLocationByPlatform( true );
    }
  }


  public void doClose()
  {
    setVisible( false );
    dispose();
  }


  public void showHomePage()
  {
    setPage( true, this.urlHome );
  }


  public void showLicensePage()
  {
    setPage( true, this.urlLicense );
  }


  @Override
  public void addNotify()
  {
    super.addNotify();
    if( !this.notified ) {
      this.notified = true;
      this.btnBack.addActionListener( this );
      this.btnHome.addActionListener( this );
      this.mnuClose.addActionListener( this );
      this.mnuCopy.addActionListener( this );
      this.mnuSelectAll.addActionListener( this );
      this.mnuEdit.addMenuListener( this );
      this.editorPane.addHyperlinkListener( this );
    }
  }


  @Override
  public void removeNotify()
  {
    super.removeNotify();
    if( this.notified ) {
      this.notified = false;
      this.btnBack.removeActionListener( this );
      this.btnHome.removeActionListener( this );
      this.mnuClose.removeActionListener( this );
      this.mnuCopy.removeActionListener( this );
      this.mnuSelectAll.removeActionListener( this );
      this.mnuEdit.removeMenuListener( this );
      this.editorPane.removeHyperlinkListener( this );
    }
  }


	/* --- ActionListener --- */

  @Override
  public void actionPerformed( ActionEvent e )
  {
    Object src = e.getSource();
    if( src == this.btnBack ) {
      doBack();
    } else if( src == this.btnHome ) {
      doHome();
    } else if( src == this.mnuClose ) {
      doClose();
    } else if( src == this.mnuCopy ) {
      this.editorPane.copy();
      this.editorPane.requestFocus();
    } else if( src == this.mnuSelectAll ) {
      this.editorPane.selectAll();
      this.editorPane.requestFocus();
    } else if( src == this.scrollTimer ) {
      doScrollTo();
    }
  }


	/* --- HyperlinkListener --- */

  @Override
  public void hyperlinkUpdate( HyperlinkEvent e )
  {
    if( e.getEventType() == HyperlinkEvent.EventType.ACTIVATED )
      setPage( true, e.getURL() );
  }


	/* --- MenuListener --- */

  @Override
  public void menuCanceled( MenuEvent e )
  {
    // empty
  }


  @Override
  public void menuDeselected( MenuEvent e )
  {
    // empty
  }


  @Override
  public void menuSelected( MenuEvent e )
  {
    if( e.getSource() == this.mnuEdit ) {
      int begPos = this.editorPane.getSelectionStart();
      this.mnuCopy.setEnabled(
	(begPos >= 0) && (begPos < this.editorPane.getSelectionEnd()) );
    }
  }


	/* --- Aktionen --- */

  private void doBack()
  {
    if( this.pageStack.size() > 1 ) {
      try {
	this.pageStack.pop();
	PageStackEntry entry = this.pageStack.pop();
	if( setPage( false, entry.url ) ) {
	  this.scrollToPos = entry.viewPos;
	  this.scrollTimer.restart();
	}
      }
      catch( EmptyStackException ex ) {}
    }
  }


  private void doHome()
  {
    this.pageStack.clear();
    showHomePage();
  }


  private void doScrollTo()
  {
    this.scrollTimer.stop();
    Double scrollToPos = this.scrollToPos;
    this.scrollToPos   = null;
    if( scrollToPos != null ) {
      double d = scrollToPos.doubleValue();
      int    h = this.editorPane.getHeight();
      if( (d > 0.0) && (d <= 1.0) && (h > 0) ) {
	JViewport vp = this.scrollPane.getViewport();
	if( vp != null ) {
	  vp.setViewPosition(
		new Point( 0, (int) Math.round( d * (double) h ) ) );
	}
      }
    }
  }


	/* --- Private Methods --- */

  private boolean setPage( boolean saveCurViewPos, URL url )
  {
    boolean status = false;
    if( url == null ) {
      this.pageStack.clear();
      url = this.urlHome;
    }
    if( url != null ) {

      // show page only if it differs from the previous one
      boolean        alreadyVisible = false;
      PageStackEntry topEntry       = null;
      if( !this.pageStack.isEmpty() ) {
	topEntry = this.pageStack.peek();
	if( (url != null) && (topEntry.url != null) ) {
	  if( topEntry.url.equals( url ) ) {
	    alreadyVisible = true;
	  }
	}
      }
      if( !alreadyVisible ) {
	try {

	  // keep current position in mind
	  if( saveCurViewPos && (topEntry != null) ) {
	    topEntry.viewPos = null;
	    int h = this.editorPane.getHeight();
	    if( h > 0 ) {
	      JViewport vp = this.scrollPane.getViewport();
	      if( vp != null ) {
		Point pt = vp.getViewPosition();
		if( pt != null ) {
		  double d = (double) pt.y / (double) h;
		  if( (d > 0.0) && (d <= 1.0) ) {
		    topEntry.viewPos = d;
		  }
		}
	      }
	    }
	  }

	  // show new page
	  if( url != null ) {
	    Document doc = this.editorPane.getDocument();
	    if( doc != null ) {
	      doc.putProperty( Document.StreamDescriptionProperty, null );
	    }
	    this.editorPane.setPage( url );
	    status = true;
	  }

	  /*
	   * If the page could be displayed
	   * then create a new stack entry
	   */
	  this.pageStack.push( new PageStackEntry( url ) );

	  // update back action
	  boolean stateBack = (this.pageStack.size() > 1);
	  this.btnBack.setEnabled( stateBack );
	}
	catch( Exception ex ) {
	  UIUtil.showErrorMsg( this, null, ex );
	}
      }
    }
    return status;
  }
}
