/*
 * (c) 2025 Jens Mueller
 *
 * JBASICODE
 *
 * Log Component
 */

package jbasicode.ui;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Font;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import jbasicode.Main;


public class LogFld extends JPanel implements ActionListener, MouseListener
{
  private MainFrm     mainFrm;
  private boolean     notified;
  private int         editorLineNum;
  private String      errLineStart;
  private String      warnLineStart;
  private JPopupMenu  popupMenu;
  private JMenuItem   popupGotoEditorLine;
  private JMenuItem   popupCopy;
  private JMenuItem   popupSelectAll;
  private JTextArea   textArea;
  private JScrollPane scrollPane;


  public LogFld( MainFrm mainFrm )
  {
    this.mainFrm       = mainFrm;
    this.notified      = false;
    this.editorLineNum = -1;

    this.errLineStart  = getTrimmedAtFieldText( "bc.error_in_line" );
    this.warnLineStart = getTrimmedAtFieldText( "bc.warning_in_line" );

    this.popupMenu = new JPopupMenu();

    this.popupGotoEditorLine = new JMenuItem(
			Main.getText( "action.log.goto_editor_line" ) );
    this.popupMenu.add( this.popupGotoEditorLine );
    this.popupMenu.addSeparator();

    this.popupCopy = new JMenuItem( Main.getText( "action.edit.copy" ) );
    this.popupMenu.add( this.popupCopy );

    this.popupSelectAll = new JMenuItem(
				Main.getText( "action.edit.select_all" ) );
    this.popupMenu.add( this.popupSelectAll );

    setLayout( new BorderLayout() );

    this.textArea = new JTextArea( 4, 1 );
    this.textArea.setEditable( false );

    Font font = this.textArea.getFont();
    this.textArea.setFont(
		new Font(
			Font.MONOSPACED,
			Font.PLAIN,
			font != null ? font.getSize() : 13 ) );
    this.scrollPane = new JScrollPane( this.textArea );
    add( this.scrollPane, BorderLayout.CENTER );
  }


  public void appendText( String text )
  {
    if( text != null )
      this.textArea.append( text );
  }


  public void clear()
  {
    this.textArea.setText( "" );
  }


  public void resetPrefSize()
  {
    this.textArea.setColumns( 0 );
    this.textArea.setRows( 0 );
  }


  @Override
  public void addNotify()
  {
    super.addNotify();
    if( !this.notified ) {
      this.notified = true;
      this.popupGotoEditorLine.addActionListener( this );
      this.popupCopy.addActionListener( this );
      this.popupSelectAll.addActionListener( this );
      this.textArea.addCaretListener( this.mainFrm );
      this.textArea.addMouseListener( this );
      this.scrollPane.addMouseListener( this );
    }
  }


  @Override
  public void removeNotify()
  {
    super.removeNotify();
    if( this.notified ) {
      this.notified = false;
      this.popupGotoEditorLine.removeActionListener( this );
      this.popupCopy.removeActionListener( this );
      this.popupSelectAll.removeActionListener( this );
      this.textArea.removeCaretListener( this.mainFrm );
      this.textArea.removeMouseListener( this );
      this.scrollPane.removeMouseListener( this );
    }
  }


	/* --- ActionListener --- */

  @Override
  public void actionPerformed( ActionEvent e )
  {
    Object src = e.getSource();
    if( src == this.popupGotoEditorLine ) {
      gotoEditorLine( this.editorLineNum );
    } else if( src == this.popupCopy ) {
      this.textArea.copy();
    } else if( src == this.popupSelectAll ) {
      this.textArea.selectAll();
    }
  }


	/* --- MouseListener --- */

  @Override
  public void mouseClicked( MouseEvent e )
  {
    if( e.isPopupTrigger() ) {
      showPopupMenu( e );
      e.consume();
    }
    else if( (e.getButton() == MouseEvent.BUTTON1)
	     && (e.getClickCount() > 1) )
    {
      gotoEditorLine( getEditorLineNum( e ) );
      e.consume();
    }
  }


  @Override
  public void mouseEntered( MouseEvent e )
  {
    // empty
  }


  @Override
  public void mouseExited( MouseEvent e )
  {
    // empty
  }


  @Override
  public void mousePressed( MouseEvent e )
  {
    if( e.isPopupTrigger() ) {
      showPopupMenu( e );
      e.consume();
    }
  }


  @Override
  public void mouseReleased( MouseEvent e )
  {
    if( e.isPopupTrigger() ) {
      showPopupMenu( e );
      e.consume();
    }
  }


	/* --- Private Methods --- */

  private static String getTrimmedAtFieldText( String textKey )
  {
    String text = Main.getText( textKey );
    int    idx  = text.indexOf( '{' );
    if( idx >= 0 ) {
      text = text.substring( 0, idx );
    }
    return text;
  }


  private int getEditorLineNum( MouseEvent e )
  {
    int lineNum = -1;
    if( e.getComponent() == this.textArea ) {
      int idx = UIUtil.viewToModel( this.textArea, e.getPoint() );
      if( idx >= 0 ) {
	String text = this.textArea.getText();
	if( text != null ) {
	  int len = text.length();
	  if( idx <= len ) {
	    while( idx > 0 ) {
	      if( text.charAt( idx - 1 ) == '\n' ) {
		break;
	      }
	      --idx;
	    }
	    if( idx < len ) {
	      int pos = -1;
	      if( text.indexOf( this.errLineStart, idx ) == idx ) {
		pos = idx + this.errLineStart.length();
	      } else if( text.indexOf( this.warnLineStart, idx ) == idx ) {
		pos = idx + this.warnLineStart.length();
	      }
	      if( pos >= 0 ) {
		char ch = text.charAt( pos++ );
		if( (ch >= '0') && (ch <= '9') ) {
		  lineNum = ch - '0';
		  ch = text.charAt( pos++ );
		  while( (ch >= '0') && (ch <= '9') ) {
		    lineNum = (lineNum * 10) + (ch - '0');
		    ch = text.charAt( pos++ );
		  }
		}
	      }
	    }
	  }
	}
      }
    }
    return lineNum;
  }


  private void gotoEditorLine( int editorLineNum )
  {
    if( editorLineNum > 0 )
      this.mainFrm.fireGotoEditorLine( editorLineNum );
  }


  private void showPopupMenu( MouseEvent e )
  {
    this.editorLineNum = getEditorLineNum( e );
    this.popupGotoEditorLine.setEnabled( this.editorLineNum > 0 );

    boolean state = UIUtil.hasSelection( this.textArea );
    this.popupCopy.setEnabled( state );

    this.popupMenu.show(
		e.getComponent(),
		e.getX(),
		e.getY() );
  }
}
