/*
 * (c) 2025 Jens Mueller
 *
 * JBASICODE
 *
 * Main Window
 */

package jbasicode.ui;

import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.EventQueue;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.Image;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.FlavorEvent;
import java.awt.datatransfer.FlavorListener;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.InputEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.net.URL;
import java.text.CharacterIterator;
import java.text.MessageFormat;
import java.text.StringCharacterIterator;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JEditorPane;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JSpinner;
import javax.swing.JTextArea;
import javax.swing.JToolBar;
import javax.swing.KeyStroke;
import javax.swing.SpinnerListModel;
import javax.swing.Timer;
import javax.swing.event.CaretEvent;
import javax.swing.event.CaretListener;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.swing.text.Document;
import javax.swing.text.JTextComponent;
import jbasicode.Main;
import jbasicode.bc.BCAudioCreator;
import jbasicode.bc.BCRuntime;
import jbasicode.bc.BCScreen;
import jbasicode.bc.ProgramParser;
import jbasicode.etc.IOUtil;


public class MainFrm extends JFrame implements
					ActionListener,
					CaretListener,
					ChangeListener,
					FlavorListener
{
  public static final int STATUS_OTHER_FILE  = 0x01;
  public static final int STATUS_SAVE_ACTION = 0x02;
  public static final int STATUS_TITLE       = 0x04;

  private static final String CARD_DEBUG        = "debug";
  private static final String CARD_EDIT         = "edit";
  private static final String CARD_NORMAL       = "normal";
  private static final String CARD_PRESENTATION = "presentation";

  private static final String PROP_PREFIX  = Main.PROP_PREFIX + "main.";
  private static final String PROP_VERSION = "version";
  private static final String PROP_AUDIO_ENABLED = "audio.enabled";

  private static final String MAX_SPEED_ITEM = "Max";

  private enum RunMode { NORMAL, DEBUG, PRESENTATION };

  private Object              bcLock;
  private volatile BCRuntime  bcRuntime;
  private BCScreen            bcScreen;
  private Clipboard           clipboard;
  private JTextComponent      curTextComp;
  private int                 curPrintDocNum;
  private Map<Integer,Object> printDocs;
  private PrinterOutputFrm    printerOutputFrm;
  private HelpFrm             helpFrm;
  private BCRuntime.EOR       convertToEOR;
  private File                recentDataFile;
  private File                recentWavFile;
  private String              defaultStatusText;
  private TextFinder          textFinder;
  private volatile boolean    bcBreakKeyEnabled;
  private RunMode             runMode;
  private CardLayout          layoutLeft;
  private CardLayout          layoutMain;
  private JPanel              panelLeft;
  private JPanel              panelMain;
  private JButton             btnFileNew;
  private JButton             btnFileOpen;
  private JButton             btnFileSave;
  private JButton             btnFilePrint;
  private JButton             btnEditUndo;
  private JButton             btnEditCut;
  private JButton             btnEditCopy;
  private JButton             btnEditPaste;
  private JButton             btnEditFind;
  private JButton             btnRun;
  private JButton             btnStop;
  private JButton             btnDebugStepInto;
  private JButton             btnDebugStepOver;
  private JButton             btnDebugStepUp;
  private JCheckBox           cbAudioEnabled;
  private JMenuItem           mnuFileNew;
  private JMenuItem           mnuFileOpen;
  private JMenuItem           mnuFileSave;
  private JMenuItem           mnuFileSaveAs;
  private JMenuItem           mnuFilePrint;
  private JMenuItem           mnuFilePrintOptions;
  private JMenuItem           mnuExportPrgToWAV;
  private JMenuItem           mnuExportDataFileToWAV;
  private JMenuItem           mnuConvertDataFileEOR;
  private JMenuItem           mnuQuit;
  private JMenuItem           mnuEditCopy;
  private JMenuItem           mnuEditCut;
  private JMenuItem           mnuEditFind;
  private JMenuItem           mnuEditFindNext;
  private JMenuItem           mnuEditFindPrev;
  private JMenuItem           mnuEditGotoLine;
  private JMenuItem           mnuEditPaste;
  private JMenuItem           mnuEditReplace;
  private JMenuItem           mnuEditToUpper;
  private JMenuItem           mnuEditSelectAll;
  private JMenuItem           mnuEditUndo;
  private JMenuItem           mnuPrgCheckSyntax;
  private JMenuItem           mnuPrgPrinterOutput;
  private JMenuItem           mnuPrgRunNormal;
  private JMenuItem           mnuPrgRunLarge;
  private JMenuItem           mnuPrgAbort;
  private JMenuItem           mnuPrgAbortESC;
  private JMenuItem           mnuDebugBpAdd;
  private JMenuItem           mnuDebugBpRemoveAll;
  private JMenuItem           mnuDebugOpen;
  private JMenuItem           mnuDebugStop;
  private JMenuItem           mnuDebugRun;
  private JMenuItem           mnuDebugStepOver;
  private JMenuItem           mnuDebugStepInto;
  private JMenuItem           mnuDebugStepUp;
  private JMenuItem           mnuSettingsOpen;
  private JMenuItem           mnuSettingsSave;
  private JMenuItem           mnuSettingsDelete;
  private JMenuItem           mnuHelpAbout;
  private JMenuItem           mnuHelpHome;
  private JMenuItem           mnuHelpLicense;
  private JSpinner            spinnerSpeed;
  private JLabel              labelStatus;
  private DebugFld            debugFld;
  private LogFld              logFld;
  private TextEditFld         textEditFld;
  private ScreenFld           smallScreenFld;
  private ScreenFld           largeScreenFld;
  private WindowAdapter       windowAdapter;
  private Timer               statusTimer;


  public MainFrm( String prgFileName, boolean runFlag )
  {
    this.bcLock             = new Object();
    this.bcRuntime          = null;
    this.bcScreen           = new BCScreen( this );
    this.clipboard          = getToolkit().getSystemClipboard();
    this.curTextComp        = null;
    this.curPrintDocNum     = 0;
    this.printDocs          = new HashMap<>();
    this.printerOutputFrm   = null;
    this.helpFrm            = null;
    this.convertToEOR       = BCRuntime.getEndOfRecord();
    this.recentDataFile     = null;
    this.recentWavFile      = null;
    this.defaultStatusText  = Main.getText( "status.default" );
    this.textFinder         = new TextFinder();
    this.bcBreakKeyEnabled  = true;
    this.runMode            = RunMode.NORMAL;
    UIUtil.setIconImagesAt( this );
    setDefaultCloseOperation( JFrame.DO_NOTHING_ON_CLOSE );
    setLayout( new BorderLayout( 0, 0 ) );


    // File menu
    JMenu mnuFile = new JMenu( Main.getText( "menu.file" ) );

    this.mnuFileNew = new JMenuItem( Main.getText( "action.file.new" ) );
    UIUtil.setShortcutAt( this.mnuFileNew, KeyEvent.VK_N, false );
    mnuFile.add( this.mnuFileNew );

    this.mnuFileOpen = new JMenuItem( Main.getText( "action.file.open" ) );
    UIUtil.setShortcutAt( this.mnuFileOpen, KeyEvent.VK_O, false );
    mnuFile.add( this.mnuFileOpen );
    mnuFile.addSeparator();

    this.mnuFileSave = new JMenuItem( Main.getText( "action.file.save" ) );
    UIUtil.setShortcutAt( this.mnuFileSave, KeyEvent.VK_S, false );
    mnuFile.add( this.mnuFileSave );

    this.mnuFileSaveAs = new JMenuItem(
				Main.getText( "action.file.save_as" ) );
    UIUtil.setShortcutAt( this.mnuFileSaveAs, KeyEvent.VK_S, true );
    mnuFile.add( this.mnuFileSaveAs );
    mnuFile.addSeparator();

    this.mnuFilePrintOptions = new JMenuItem(
			Main.getText( "action.file.print_options" ) );
    mnuFile.add( this.mnuFilePrintOptions );

    this.mnuFilePrint = new JMenuItem( Main.getText( "action.file.print" ) );
    UIUtil.setShortcutAt( this.mnuFilePrint, KeyEvent.VK_P, false );
    mnuFile.add( this.mnuFilePrint );
    mnuFile.addSeparator();

    this.mnuExportPrgToWAV = new JMenuItem(
			Main.getText( "action.export.prg_to_wav" ) );
    mnuFile.add( this.mnuExportPrgToWAV );

    this.mnuExportDataFileToWAV = new JMenuItem(
			Main.getText( "action.export.data_file_to_wav" ) );
    mnuFile.add( this.mnuExportDataFileToWAV );
    mnuFile.addSeparator();

    this.mnuConvertDataFileEOR = new JMenuItem(
			Main.getText( "action.convert_data_file_eor" ) );
    mnuFile.add( this.mnuConvertDataFileEOR );
    mnuFile.addSeparator();

    this.mnuQuit = new JMenuItem( Main.getText( "action.quit" ) );
    mnuFile.add( this.mnuQuit );

    // Edit menu
    JMenu mnuEdit = new JMenu( Main.getText( "menu.edit" ) );

    this.mnuEditUndo = new JMenuItem( Main.getText( "action.edit.undo" ) );
    UIUtil.setShortcutAt( this.mnuEditUndo, KeyEvent.VK_Z, false );
    mnuEdit.add( this.mnuEditUndo );
    mnuEdit.addSeparator();

    this.mnuEditCut = new JMenuItem( Main.getText( "action.edit.cut" ) );
    UIUtil.setShortcutAt( this.mnuEditCut, KeyEvent.VK_X, false );
    mnuEdit.add( this.mnuEditCut );

    this.mnuEditCopy = new JMenuItem( Main.getText( "action.edit.copy" ) );
    UIUtil.setShortcutAt( this.mnuEditCopy, KeyEvent.VK_C, false );
    mnuEdit.add( this.mnuEditCopy );

    this.mnuEditPaste = new JMenuItem( Main.getText( "action.edit.paste" ) );
    UIUtil.setShortcutAt( this.mnuEditPaste, KeyEvent.VK_V, false );
    mnuEdit.add( this.mnuEditPaste );
    mnuEdit.addSeparator();

    this.mnuEditFind = new JMenuItem( Main.getText( "action.edit.find" ) );
    UIUtil.setShortcutAt( this.mnuEditFind, KeyEvent.VK_F, false );
    mnuEdit.add( this.mnuEditFind );

    this.mnuEditFindNext = new JMenuItem(
				Main.getText( "action.edit.find_next" ) );
    this.mnuEditFindNext.setAccelerator(
			KeyStroke.getKeyStroke( KeyEvent.VK_F3, 0 ) );
    mnuEdit.add( this.mnuEditFindNext );

    this.mnuEditFindPrev = new JMenuItem(
				Main.getText( "action.edit.find_prev" ) );
    this.mnuEditFindPrev.setAccelerator(
			KeyStroke.getKeyStroke(
					KeyEvent.VK_F3,
					InputEvent.SHIFT_DOWN_MASK ) );
    mnuEdit.add( this.mnuEditFindPrev );

    this.mnuEditReplace = new JMenuItem(
				Main.getText( "action.edit.replace" ) );
    this.mnuEditReplace.setEnabled( false );
    UIUtil.setShortcutAt( this.mnuEditReplace, KeyEvent.VK_R, false );
    mnuEdit.add( this.mnuEditReplace );
    mnuEdit.addSeparator();

    this.mnuEditToUpper = new JMenuItem(
				Main.getText( "action.edit.to_upper" ) );
    this.mnuEditToUpper.setEnabled( false );
    UIUtil.setShortcutAt( this.mnuEditToUpper, KeyEvent.VK_U, false );
    mnuEdit.add( this.mnuEditToUpper );

    this.mnuEditGotoLine = new JMenuItem(
				Main.getText( "action.edit.goto_line" ) );
    this.mnuEditGotoLine.setEnabled( false );
    UIUtil.setShortcutAt( this.mnuEditGotoLine, KeyEvent.VK_G, false );
    mnuEdit.add( this.mnuEditGotoLine );
    mnuEdit.addSeparator();

    this.mnuEditSelectAll = new JMenuItem(
				Main.getText( "action.edit.select_all" ) );
    UIUtil.setShortcutAt( this.mnuEditSelectAll, KeyEvent.VK_A, false );
    mnuEdit.add( this.mnuEditSelectAll );


    // Program menu
    JMenu mnuPrg = new JMenu( Main.getText( "menu.program" ) );

    this.mnuPrgCheckSyntax = new JMenuItem(
			Main.getText( "action.prg.check_syntax" ) );
    UIUtil.setShortcutAt( this.mnuPrgCheckSyntax, KeyEvent.VK_T, false );
    mnuPrg.add( this.mnuPrgCheckSyntax );

    this.mnuPrgRunNormal = new JMenuItem( Main.getText( "action.prg.run" ) );
    this.mnuPrgRunNormal.setAccelerator(
			KeyStroke.getKeyStroke( KeyEvent.VK_F9, 0 ) );
    mnuPrg.add( this.mnuPrgRunNormal );

    this.mnuPrgRunLarge = new JMenuItem(
			Main.getText( "action.prg.run_large" ) );
    this.mnuPrgRunLarge.setAccelerator(
			KeyStroke.getKeyStroke(
					KeyEvent.VK_F9,
					InputEvent.SHIFT_DOWN_MASK ) );
    mnuPrg.add( this.mnuPrgRunLarge );
    mnuPrg.addSeparator();

    this.mnuPrgAbort = new JMenuItem( Main.getText( "action.prg.abort" ) );
    mnuPrg.add( this.mnuPrgAbort );

    this.mnuPrgAbortESC = new JMenuItem( this.mnuPrgAbort.getText() );
    this.mnuPrgAbortESC.setAccelerator(
		KeyStroke.getKeyStroke( KeyEvent.VK_ESCAPE, 0 ) );
    mnuPrg.add( this.mnuPrgAbortESC );
    mnuPrg.addSeparator();

    this.mnuPrgPrinterOutput = new JMenuItem(
			Main.getText( "action.prg.printer_output" ) );
    mnuPrg.add( this.mnuPrgPrinterOutput );


    // Debugger menu
    JMenu mnuDebug = new JMenu( Main.getText( "menu.debug" ) );

    this.mnuDebugOpen = new JMenuItem( Main.getText( "action.debug.open" ) );
    UIUtil.setShortcutAt( this.mnuDebugOpen, KeyEvent.VK_D, false );
    mnuDebug.add( this.mnuDebugOpen );
    mnuDebug.addSeparator();

    this.mnuDebugStop = new JMenuItem( Main.getText( "action.debug.stop" ) );
    this.mnuDebugStop.setAccelerator(
		KeyStroke.getKeyStroke( KeyEvent.VK_F4, 0 ) );
    mnuDebug.add( this.mnuDebugStop );

    this.mnuDebugRun = new JMenuItem( Main.getText( "action.debug.run" ) );
    this.mnuDebugRun.setAccelerator(
		KeyStroke.getKeyStroke( KeyEvent.VK_F5, 0 ) );
    mnuDebug.add( this.mnuDebugRun );

    this.mnuDebugStepOver = new JMenuItem(
				Main.getText( "action.debug.step_over" ) );
    this.mnuDebugStepOver.setAccelerator(
		KeyStroke.getKeyStroke( KeyEvent.VK_F6, 0 ) );
    mnuDebug.add( this.mnuDebugStepOver );

    this.mnuDebugStepInto = new JMenuItem(
				Main.getText( "action.debug.step_into" ) );
    this.mnuDebugStepInto.setAccelerator(
		KeyStroke.getKeyStroke( KeyEvent.VK_F7, 0 ) );
    mnuDebug.add( this.mnuDebugStepInto );

    this.mnuDebugStepUp = new JMenuItem(
				Main.getText( "action.debug.step_up" ) );
    this.mnuDebugStepUp.setAccelerator(
		KeyStroke.getKeyStroke( KeyEvent.VK_F8, 0 ) );
    mnuDebug.add( this.mnuDebugStepUp );
    mnuDebug.addSeparator();

    this.mnuDebugBpAdd = new JMenuItem(
				Main.getText( "action.debug.bp.add" ) );
    mnuDebug.add( this.mnuDebugBpAdd );

    this.mnuDebugBpRemoveAll = new JMenuItem(
			Main.getText( "action.debug.bp.remove_all" ) );
    mnuDebug.add( this.mnuDebugBpRemoveAll );


    // Settings menu
    JMenu mnuSettings = null;

    mnuSettings = new JMenu( Main.getText( "menu.settings" ) );

    this.mnuSettingsOpen = new JMenuItem(
				Main.getText( "action.settings.open" ) );
    mnuSettings.add( this.mnuSettingsOpen );

    File settingsDir = Main.getSettingsDir();
    if( settingsDir != null ) {
      mnuSettings.addSeparator();

      this.mnuSettingsSave = new JMenuItem(
				Main.getText( "action.settings.save" ) );
      mnuSettings.add( this.mnuSettingsSave );

      this.mnuSettingsDelete = new JMenuItem(
				Main.getText( "action.settings.delete" ) );
      this.mnuSettingsDelete.setEnabled( settingsDir.exists() );
      mnuSettings.add( this.mnuSettingsDelete );
    } else {
      this.mnuSettingsSave   = null;
      this.mnuSettingsDelete = null;
    }


    // Help menu
    JMenu mnuHelp = new JMenu( Main.getText( "menu.help" ) );

    this.mnuHelpHome = new JMenuItem(
				MessageFormat.format(
					Main.getText( "action.help.home" ),
					Main.APPNAME ) );
    mnuHelp.add( this.mnuHelpHome );
    mnuHelp.addSeparator();

    this.mnuHelpAbout = new JMenuItem(
				MessageFormat.format(
					Main.getText( "action.help.about" ),
					Main.APPNAME ) );
    mnuHelp.add( this.mnuHelpAbout );

    this.mnuHelpLicense = new JMenuItem(
				Main.getText( "action.help.license" ) );
    mnuHelp.add( this.mnuHelpLicense );


    // Menu bar
    JMenuBar mnuBar = new JMenuBar();
    mnuBar.add( mnuFile );
    mnuBar.add( mnuEdit );
    mnuBar.add( mnuPrg );
    mnuBar.add( mnuDebug );
    mnuBar.add( mnuSettings );
    mnuBar.add( mnuHelp );
    setJMenuBar( mnuBar );


    // ToolBar
    JToolBar toolBar = new JToolBar();
    toolBar.setFloatable( false );
    toolBar.setBorderPainted( false );
    toolBar.setOrientation( JToolBar.HORIZONTAL );
    toolBar.setRollover( true );
    toolBar.setBorder( BorderFactory.createEmptyBorder( 5, 5, 5, 5 ) );
    add( toolBar, BorderLayout.NORTH );

    this.btnFileNew = UIUtil.createImageButton(
				this,
				"/images/file/new.png",
				Main.getText( "action.file.new" ) );
    toolBar.add( this.btnFileNew );

    this.btnFileOpen = UIUtil.createImageButton(
				this,
				"/images/file/open.png",
				Main.getText( "action.file.open" ) );
    toolBar.add( this.btnFileOpen );

    this.btnFileSave = UIUtil.createImageButton(
				this,
				"/images/file/save.png",
				Main.getText( "action.file.save" ) );
    toolBar.add( this.btnFileSave );

    this.btnFilePrint = UIUtil.createImageButton(
				this,
				"/images/file/print.png",
				Main.getText( "action.file.print" ) );
    toolBar.add( this.btnFilePrint );
    toolBar.addSeparator();

    this.btnEditUndo = UIUtil.createImageButton(
				this,
				"/images/edit/undo.png",
				Main.getText( "action.edit.undo" ) );
    toolBar.add( this.btnEditUndo );
    toolBar.addSeparator();

    this.btnEditCut = UIUtil.createImageButton(
				this,
				"/images/edit/cut.png",
				Main.getText( "action.edit.cut" ) );
    toolBar.add( this.btnEditCut );

    this.btnEditCopy = UIUtil.createImageButton(
				this,
				"/images/edit/copy.png",
				Main.getText( "action.edit.copy" ) );
    toolBar.add( this.btnEditCopy );

    this.btnEditPaste = UIUtil.createImageButton(
				this,
				"/images/edit/paste.png",
				Main.getText( "action.edit.paste" ) );
    toolBar.add( this.btnEditPaste );
    toolBar.addSeparator();

    this.btnEditFind = UIUtil.createImageButton(
				this,
				"/images/edit/find.png",
				Main.getText( "action.edit.find" ) );
    toolBar.add( this.btnEditFind );
    toolBar.addSeparator();

    this.btnRun = UIUtil.createImageButton(
				this,
				"/images/debug/run.png",
				Main.getText( "action.run" ) );
    this.btnRun.setToolTipText( Main.getText( "action.prg.run" ) );
    toolBar.add( this.btnRun );

    this.btnStop = UIUtil.createImageButton(
				this,
				"/images/debug/stop.png",
				Main.getText( "action.stop" ) );
    this.btnStop.setToolTipText( Main.getText( "action.prg.abort" ) );
    toolBar.add( this.btnStop );

    this.btnDebugStepOver = UIUtil.createImageButton(
				this,
				"/images/debug/step_over.png",
				Main.getText( "action.debug.step_over" ) );
    toolBar.add( this.btnDebugStepOver );

    this.btnDebugStepInto = UIUtil.createImageButton(
				this,
				"/images/debug/step_into.png",
				Main.getText( "action.debug.step_into" ) );
    toolBar.add( this.btnDebugStepInto );

    this.btnDebugStepUp = UIUtil.createImageButton(
				this,
				"/images/debug/step_up.png",
				Main.getText( "action.debug.step_up" ) );
    toolBar.add( this.btnDebugStepUp );
    toolBar.addSeparator();

    String text = Main.getText( "action.audio.on_off" );
    URL urlOn   = getClass().getResource( "/images/audio/speaker_on.png" );
    URL urlOff  = getClass().getResource( "/images/audio/speaker_off.png" );
    if( (urlOn != null) && (urlOff != null) ) {
      this.cbAudioEnabled = new JCheckBox( new ImageIcon( urlOff ) );
      this.cbAudioEnabled.setSelectedIcon( new ImageIcon( urlOn ) );
      this.cbAudioEnabled.setToolTipText( text );
    } else {
      this.cbAudioEnabled = new JCheckBox( text );
    }
    this.cbAudioEnabled.setSelected(
		Main.getBooleanProperty( PROP_AUDIO_ENABLED, true ) );
    toolBar.add( this.cbAudioEnabled );
    toolBar.addSeparator();

    JPanel panelSpeed = new JPanel(
				new FlowLayout( FlowLayout.LEFT, 5, 5 ) );
    panelSpeed.add( new JLabel( Main.getText( "label.speed" ) ) );

    this.spinnerSpeed = new JSpinner(
				new SpinnerListModel(
					new Object[] {
						1, 2, 3, 4, 5, 6, 7, 8, 9,
						MAX_SPEED_ITEM } ) );
    try {
      int speed = Main.getIntProperty(
				BCRuntime.PROP_SPEED,
				BCRuntime.DEFAULT_SPEED );
      if( (speed >= 1) && (speed <= 9) ) {
	this.spinnerSpeed.setValue( speed );
      } else {
	this.spinnerSpeed.setValue( MAX_SPEED_ITEM );
      }
    }
    catch( IllegalArgumentException ex ) {}
    panelSpeed.add( this.spinnerSpeed );

    toolBar.add( panelSpeed );


    // Main window content
    this.layoutMain = new CardLayout();
    this.panelMain  = new JPanel( this.layoutMain );
    add( this.panelMain, BorderLayout.CENTER );

    JPanel panelNormalMode = new JPanel( new BorderLayout( 0, 0 ) );
    this.panelMain.add( panelNormalMode, CARD_NORMAL );
    
    this.layoutLeft = new CardLayout();
    this.panelLeft  = new JPanel( this.layoutLeft );
    panelNormalMode.add( this.panelLeft, BorderLayout.CENTER );

    this.textEditFld = new TextEditFld( this );
    this.textEditFld.setBorder(
		BorderFactory.createTitledBorder(
			Main.getText( "title.area.edit" ) ) );
    this.panelLeft.add( this.textEditFld, CARD_EDIT );

    this.debugFld = new DebugFld( this );
    this.debugFld.setBorder(
		BorderFactory.createTitledBorder(
			Main.getText( "title.area.debug" ) ) );
    this.panelLeft.add( this.debugFld, CARD_DEBUG );

    JPanel panelRight = new JPanel( new BorderLayout( 5, 5 ) );
    panelNormalMode.add( panelRight, BorderLayout.EAST );

    this.smallScreenFld = new ScreenFld( this.bcScreen );
    this.smallScreenFld.setBorder(
			BorderFactory.createLoweredBevelBorder() );
    this.smallScreenFld.setEnabled( false );

    JPanel panelSmallScreen = new JPanel( new BorderLayout( 0, 0 ) );
    panelSmallScreen.add( this.smallScreenFld, BorderLayout.CENTER );
    panelSmallScreen.setBorder(
		BorderFactory.createTitledBorder(
			Main.getText( "title.area.screen" ) ) );
    panelRight.add( panelSmallScreen, BorderLayout.NORTH );

    this.logFld = new LogFld( this );
    this.logFld.setBorder(
		BorderFactory.createTitledBorder(
			Main.getText( "title.area.log" ) ) );
    this.logFld.setPreferredSize( new java.awt.Dimension( 1, 1 ) );
    panelRight.add( this.logFld, BorderLayout.CENTER );

    JPanel panelPresentationMode = new JPanel( new BorderLayout( 0, 0 ) );
    this.panelMain.add( panelPresentationMode, CARD_PRESENTATION );

    this.largeScreenFld = new ScreenFld( this.bcScreen );
    this.largeScreenFld.setBorder(
			BorderFactory.createLoweredBevelBorder() );
    this.largeScreenFld.setEnabled( false );
    panelPresentationMode.add( this.largeScreenFld, BorderLayout.CENTER );

    this.labelStatus = new JLabel( this.defaultStatusText );
    this.labelStatus.setBorder(
		BorderFactory.createEmptyBorder( 0, 5, 5, 5 ) );
    add( this.labelStatus, BorderLayout.SOUTH );


    // Listeners
    this.windowAdapter = new WindowAdapter()
				{
				  @Override
				  public void windowClosing( WindowEvent e )
				  {
				    doQuit();
				  }
				};
    addWindowListener( this.windowAdapter );
    this.mnuFileNew.addActionListener( this );
    this.mnuFileOpen.addActionListener( this );
    this.mnuFileSave.addActionListener( this );
    this.mnuFileSaveAs.addActionListener( this );
    this.mnuFilePrint.addActionListener( this );
    this.mnuFilePrintOptions.addActionListener( this );
    this.mnuExportPrgToWAV.addActionListener( this );
    this.mnuExportDataFileToWAV.addActionListener( this );
    this.mnuConvertDataFileEOR.addActionListener( this );
    this.mnuQuit.addActionListener( this );
    this.mnuEditCopy.addActionListener( this );
    this.mnuEditCut.addActionListener( this );
    this.mnuEditGotoLine.addActionListener( this );
    this.mnuEditPaste.addActionListener( this );
    this.mnuEditFind.addActionListener( this );
    this.mnuEditFindNext.addActionListener( this );
    this.mnuEditFindPrev.addActionListener( this );
    this.mnuEditReplace.addActionListener( this );
    this.mnuEditSelectAll.addActionListener( this );
    this.mnuEditToUpper.addActionListener( this );
    this.mnuEditUndo.addActionListener( this );
    this.mnuPrgCheckSyntax.addActionListener( this );
    this.mnuPrgPrinterOutput.addActionListener( this );
    this.mnuPrgRunNormal.addActionListener( this );
    this.mnuPrgRunLarge.addActionListener( this );
    this.mnuPrgAbort.addActionListener( this );
    this.mnuPrgAbortESC.addActionListener( this );
    this.mnuDebugBpAdd.addActionListener( this );
    this.mnuDebugBpRemoveAll.addActionListener( this );
    this.mnuDebugOpen.addActionListener( this );
    this.mnuDebugRun.addActionListener( this );
    this.mnuDebugStop.addActionListener( this );
    this.mnuDebugStepOver.addActionListener( this );
    this.mnuDebugStepInto.addActionListener( this );
    this.mnuDebugStepUp.addActionListener( this );
    this.mnuSettingsDelete.addActionListener( this );
    this.mnuSettingsOpen.addActionListener( this );
    this.mnuSettingsSave.addActionListener( this );
    this.mnuHelpAbout.addActionListener( this );
    this.mnuHelpHome.addActionListener( this );
    this.mnuHelpLicense.addActionListener( this );
    this.btnFileNew.addActionListener( this );
    this.btnFileOpen.addActionListener( this );
    this.btnFileSave.addActionListener( this );
    this.btnFilePrint.addActionListener( this );
    this.btnEditCopy.addActionListener( this );
    this.btnEditCut.addActionListener( this );
    this.btnEditPaste.addActionListener( this );
    this.btnEditFind.addActionListener( this );
    this.btnEditUndo.addActionListener( this );
    this.btnRun.addActionListener( this );
    this.btnStop.addActionListener( this );
    this.btnDebugStepOver.addActionListener( this );
    this.btnDebugStepInto.addActionListener( this );
    this.btnDebugStepUp.addActionListener( this );
    this.cbAudioEnabled.addActionListener( this );
    this.spinnerSpeed.addChangeListener( this );
    if( this.clipboard != null ) {
      this.clipboard.addFlavorListener( this );
    }


    // Etc
    this.statusTimer = new Timer( 5000, this );
    this.statusTimer.setRepeats( false );
    updPrgActionsEnabled();
    findNextPrevStateChanged( false );
    fireFileStatusChanged( 0xFF );
    boolean windowBoundsDone = UIUtil.restoreWindowBounds(
					this,
					Main.getProperties(),
					PROP_PREFIX );
    if( !windowBoundsDone ) {
      setLocationByPlatform( true );
    }
    EventQueue.invokeLater(
		()->initApp(
			!windowBoundsDone,
			prgFileName,
			runFlag ) );
  }


  public void bcThreadFinished( String msg )
  {
    synchronized( this.bcLock ) {
      this.bcRuntime = null;
    }
    EventQueue.invokeLater( ()->bcThreadFinishedInternal( msg ) );
  }


  public void findNextPrevStateChanged( boolean state )
  {
    if( this.runMode.equals( RunMode.PRESENTATION ) ) {
      state = false;
    }
    this.mnuEditFindNext.setEnabled( state );
    this.mnuEditFindPrev.setEnabled( state );
  }


  public void fireAddPrintDoc( StringBuffer content )
  {
    EventQueue.invokeLater( ()->addPrintDoc( content ) );
  }


  public void fireAppendToLog( String text )
  {
    EventQueue.invokeLater( ()->this.logFld.appendText( text ) );
  }


  public void fireBreakpointsChanged()
  {
    EventQueue.invokeLater( ()->breakpointsChanged() );
  }


  public void fireClearLog()
  {
    EventQueue.invokeLater( ()->this.logFld.clear() );
  }


  public void fireCurPrintDocChanged()
  {
    EventQueue.invokeLater( ()->curPrintDocChanged() );
  }


  public void fireDebugStatusChanged( boolean prgStopped )
  {
    EventQueue.invokeLater( ()->debugStatusChanged( prgStopped ) );
  }


  public void fireFileStatusChanged( int mask )
  {
    EventQueue.invokeLater( ()->fileStatusChanged( mask ) );
  }


  public void fireGotoEditorLine( int lineNum )
  {
    this.textEditFld.fireGotoLine( lineNum );
  }


  public void fireSetDefaultStatusText()
  {
    EventQueue.invokeLater( ()->setDefaultStatusText() );
  }


  public void fireSetStatusText( String text )
  {
    EventQueue.invokeLater( ()->setStatusText( text ) );
  }


  public void fireShowDefaultStatusText()
  {
    EventQueue.invokeLater( ()->setDefaultStatusText() );
  }


  public void fireShowStatusText( String text )
  {
    EventQueue.invokeLater( ()->showStatusText( text ) );
  }


  public void fireShowStatusTextByKey( String textKey )
  {
    fireShowStatusText( Main.getText( textKey ) );
  }


  public BCScreen getBCScreen()
  {
    return this.bcScreen;
  }


  public TextFinder getTextFinder()
  {
    return this.textFinder;
  }


  public void setBCBreakKeyEnabled( boolean state )
  {
    this.bcBreakKeyEnabled = state;
    this.bcScreen.setBCBreakKeyEnabled( state );
    EventQueue.invokeLater( ()->updPrgActionsEnabled() );
  }


  public void setRecentDataFile( File file )
  {
    this.recentDataFile = file;
  }


  public void setStatusText( String text )
  {
    this.labelStatus.setText( text != null ? text : this.defaultStatusText );
  }


  public void settingsChanged()
  {
    this.bcScreen.settingsChanged();
    BCRuntime bcRuntime = this.bcRuntime;
    if( bcRuntime != null ) {
      bcRuntime.settingsChanged();
    }
  }


  public void showStatusText( String text )
  {
    if( text != null ) {
      if( !text.isEmpty() ) {
	this.labelStatus.setText( text );
	this.statusTimer.restart();
      }
    }
  }


  public void undoStateChanged( boolean state )
  {
    if( !this.runMode.equals( RunMode.NORMAL ) ) {
      state = false;
    }
    this.mnuEditUndo.setEnabled( state );
    this.btnEditUndo.setEnabled( state );
  }


	/* --- ActionListener --- */

  @Override
  public void actionPerformed( ActionEvent e )
  {
    Object src = e.getSource();
    if( src == this.statusTimer ) {
      this.statusTimer.stop();
      setDefaultStatusText();
    } else if( (src == this.mnuFileNew) || (src == this.btnFileNew) ) {
      this.textEditFld.doNew();
    } else if( (src == this.mnuFileOpen) || (src == this.btnFileOpen) ) {
      this.textEditFld.doOpen();
    } else if( (src == this.mnuFileSave) || (src == this.btnFileSave) ) {
      this.textEditFld.doSave( false );
    } else if( src == this.mnuFileSaveAs ) {
      this.textEditFld.doSave( true );
    } else if( (src == this.mnuFilePrint) || (src == this.btnFilePrint) ) {
      doFilePrint();
    } else if( src == this.mnuFilePrintOptions ) {
      PrintOptionsDlg.open( this, true );
    } else if( src == this.mnuExportPrgToWAV ) {
      doExportPrgToWAV();
    } else if( src == this.mnuExportDataFileToWAV ) {
      doExportDataFileToWAV();
    } else if( src == this.mnuConvertDataFileEOR ) {
      doConvertDataFileEOR();
    } else if( src == this.mnuQuit ) {
      doQuit();
    } else if( (src == this.mnuEditUndo) || (src == this.btnEditUndo) ) {
      this.textEditFld.doUndo();
    } else if( (src == this.mnuEditCut) || (src == this.btnEditCut) ) {
      doEditCut();
    } else if( (src == this.mnuEditCopy) || (src == this.btnEditCopy) ) {
      doEditCopy();
    } else if( (src == this.mnuEditPaste) || (src == this.btnEditPaste) ) {
      doEditPaste();
    } else if( (src == this.mnuEditFind) || (src == this.btnEditFind) ) {
      doEditFind();
    } else if( src == this.mnuEditFindNext ) {
      doEditFindNext();
    } else if( src == this.mnuEditFindPrev ) {
      doEditFindPrev();
    } else if( src == this.mnuEditReplace ) {
      doEditReplace();
    } else if( src == this.mnuEditToUpper ) {
      if( this.runMode.equals( RunMode.NORMAL ) ) {
	this.textEditFld.doToUpper();
      }
    } else if( src == this.mnuEditGotoLine ) {
      if( this.runMode.equals( RunMode.NORMAL ) ) {
	this.textEditFld.doGotoLine();
      }
    } else if( src == this.mnuEditSelectAll ) {
      doEditSelectAll();
    } else if( src == this.mnuPrgCheckSyntax ) {
      doPrgCheckSyntax();
    } else if( src == this.mnuPrgRunNormal ) {
      doPrgRun( false );
    } else if( src == this.mnuPrgRunLarge ) {
      doPrgRun( true );
    } else if( src == this.mnuPrgAbort ) {
      doPrgStop();
    } else if( src == this.mnuPrgAbortESC ) {
      if( this.mnuPrgAbortESC.isVisible() ) {
	doPrgStop();
      }
    } else if( src == this.mnuDebugBpAdd ) {
      this.debugFld.doAddBreakpoint();
    } else if( src == this.mnuDebugBpRemoveAll ) {
      this.debugFld.removeAllBreakpoints();
    } else if( src == this.mnuDebugOpen ) {
      doDebugOpen();
    } else if( src == this.mnuPrgPrinterOutput ) {
      doPrgPrinterOutput();
    } else if( src == this.mnuDebugRun ) {
      doDebugAction( DebugFld.DebugAction.RUN );
    } else if( (src == this.mnuDebugStepInto)
	       || src == this.btnDebugStepInto )
    {
      doDebugAction( DebugFld.DebugAction.STEP_INTO );
    } else if( (src == this.mnuDebugStepOver)
	       || src == this.btnDebugStepOver )
    {
      doDebugAction( DebugFld.DebugAction.STEP_OVER );
    } else if( (src == this.mnuDebugStepUp)
	       || src == this.btnDebugStepUp )
    {
      doDebugAction( DebugFld.DebugAction.STEP_UP );
    } else if( src == this.mnuDebugStop ) {
      doDebugAction( DebugFld.DebugAction.STOP );
    } else if( src == this.mnuSettingsOpen ) {
      SettingsDlg.open( this );
    } else if( src == this.mnuSettingsSave ) {
      doSettingsSave();
    } else if( src == this.mnuSettingsDelete ) {
      doSettingsDelete();
    } else if( src == this.mnuHelpAbout ) {
      doHelpAbout();
    } else if( src == this.mnuHelpHome ) {
      doHelp().showHomePage();
    } else if( src == this.mnuHelpLicense ) {
      doHelp().showLicensePage();
    } else if( src == this.btnRun ) {
      if( this.runMode.equals( RunMode.DEBUG ) ) {
	doDebugAction( DebugFld.DebugAction.RUN );
      } else {
	doPrgRun( false );
      }
    } else if( src == this.btnStop ) {
      if( this.runMode.equals( RunMode.DEBUG ) ) {
	doDebugAction( DebugFld.DebugAction.STOP );
      } else {
	doPrgStop();
      }
    } else if( src == this.cbAudioEnabled ) {
      doAudioOnOff();
    }
  }


	/* --- CaretListener --- */

  @Override
  public void caretUpdate( CaretEvent e )
  {
    Object src = e.getSource();
    if( src != null ) {
      if( src instanceof JTextComponent ) {
	setCurTextComponent( (JTextComponent) src );
      }
    }
  }


	/* --- ChangeListener --- */

  @Override
  public void stateChanged( ChangeEvent e )
  {
    if( e.getSource() == this.spinnerSpeed ) {
      Object item = this.spinnerSpeed.getValue();
      if( item != null ) {
	int value = 0;
	if( item instanceof Integer ) {
	  value = ((Integer) item).intValue();
	}
	BCRuntime bcRuntime = this.bcRuntime;
	if( bcRuntime != null ) {
	  bcRuntime.setSpeed( value );
	}
	Main.setProperty( BCRuntime.PROP_SPEED, value );
      }
    }
  }


	/* --- FlavorListener --- */

  @Override
  public void flavorsChanged( FlavorEvent e )
  {
    if( e.getSource() == this.clipboard )
      updPasteActionsEnabled();
  }


	/* --- Actions --- */

  private void doAudioOnOff()
  {
    boolean   enabled   = this.cbAudioEnabled.isSelected();
    BCRuntime bcRuntime = this.bcRuntime;
    if( bcRuntime != null ) {
      bcRuntime.setAudioEnabled( enabled );
    }
    fireShowStatusTextByKey(
		enabled ? "status.audio_on" : "status.audio_off" );
  }


  private void doConvertDataFileEOR()
  {
    File inFile = UIUtil.showOpenFileDlg(
			this,
			Main.getText( "title.open_data_file" ),
			this.recentDataFile != null ?
				this.recentDataFile.getParentFile()
				: null,
			null,
			null );
    if( inFile != null ) {
      JPanel panelEOR = new JPanel();
      panelEOR.setLayout( new BoxLayout( panelEOR, BoxLayout.Y_AXIS ) );

      ButtonGroup grpEOR = new ButtonGroup();

      JRadioButton rbCR = new JRadioButton(
			Main.getText( "option.eor.cr" ),
			this.convertToEOR.equals( BCRuntime.EOR.CR ) );
      grpEOR.add( rbCR );
      rbCR.setAlignmentX( 0F );
      panelEOR.add( rbCR, true );

      JRadioButton rbCRNL = new JRadioButton(
			Main.getText( "option.eor.crnl" ),
			this.convertToEOR.equals( BCRuntime.EOR.CRNL ) );
      grpEOR.add( rbCRNL );
      rbCRNL.setAlignmentX( 0F );
      panelEOR.add( rbCRNL );

      JRadioButton rbNL = new JRadioButton(
			Main.getText( "option.eor.nl" ),
			this.convertToEOR.equals( BCRuntime.EOR.NL ) );
      grpEOR.add( rbNL );
      rbNL.setAlignmentX( 0F );
      panelEOR.add( rbNL );

      if( JOptionPane.showConfirmDialog(
		this,
		panelEOR,
		Main.getText( "title.settings.eor_bytes" ),
		JOptionPane.OK_CANCEL_OPTION,
		JOptionPane.QUESTION_MESSAGE ) == JOptionPane.OK_OPTION )
      {
	byte[] eor = null;
	if( rbCR.isSelected() ) {
	  eor               = new byte[] { (byte) 0x0D };
	  this.convertToEOR = BCRuntime.EOR.CR;
	} else if( rbNL.isSelected() ) {
	  eor               = new byte[] { (byte) 0x0A };
	  this.convertToEOR = BCRuntime.EOR.NL;
	} else {
	  eor               = new byte[] { (byte) 0x0D, (byte) 0x0A };
	  this.convertToEOR = BCRuntime.EOR.CRNL;
	}
	String fileName = inFile.getName();
	if( fileName != null ) {
	  String baseName = fileName;
	  String fileExt  = "";
	  int    idx      = fileName.lastIndexOf( '.' );
	  if( idx >= 0 ) {
	    baseName = fileName.substring( 0, idx );
	    fileExt  = fileName.substring( idx );
	  }
	  StringBuilder buf = new StringBuilder( fileName.length() * 6 );
	  buf.append( baseName );
	  for( byte b : eor ) {
	    buf.append( String.format( "_%02X", (int) b & 0xFF ) );
	  }
	  buf.append( fileExt );
	  fileName = buf.toString();
	}
	File outFile = UIUtil.showOpenFileDlg(
				this,
				Main.getText( "title.save_data_file" ),
				inFile.getParentFile(),
				fileName,
				null );
	if( outFile != null ) {
	  OutputStream out = null;
	  try( InputStream in = new BufferedInputStream(
					new FileInputStream( inFile ) ) )
	  {
	    try {
	      out = new BufferedOutputStream(
					new FileOutputStream( outFile ) );
	      boolean cr = false;
	      int     b  = in.read();
	      while( b >= 0 ) {
		if( b == 0x0D ) {
		  if( cr ) {
		    out.write( eor );
		  }
		  cr = true;
		} else if( b == 0x0A ) {
		  out.write( eor );
		  cr = false;
		} else {
		  if( cr ) {
		    out.write( eor );
		  }
		  out.write( b );
		  cr = false;
		}
		b = in.read();
	      }
	      if( cr ) {
		out.write( eor );
	      }
	      out.close();
	      out = null;
	      showStatusText( Main.getText( "status.file_exported" ) );
	      this.recentDataFile = inFile;
	    }
	    finally {
	      IOUtil.closeSilently( out );
	    }
	  }
	  catch( IOException ex ) {
	    UIUtil.showErrorMsg( this, null, ex );
	  }
	}
      }
    }
  }


  private void doDebugAction( DebugFld.DebugAction debugAction )
  {
    if( this.runMode.equals( RunMode.DEBUG ) )
      this.debugFld.setDebugAction( debugAction );
  }


  private void doDebugOpen()
  {
    if( !this.runMode.equals( RunMode.DEBUG ) ) {
      BCRuntime bcRuntime = null;
      String    prgText   = null;
      synchronized( this.bcLock ) {
	bcRuntime = this.bcRuntime;
	if( bcRuntime != null ) {
	  prgText = bcRuntime.getProgramText();
	} else {
	  this.logFld.clear();
	  prgText = getRunnablePrgText();
	  if( prgText != null ) {
	    bcRuntime = new BCRuntime(
				this,
				prgText,
				this.textEditFld.getFile(),
				this.cbAudioEnabled.isSelected() );
	  } else {
	    prgText = null;
	  }
	}
      }
      if( bcRuntime != null ) {
	this.runMode = RunMode.DEBUG;
	this.bcRuntime = bcRuntime;
	this.bcScreen.setScreenFld( this.smallScreenFld );
	this.layoutLeft.show( this.panelLeft, CARD_DEBUG );
	this.debugFld.setData( prgText, bcRuntime );
	this.smallScreenFld.setEnabled( true );
	this.smallScreenFld.requestFocus();
	this.btnRun.setToolTipText( Main.getText( "action.debug.run" ) );
	this.btnStop.setToolTipText( Main.getText( "action.debug.stop" ) );
	updFileActionsEnabled();
	updPrgActionsEnabled();
      }
    }
  }


  private void doEditCopy()
  {
    JTextComponent tc = this.curTextComp;
    if( tc != null ) {
      tc.copy();
      tc.requestFocus();
    }
  }


  private void doEditCut()
  {
    JTextComponent tc = this.curTextComp;
    if( tc != null ) {
      if( tc.isEditable() ) {
	tc.cut();
	tc.requestFocus();
      }
    }
  }


  private void doEditFind()
  {
    if( this.runMode.equals( RunMode.DEBUG ) ) {
      this.textFinder.openFindDlg( this.debugFld.getTextArea() );
      if( this.textFinder.hasSearchText() ) {
	findNextPrevStateChanged( true );
      }
    } else {
      this.textEditFld.doFind();
    }
  }


  private void doEditFindNext()
  {
    if( this.runMode.equals( RunMode.DEBUG ) ) {
      this.textFinder.findNext( this.debugFld.getTextArea() );
    } else {
      this.textEditFld.doFindNext();
    }
  }


  private void doEditFindPrev()
  {
    if( this.runMode.equals( RunMode.DEBUG ) ) {
      this.textFinder.findPrev( this.debugFld.getTextArea() );
    } else {
      this.textEditFld.doFindPrev();
    }
  }


  private void doEditPaste()
  {
    JTextComponent tc = this.curTextComp;
    if( tc != null ) {
      if( tc.isEditable() ) {
	tc.paste();
	tc.requestFocus();
      }
    }
  }


  private void doEditReplace()
  {
    JTextComponent tc = this.curTextComp;
    String         text = this.textFinder.getReplaceText();
    if( (tc != null) && (text != null) ) {
      if( tc.isEditable() && !text.isEmpty() ) {
	tc.replaceSelection( text );
	tc.requestFocus();
      }
    }
  }


  private void doEditSelectAll()
  {
    JTextComponent tc = this.curTextComp;
    if( tc != null ) {
      tc.selectAll();
      tc.requestFocus();
    }
  }


  private void doExportDataFileToWAV()
  {
    File dataFile = UIUtil.showOpenFileDlg(
			this,
			Main.getText( "title.open_data_file" ),
			this.recentDataFile != null ?
				this.recentDataFile.getParentFile()
				: null,
			null,
			null );
    if( dataFile != null ) {
      try {
	InputStream in = null;
	try {
	  in = new BufferedInputStream( new FileInputStream( dataFile ) );

	  File wavFile = showSaveWavFileDlg( dataFile );
	  if( wavFile != null ) {
	    if( wavFile.equals( dataFile ) ) {
	      throw new IOException(
			"error.identical_in_and_out_file" );
	    }
	    BCAudioCreator audioCreator = new BCAudioCreator( 24000 );
	    audioCreator.writeDataWAV( wavFile, in );
	    this.recentWavFile = wavFile;
	    showStatusText( Main.getText( "status.file_exported" ) );
	  }
	}
	finally {
	  IOUtil.closeSilently( in );
	}
      }
      catch( IOException ex ) {
	UIUtil.showErrorMsg( this, null, ex );
      }
    }
  }


  private void doExportPrgToWAV()
  {
    String prgText = getRunnablePrgText();
    if( prgText != null ) {
      ProgramParser prgParser = doPrgCheckSyntax();
      if( prgParser != null ) {
	boolean status = true;
	String  msgKey = null;
	if( prgParser.hasErrors() ) {
	  msgKey = "confirm.export_with_errors";
	}
	else if( prgParser.hasWarnings() ) {
	  msgKey = "confirm.export_with_warnings";
	}
	else if( !prgParser.getWarnNonAsciiEnabled()
			|| !prgParser.getWarnSpecViolatedEnabled() )
	{
	  msgKey = "confirm.export_with_warnings_disabled";
	}
	else if( prgParser.getIgnoreCase() ) {
	  msgKey = "confirm.export_with_ignored_case";
	}
	if( msgKey != null ) {
	  if( JOptionPane.showConfirmDialog(
		this,
		Main.getText( msgKey ),
		Main.getText( "title.confirm" ),
		JOptionPane.YES_NO_OPTION,
		JOptionPane.WARNING_MESSAGE ) != JOptionPane.YES_OPTION )
	  {
	    status = false;
	  }
	}
	if( status ) {
	  File file = showSaveWavFileDlg( this.textEditFld.getFile() );
	  if( file != null ) {
	    try {
	      BCAudioCreator audioCreator = new BCAudioCreator( 24000 );
	      audioCreator.writeProgramWAV( file, prgText );
	      this.recentWavFile = file;
	      showStatusText( Main.getText( "status.file_exported" ) );
	    }
	    catch( IOException ex ) {
	      UIUtil.showErrorMsg( this, null, ex );
	    }
	  }
	}
      }
    }
  }


  private void doFilePrint()
  {
    if( getRunnablePrgText() != null )
      this.textEditFld.doPrint();
  }


  private HelpFrm doHelp()
  {
    if( this.helpFrm == null ) {
      this.helpFrm = new HelpFrm();
    }
    this.helpFrm.setVisible( true );
    this.helpFrm.toFront();
    return this.helpFrm;
  }


  private void doHelpAbout()
  {
    JLabel label = new JLabel( Main.APPNAME );
    label.setFont( new Font( Font.SANS_SERIF, Font.BOLD, 18 ) );

    String textVersion = Main.getText( "text.version" );

    Object[] msg = new Object[] {
			label,
			textVersion + "\u0020" + Main.VERSION + "\n\n"
				+ Main.COPYRIGHT
				+ "\n\n"
				+ Main.getText( "label.license" )
				+ "\nGNU General Public License (GNU-GPL) "
				+ textVersion + " 3\n\n"
				+ Main.getText(
					"about.warranty_liability" ) };

    String title = Main.getText( "title.about" );

    Image iconImage = UIUtil.getLargestIconImage();
    if( iconImage != null ) {
      JOptionPane.showMessageDialog(
		this,
		msg,
		title,
		JOptionPane.INFORMATION_MESSAGE,
		new ImageIcon( iconImage ) );
    } else {
      JOptionPane.showMessageDialog(
		this,
		msg,
		title,
		JOptionPane.INFORMATION_MESSAGE );
    }
  }


  private ProgramParser doPrgCheckSyntax()
  {
    ProgramParser prgParser = null;
    if( this.bcRuntime == null ) {
      String prgText = getRunnablePrgText();
      if( prgText != null ) {
	this.logFld.clear();
	try {
	  prgParser = new ProgramParser( this, null );
	  prgParser.parse( prgText );
	  String textKey = "bc.syntax_check_successfully";
	  if( prgParser.hasErrors() ) {
	    textKey = "bc.syntax_check_failed";
	  }
	  else if( prgParser.hasWarnings() ) {
	    textKey = "bc.syntax_check_with_warnings";
	  }
	  else if( !prgParser.getWarnNonAsciiEnabled()
			|| !prgParser.getWarnSpecViolatedEnabled() )
	  {
	    textKey = "bc.syntax_check_warnings_disabled";
	  }
	  else if( prgParser.getIgnoreCase() ) {
	    textKey = "bc.syntax_check_ignored_case";
	  }
	  fireAppendToLog( Main.getText( textKey ) + "\n" );
	}
	catch( InterruptedException ex ) {}
      }
    }
    return prgParser;
  }


  private void doPrgPrinterOutput()
  {
    if( this.printerOutputFrm == null ) {
      this.printerOutputFrm = new PrinterOutputFrm( this.printDocs );
    }
    this.printerOutputFrm.setVisible( true );
    this.printerOutputFrm.toFront();
  }


  private void doPrgRun( boolean presentationMode )
  {
    if( this.bcRuntime == null ) {
      String prgText = getRunnablePrgText();
      if( prgText != null ) {
	this.logFld.clear();
	if( presentationMode ) {
	  this.largeScreenFld.setEnabled( true );
	  this.largeScreenFld.requestFocus();
	  this.bcScreen.setScreenFld( this.largeScreenFld );
	  this.layoutMain.show( this.panelMain, CARD_PRESENTATION );
	  this.runMode = RunMode.PRESENTATION;
	} else {
	  this.smallScreenFld.setEnabled( true );
	  this.smallScreenFld.requestFocus();
	  this.bcScreen.setScreenFld( this.smallScreenFld );
	  this.layoutMain.show( this.panelMain, CARD_NORMAL );
	  this.runMode = RunMode.NORMAL;
	}
	synchronized( this.bcLock ) {
	  this.bcRuntime = new BCRuntime(
				this,
				prgText,
				this.textEditFld.getFile(),
				this.cbAudioEnabled.isSelected() );
	  this.bcRuntime.start();
	}
      }
    }
    updFileActionsEnabled();
    updPrgActionsEnabled();
  }


  private void doPrgStop()
  {
    BCRuntime bcRuntime = this.bcRuntime;
    if( bcRuntime != null ) {
      bcRuntime.fireAbort();
    }
  }


  public void doQuit()
  {
    if( this.textEditFld.confirmTextSaved() ) {
      BCRuntime bcRuntime = this.bcRuntime;
      if( bcRuntime != null ) {
	try {
	  bcRuntime.fireAbort();
	  bcRuntime.join( 1000 );
	}
	catch( Exception ex ) {}
      }
      if( this.helpFrm != null ) {
	this.helpFrm.doClose();
      }
      if( this.printerOutputFrm != null ) {
	this.printerOutputFrm.doClose();
      }
      removeWindowListener( this.windowAdapter );
      this.mnuFileNew.removeActionListener( this );
      this.mnuFileOpen.removeActionListener( this );
      this.mnuFileSave.removeActionListener( this );
      this.mnuFileSaveAs.removeActionListener( this );
      this.mnuFilePrint.removeActionListener( this );
      this.mnuFilePrintOptions.removeActionListener( this );
      this.mnuExportPrgToWAV.removeActionListener( this );
      this.mnuExportDataFileToWAV.removeActionListener( this );
      this.mnuConvertDataFileEOR.removeActionListener( this );
      this.mnuQuit.removeActionListener( this );
      this.mnuEditCopy.removeActionListener( this );
      this.mnuEditCut.removeActionListener( this );
      this.mnuEditGotoLine.removeActionListener( this );
      this.mnuEditPaste.removeActionListener( this );
      this.mnuEditFind.removeActionListener( this );
      this.mnuEditFindNext.removeActionListener( this );
      this.mnuEditFindPrev.removeActionListener( this );
      this.mnuEditReplace.removeActionListener( this );
      this.mnuEditSelectAll.removeActionListener( this );
      this.mnuEditToUpper.removeActionListener( this );
      this.mnuEditUndo.removeActionListener( this );
      this.mnuPrgCheckSyntax.removeActionListener( this );
      this.mnuPrgPrinterOutput.removeActionListener( this );
      this.mnuPrgRunNormal.removeActionListener( this );
      this.mnuPrgRunLarge.removeActionListener( this );
      this.mnuPrgAbort.removeActionListener( this );
      this.mnuPrgAbortESC.removeActionListener( this );
      this.mnuDebugBpAdd.removeActionListener( this );
      this.mnuDebugBpRemoveAll.removeActionListener( this );
      this.mnuDebugOpen.removeActionListener( this );
      this.mnuDebugRun.removeActionListener( this );
      this.mnuDebugStop.removeActionListener( this );
      this.mnuDebugStepOver.removeActionListener( this );
      this.mnuDebugStepInto.removeActionListener( this );
      this.mnuDebugStepUp.removeActionListener( this );
      this.mnuSettingsDelete.removeActionListener( this );
      this.mnuSettingsOpen.removeActionListener( this );
      this.mnuSettingsSave.removeActionListener( this );
      this.mnuHelpAbout.removeActionListener( this );
      this.mnuHelpHome.removeActionListener( this );
      this.mnuHelpLicense.removeActionListener( this );
      this.btnFileNew.removeActionListener( this );
      this.btnFileOpen.removeActionListener( this );
      this.btnFileSave.removeActionListener( this );
      this.btnFilePrint.removeActionListener( this );
      this.btnEditCopy.removeActionListener( this );
      this.btnEditCut.removeActionListener( this );
      this.btnEditPaste.removeActionListener( this );
      this.btnEditFind.removeActionListener( this );
      this.btnEditUndo.removeActionListener( this );
      this.btnRun.removeActionListener( this );
      this.btnStop.removeActionListener( this );
      this.btnDebugStepOver.removeActionListener( this );
      this.btnDebugStepInto.removeActionListener( this );
      this.btnDebugStepUp.removeActionListener( this );
      this.cbAudioEnabled.removeActionListener( this );
      this.spinnerSpeed.removeChangeListener( this );
      if( this.clipboard != null ) {
	this.clipboard.removeFlavorListener( this );
      }
      setVisible( false );
      dispose();
      System.exit( 0 );
    }
  }


  private void doSettingsDelete()
  {
    File settingsDir = Main.getSettingsDir();
    if( settingsDir != null ) {
      if( settingsDir.exists() ) {
	if( JOptionPane.showConfirmDialog(
		this,
		Main.getText( "confirm.delete_settings" ),
		Main.getText( "title.confirm" ),
		JOptionPane.YES_NO_OPTION,
		JOptionPane.WARNING_MESSAGE ) == JOptionPane.YES_OPTION )
	{
	  File settingsFile = Main.getSettingsFile();
	  if( settingsFile != null ) {
	    settingsFile.delete();
	  }
	  File prgPathFile = Main.getProgramPathFile();
	  if( prgPathFile != null ) {
	    prgPathFile.delete();
	  }
	  if( settingsDir.delete() ) {
	    if( this.mnuSettingsDelete != null ) {
	      this.mnuSettingsDelete.setEnabled( false );
	    }
	    fireShowStatusTextByKey( "status.settings_deleted" );
	  } else {
	    fireShowStatusTextByKey( "error.settings_not_deleted" );
	  }
	}
      }
    }
  }


  private void doSettingsSave()
  {
    File settingsFile = Main.getSettingsFile();
    if( settingsFile != null ) {
      Main.setProperty( PROP_VERSION, Main.VERSION );
      Main.setProperty(
		PROP_AUDIO_ENABLED,
		String.valueOf( this.cbAudioEnabled.isSelected() ) );

      Properties props = Main.getProperties();
      UIUtil.getWindowBounds( this, props, PROP_PREFIX );
      if( this.helpFrm != null ) {
	UIUtil.getWindowBounds(
			this.helpFrm,
			props,
			HelpFrm.PROP_PREFIX );
      }
      if( this.printerOutputFrm != null ) {
	UIUtil.getWindowBounds(
			this.printerOutputFrm,
			props,
			PrinterOutputFrm.PROP_PREFIX );
      }
      Main.createSettingsDirIfNotExists();
      boolean saved = false;
      try( OutputStream out = new BufferedOutputStream(
				new FileOutputStream( settingsFile ) ) )
      {
	props.storeToXML( out, Main.APPNAME + " Settings" );
	saved = true;
      }
      catch( Exception ex ) {
	saved = false;
	UIUtil.showErrorMsg( this, null, ex );
      }
      if( saved ) {
	if( this.mnuSettingsDelete != null ) {
	  this.mnuSettingsDelete.setEnabled( true );
	}
	showStatusText(
		MessageFormat.format(
			Main.getText( "status.settings_saved" ),
			settingsFile.getPath() ) );
      }
    }
  }


	/* --- Private Methods --- */

  public void addPrintDoc( StringBuffer contentBuf )
  {
    if( contentBuf != null ) {

      // convert old document from StringBuffer to String
      Object oldDoc = this.printDocs.get( this.curPrintDocNum );
      if( oldDoc != null ) {
	this.printDocs.put( this.curPrintDocNum, oldDoc.toString() );
      }

      // add new document
      this.curPrintDocNum++;
      this.printDocs.put( this.curPrintDocNum, contentBuf );
      if( this.printerOutputFrm != null ) {
	this.printerOutputFrm.printDocAdded( this.curPrintDocNum );
      }
    }
  }


  private void bcThreadFinishedInternal( String msg )
  {
    this.logFld.appendText( msg );
    this.logFld.appendText( "\n" );
    if( !this.runMode.equals( RunMode.NORMAL ) ) {
      this.runMode = RunMode.NORMAL;
      this.layoutLeft.show( this.panelLeft, CARD_EDIT );
      this.debugFld.setData( null, null );
      debugStatusChanged( false );
    }
    this.layoutMain.show( this.panelMain, CARD_NORMAL );
    this.bcScreen.setScreenFld( null );
    this.btnRun.setToolTipText( Main.getText( "action.prg.run" ) );
    this.btnStop.setToolTipText( Main.getText( "action.prg.abort" ) );
    updPrgActionsEnabled();
  }


  private void breakpointsChanged()
  {
    boolean debugMode = this.runMode.equals( RunMode.DEBUG );
    this.mnuDebugBpAdd.setEnabled( debugMode );
    this.mnuDebugBpRemoveAll.setEnabled(
			debugMode && this.debugFld.hasBreakpoints() );
  }


  private void curPrintDocChanged()
  {
    if( this.printerOutputFrm != null )
      this.printerOutputFrm.printDocChanged( this.curPrintDocNum );
  }


  private void debugStatusChanged( boolean prgStopped )
  {
    boolean debugMode = this.runMode.equals( RunMode.DEBUG );
    this.mnuDebugOpen.setEnabled( !debugMode );
    this.mnuDebugRun.setEnabled( debugMode && prgStopped );
    this.mnuDebugStop.setEnabled( debugMode && !prgStopped );
    this.mnuDebugStepInto.setEnabled( debugMode && prgStopped );
    this.mnuDebugStepOver.setEnabled( debugMode && prgStopped );
    this.mnuDebugStepUp.setEnabled( debugMode && prgStopped );
    if( debugMode ) {
      this.btnRun.setEnabled( prgStopped );
      this.btnStop.setEnabled( !prgStopped );
    }
    this.btnDebugStepInto.setEnabled( debugMode && prgStopped );
    this.btnDebugStepOver.setEnabled( debugMode && prgStopped );
    this.btnDebugStepUp.setEnabled( debugMode && prgStopped );

    breakpointsChanged();
    updFileActionsEnabled();
  }


  private void fileStatusChanged( int mask )
  {
    boolean normalMode = this.runMode.equals( RunMode.NORMAL );
    if( (mask & STATUS_OTHER_FILE) != 0 ) {
      this.debugFld.removeAllBreakpoints();
    }
    if( (mask & STATUS_SAVE_ACTION) != 0 ) {
      this.mnuFileSave.setEnabled(
		normalMode && this.textEditFld.hasTextChanged() );
      this.btnFileSave.setEnabled( this.mnuFileSave.isEnabled() );
    }
    if( (mask & STATUS_TITLE) != 0 ) {
      StringBuilder buf = new StringBuilder( 128 );
      buf.append( Main.APPNAME );
      String fileName = this.textEditFld.getFileName();
      if( fileName != null ) {
	buf.append( ": " );
	buf.append( fileName );
      } else {
	buf.append( ": " );
	buf.append( Main.getText( "title.new_program" ) );
      }
      setTitle( buf.toString() );
    }
  }


  private String getRunnablePrgText()
  {
    String prgText = this.textEditFld.getText();
    if( prgText == null ) {
      prgText = "";
    }
    if( ProgramParser.skipBlanks(
		new StringCharacterIterator( prgText ) )
					== CharacterIterator.DONE )
    {
      UIUtil.fireShowErrorMsg( this, Main.getText( "bc.no_program" ) );
      prgText = null;
    }
    return prgText;
  }


  private void initApp(
		boolean withWindowBounds,
		String  prgFileName,
		boolean runFlag )
  {
    this.layoutMain.show( this.panelMain, CARD_NORMAL );
    this.layoutLeft.show( this.panelLeft, CARD_EDIT );
    this.bcScreen.updScreenSize();
    this.smallScreenFld.updPrefSize();
    this.largeScreenFld.updPrefSize();
    if( withWindowBounds ) {
      pack();
    }
    EventQueue.invokeLater(
		()->{
			this.textEditFld.resetPrefSize();
			this.logFld.resetPrefSize();
		} );
    debugStatusChanged( false );
    JTextArea textArea = this.textEditFld.getTextArea();
    textArea.requestFocus();
    setCurTextComponent( textArea );
    if( prgFileName != null ) {
      EventQueue.invokeLater(
		()->openPrgFile( new File( prgFileName ), runFlag ) );
    }
  }


  private void openPrgFile( File file, boolean runFlag )
  {
    if( this.textEditFld.openFile( file ) ) {
      if( runFlag ) {
	doPrgRun( true );
      }
    }
  }


  private void setCurTextComponent( JTextComponent textComp )
  {
    this.curTextComp = textComp;
    boolean state    = UIUtil.hasSelection( textComp );
    this.mnuEditCopy.setEnabled( state );
    this.btnEditCopy.setEnabled( state );
    this.mnuEditToUpper.setEnabled( state );
    state &= textComp.isEditable();
    this.mnuEditCut.setEnabled( state );
    this.btnEditCut.setEnabled( state );
    this.mnuEditReplace.setEnabled(
		state && this.textFinder.hasReplaceText() );
    this.mnuEditToUpper.setEnabled( state );
    updPasteActionsEnabled();
  }


  private void setDefaultStatusText()
  {
    this.labelStatus.setText( this.defaultStatusText );
  }


  private File showSaveWavFileDlg( File srcFile )
  {
    String presetFileName = null;
    if( srcFile != null ) {
      presetFileName = srcFile.getName();
    }
    if( presetFileName != null ) {
      int pos = presetFileName.lastIndexOf( '.' );
      if( pos >= 0 ) {
	presetFileName = presetFileName.substring( 0, pos );
      }
      if( !presetFileName.isEmpty() ) {
	presetFileName += ".wav";
      }
    }
    return UIUtil.showSaveFileDlg(
			this,
			Main.getText( "title.save_wav_file" ),
			this.recentWavFile != null ?
				this.recentWavFile.getParentFile() : null,
			presetFileName,
			new FileNameExtensionFilter(
				Main.getText( "filefilter.wav" )
							+ " (*.wav)",
				"wav" ) );
  }


  private void updFileActionsEnabled()
  {
    boolean normalMode = this.runMode.equals( RunMode.NORMAL );
    this.mnuFileNew.setEnabled( normalMode );
    this.mnuFileOpen.setEnabled( normalMode );
    this.mnuFileSave.setEnabled(
			normalMode && this.textEditFld.hasTextChanged() );
    this.mnuFileSaveAs.setEnabled( normalMode );
    this.mnuFilePrint.setEnabled( normalMode );
    this.mnuEditGotoLine.setEnabled( normalMode );

    this.btnFileNew.setEnabled( normalMode );
    this.btnFileOpen.setEnabled( normalMode );
    this.btnFileSave.setEnabled( this.mnuFileSave.isEnabled() );
    this.btnFilePrint.setEnabled( normalMode );

    boolean state = this.runMode.equals( RunMode.DEBUG ) || normalMode;
    this.mnuEditFind.setEnabled( state );
    this.btnEditFind.setEnabled( state );
    findNextPrevStateChanged( state && this.textFinder.hasSearchText() );
    this.mnuEditSelectAll.setEnabled( state );

    undoStateChanged( normalMode && this.textEditFld.canUndo() );
    updPasteActionsEnabled();
  }


  private void updPasteActionsEnabled()
  {
    boolean state = false;
    if( (this.runMode.equals( RunMode.NORMAL )
		|| this.runMode.equals( RunMode.DEBUG ))
	&& (this.clipboard != null)
	&& (this.curTextComp != null) )
    {
      if( this.curTextComp.isEditable() ) {
	try {
	  for( DataFlavor df : this.clipboard.getAvailableDataFlavors() ) {
	    if( df.equals( DataFlavor.stringFlavor ) ) {
	      state = true;
	      break;
	    }
	  }
	}
	catch( IllegalStateException ex ) {}
      }
    }
    this.mnuEditPaste.setEnabled( state );
    this.btnEditPaste.setEnabled( state );
  }


  private void updPrgActionsEnabled()
  {
    boolean debugMode = this.runMode.equals( RunMode.DEBUG );
    boolean running   = (this.bcRuntime != null);
    if( !debugMode ) {
      this.btnRun.setEnabled( !running );
    }
    this.mnuExportPrgToWAV.setEnabled( !running );
    this.mnuPrgCheckSyntax.setEnabled( !running );
    this.mnuPrgRunNormal.setEnabled( !running );
    this.mnuPrgRunLarge.setEnabled( !running );
    if( this.bcBreakKeyEnabled ) {
      this.mnuPrgAbort.setVisible( false );
      this.mnuPrgAbort.setEnabled( false );
      this.mnuPrgAbortESC.setVisible( true );
      this.mnuPrgAbortESC.setEnabled( running );
    } else {
      this.mnuPrgAbort.setVisible( true );
      this.mnuPrgAbort.setEnabled( running );
      this.mnuPrgAbortESC.setVisible( false );
      this.mnuPrgAbortESC.setEnabled( false );
    }
    this.btnRun.setEnabled( !running );
    this.btnStop.setEnabled( running );
  }
}
