/*
 * (c) 2025 Jens Mueller
 *
 * JBASICODE
 *
 * Printable to print a plain text
 */

package jbasicode.ui;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterException;
import javax.swing.JTextArea;
import jbasicode.Main;


public class PlainTextPrintable implements Printable
{
  private String text;
  private int    tabSize;
  private int    contentFontSize;
  private Font   contentFont;
  private Font   footerFont;
  private String fileName;


  public PlainTextPrintable( JTextArea textArea, String fileName )
  {
    this(
	textArea.getText(),
	textArea.getTabSize(),
	textArea.getFont(),
	fileName );
  }


  public PlainTextPrintable(
			String text,
			int    tabSize,
			Font   font,
			String fileName )
  {
    this.text            = text;
    this.tabSize         = tabSize;
    this.fileName        = fileName;
    this.footerFont      = null;
    this.contentFont     = null;
    this.contentFontSize = Main.getIntProperty(
				PrintOptionsDlg.PROP_PRINT_FONT_SIZE,
				PrintOptionsDlg.DEFAULT_PRINT_FONT_SIZE );
    if( this.contentFontSize > 0 ) {
      if( font != null ) {
	if( font.getSize() == this.contentFontSize ) {
	  this.contentFont = font;
	} else {
	  this.contentFont = font.deriveFont( (float) this.contentFontSize );
	}
      } else {
	this.contentFont = new Font(
				Font.MONOSPACED,
				Font.PLAIN,
				this.contentFontSize );
      }
      int footerSize = this.contentFontSize - 2;
      if( footerSize < 6 ) {
	footerSize = 6;
      }
      this.footerFont = this.contentFont.deriveFont( (float) footerSize );
    }
  }


	/* --- Printable --- */

  @Override
  public int print(
		Graphics   g,
		PageFormat pf,
		int        pageNum ) throws PrinterException
  {
    int rv = NO_SUCH_PAGE;
    if( (this.text != null)
	&& (this.tabSize > 0)
	&& (this.contentFontSize > 0)
	&& (this.contentFont != null) )
    {
      boolean printFileName = Main.getBooleanProperty(
				PrintOptionsDlg.PROP_PRINT_FILE_NAME,
				false );
      boolean printPageNum = Main.getBooleanProperty(
				PrintOptionsDlg.PROP_PRINT_PAGE_NUM,
				false );

      // number of lines per page
      int linesPerPage = (int) (pf.getImageableHeight()
					/ this.contentFontSize);
      if( printFileName || printPageNum ) {
	linesPerPage -= 2;
      }
      if( (linesPerPage < 1) || (pf.getImageableWidth() < 1.0) ) {
	throw new PrinterException(
		"Die Seite hat keinen bedruckbaren Bereich,\n"
			+ "da die R\u00E4nder zu gro\u00DF sind." );
      }

      // skip previous pages
      int pagesToSkip = pageNum;
      int len         = this.text.length();
      int pos         = 0;
      int linesOnPage = 0;
      while( (pagesToSkip > 0) && (pos < len) ) {
	char ch = this.text.charAt( pos++ );
	if( ch == '\f' ) {
	  --pagesToSkip;
	  linesOnPage = 0;
	} else {
	  if( linesOnPage == 0 ) {
	    linesOnPage++;
	  }
	  if( ch == '\n' ) {
	    linesOnPage++;
	  }
	  if( linesOnPage > linesPerPage ) {
	    --pagesToSkip;
	    linesOnPage = 0;
	  }
	}
      }
      if( (pagesToSkip == 0) && (pos < len) ) {

	// print page
	g.setFont( this.contentFont );
	g.setColor( Color.BLACK );

	int x = (int) pf.getImageableX();
	int y = (int) pf.getImageableY() + this.contentFontSize;
	while( (linesPerPage > 0) && (pos < len) ) {
	  boolean newPage = false;
	  int     eop     = this.text.indexOf( '\f', pos );
	  int     idx     = this.text.indexOf( '\n', pos );
	  if( (idx < 0) || ((eop >= 0) && (eop < idx)) ) {
	    idx     = eop;
	    newPage = true;
	  }
	  if( idx >= pos ) {
	    g.drawString(
		expandTabs( this.text.substring( pos, idx ), this.tabSize ),
		x,
		y );
	    pos = idx + 1;
	  } else {
	    g.drawString(
		expandTabs( this.text.substring( pos ), this.tabSize ),
		x,
		y );
	    pos = len;
	  }
	  if( newPage ) {
	    break;
	  }
	  y += this.contentFontSize;
	  --linesPerPage;
	}

	// print footer
	if( this.footerFont != null ) {
	  y = (int) (pf.getImageableY() + pf.getImageableHeight()) - 1;
	  if( printFileName && (this.fileName != null) ) {
	    g.setFont( this.footerFont );
	    g.drawString( this.fileName, x, y );
	    if( printPageNum ) {
	      String s = String.valueOf( pageNum + 1 );
	      g.drawString(
		s,
		(int) (pf.getImageableX() + pf.getImageableWidth()
				- g.getFontMetrics().stringWidth( s )),
		y );
	    }
	  } else {
	    if( printPageNum ) {
	      printCenteredPageNum( g, pf, this.footerFont, pageNum + 1 );
	    }
	  }
	}
	rv = PAGE_EXISTS;
      }
    }
    return rv;
  }


	/* --- Private Methods --- */

  private static String expandTabs( String text, int tabSize )
  {
    String rv = "";
    if( text != null ) {
      if( text.indexOf( '\t' ) >= 0 ) {
	StringBuilder buf = new StringBuilder();
	int           len = text.length();
	for( int i = 0; i < len; i++ ) {
	  char ch = text.charAt( i );
	  if( ch == '\t' ) {
	    int n = tabSize - (buf.length() % tabSize);
	    for( int k = 0; k < n; k++ ) {
	      buf.append( '\u0020' );
	    }
	  } else {
	    buf.append( ch );
	  }
	}
	rv = buf.toString();
      } else {
	rv = text;
      }
    }
    return rv;
  }


  private static void printCenteredPageNum(
				Graphics   g,
				PageFormat pf,
				Font       font,
				int        pageNum )
  {
    String s = "- " + String.valueOf( pageNum ) + " -";
    g.setColor( Color.BLACK );
    g.setFont( font );
    g.drawString(
	s,
	(int) (pf.getImageableX() + (pf.getImageableWidth() / 2))
			- (g.getFontMetrics().stringWidth( s ) / 2),
	(int) (pf.getImageableY() + pf.getImageableHeight()) - 1 );
  }


  private static void printCenteredPageNum(
				Graphics   g,
				PageFormat pf,
				int        fontSize,
				int        pageNum )
  {
    printCenteredPageNum(
		g,
		pf,
		new Font( Font.MONOSPACED, Font.PLAIN, fontSize ),
		pageNum );
  }
}
