/*
 * (c) 2025 Jens Mueller
 *
 * JBASICODE
 *
 * Dialog to edit print options
 */

package jbasicode.ui;

import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JPanel;
import jbasicode.Main;


public class PrintOptionsDlg extends JDialog implements ActionListener
{
  public static final String PROP_PRINT_FONT_SIZE    = "print.font.size";
  public static final String PROP_PRINT_FILE_NAME    = "print.file.name";
  public static final String PROP_PRINT_PAGE_NUM     = "print.page.num";
  public static final int    DEFAULT_PRINT_FONT_SIZE = 10;

  private static final Integer[] fontSizes = {
			6, 7, 8, 9, 10, 11, 12, 13,
			14, 16, 18, 20, 22, 24 };

  private boolean            notified;
  private JComboBox<Integer> comboFontSize;
  private JCheckBox          cbFileName;
  private JCheckBox          cbPageNum;
  private JButton            btnOK;
  private JButton            btnCancel;


  public static void open( Window owner, boolean fileNameOption )
  {
    (new PrintOptionsDlg( owner, fileNameOption )).setVisible( true );
  }


  @Override
  public void addNotify()
  {
    super.addNotify();
    if( !this.notified ) {
      this.notified = true;
      this.btnOK.addActionListener( this );
      this.btnCancel.addActionListener( this );
    }
  }


  @Override
  public void removeNotify()
  {
    if( this.notified ) {
      this.notified = false;
      this.btnOK.removeActionListener( this );
      this.btnCancel.removeActionListener( this );
    }
    super.removeNotify();
  }

	/* --- ActionListener --- */

  @Override
  public void actionPerformed( ActionEvent e )
  {
    Object src = e.getSource();
    if( src == this.btnOK ) {
      doApply();
    }
    else if( src == this.btnCancel ) {
      doClose();
    }
  }


	/* --- Constructor --- */

  private PrintOptionsDlg( Window owner, boolean fileNameOption )
  {
    super( owner );
    setDefaultCloseOperation( JDialog.DISPOSE_ON_CLOSE );
    setTitle( Main.getText( "title.print_options" ) );
    this.notified = false;


    // Window content
    setLayout( new GridBagLayout() );

    GridBagConstraints gbc = new GridBagConstraints(
					0, 0,
					1, 1,
					0.0, 0.0,
					GridBagConstraints.WEST,
					GridBagConstraints.NONE,
					new Insets( 5, 5, 0, 5 ),
					0, 0 );

    add( new JLabel( Main.getText( "label.font_size" ) ), gbc );

    this.comboFontSize = new JComboBox<>( fontSizes );
    this.comboFontSize.setEditable( false );
    gbc.gridx++;
    add( this.comboFontSize, gbc );

    this.comboFontSize.setSelectedItem(
				Main.getIntProperty(
					PROP_PRINT_FONT_SIZE,
					DEFAULT_PRINT_FONT_SIZE ) );

    if( fileNameOption ) {
      this.cbFileName = new JCheckBox(
				Main.getText( "option.print_file_name" ),
				Main.getBooleanProperty(
					PROP_PRINT_FILE_NAME,
					false ) );
      gbc.gridwidth = 2;
      gbc.gridx     = 0;
      gbc.gridy++;
      add( this.cbFileName, gbc );
    } else {
      this.cbFileName = null;
    }

    this.cbPageNum = new JCheckBox(
				Main.getText( "option.print_page_num" ),
				Main.getBooleanProperty(
					PROP_PRINT_PAGE_NUM,
					false ) );
    gbc.insets.top = 0;
    gbc.gridwidth = 2;
    gbc.gridx     = 0;
    gbc.gridy++;
    add( this.cbPageNum, gbc );


    // Buttons
    JPanel panelBtn   = new JPanel( new GridLayout( 1, 2, 5, 5 ) );
    gbc.insets.top    = 10;
    gbc.insets.left   = 10;
    gbc.insets.right  = 10;
    gbc.insets.bottom = 10;
    gbc.gridy++;
    add( panelBtn, gbc );

    this.btnOK = new JButton( Main.getText( "action.ok" ) );
    panelBtn.add( this.btnOK );

    this.btnCancel = new JButton( Main.getText( "action.cancel" ) );
    panelBtn.add( this.btnCancel );


    // Window size
    pack();
    UIUtil.setParentCentered( this );
    setResizable( false );
  }


  private void doApply()
  {
    Object obj = this.comboFontSize.getSelectedItem();
    if( obj != null ) {
      String text = obj.toString();
      if( text != null ) {
	Main.setProperty( PROP_PRINT_FONT_SIZE, text );
	if( this.cbFileName != null ) {
	  Main.setProperty(
			PROP_PRINT_FILE_NAME,
			String.valueOf( this.cbFileName.isSelected() ) );
	}
	Main.setProperty(
			PROP_PRINT_PAGE_NUM,
			String.valueOf( this.cbPageNum.isSelected() ) );
	doClose();
      }
    }
  }


  private void doClose()
  {
    setVisible( false );
    dispose();
  }
}
