/*
 * (c) 2025 Jens Mueller
 *
 * JBASICODE
 *
 * Window with printer output of the BASICODE program
 */

package jbasicode.ui;

import java.awt.BorderLayout;
import java.awt.EventQueue;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.StringReader;
import java.util.Collections;
import java.util.Map;
import java.util.Vector;
import javax.swing.BorderFactory;
import javax.swing.JFrame;
import javax.swing.JList;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.KeyStroke;
import javax.swing.ListModel;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.swing.text.Document;
import jbasicode.Main;
import jbasicode.etc.IOUtil;


public class PrinterOutputFrm extends JFrame implements
						ActionListener,
						ListSelectionListener
{
  public static final String PROP_PREFIX  = Main.PROP_PREFIX + "printdoc.";

  private Map<Integer,Object> printDocs;
  private int                 curDocNum;
  private boolean             notified;
  private File                recentDirFile;
  private JList<Integer>      list;
  private JTextArea           textArea;
  private JMenuItem           mnuClose;
  private JMenuItem           mnuPrint;
  private JMenuItem           mnuPrintOptions;
  private JMenuItem           mnuRemove;
  private JMenuItem           mnuRemoveAll;
  private JMenuItem           mnuSaveAs;


  public PrinterOutputFrm( Map<Integer,Object> printDocs )
  {
    this.printDocs     = printDocs;
    this.curDocNum     = -1;
    this.notified      = false;
    this.recentDirFile = null;
    UIUtil.setIconImagesAt( this );
    setDefaultCloseOperation( JFrame.DISPOSE_ON_CLOSE );
    setTitle( Main.APPNAME + "\u0020"
			+ Main.getText( "title.window.printer_output" ) );


    // Menu
    JMenu mnuFile = new JMenu( Main.getText( "menu.file" ) );

    this.mnuPrintOptions = new JMenuItem(
			Main.getText( "action.file.print_options" ) );
    mnuFile.add( this.mnuPrintOptions );

    this.mnuPrint = new JMenuItem( Main.getText( "action.file.print" ) );
    this.mnuPrint.setEnabled( false );
    mnuFile.add( this.mnuPrint );

    this.mnuSaveAs = new JMenuItem( Main.getText( "action.file.save_as" ) );
    this.mnuSaveAs.setEnabled( false );
    mnuFile.add( this.mnuSaveAs );
    mnuFile.addSeparator();

    this.mnuRemove = new JMenuItem( Main.getText( "action.remove" ) );
    this.mnuRemove.setAccelerator(
		KeyStroke.getKeyStroke( KeyEvent.VK_DELETE, 0 ) );
    this.mnuRemove.setEnabled( false );
    mnuFile.add( this.mnuRemove );

    this.mnuRemoveAll = new JMenuItem( Main.getText( "action.remove_all" ) );
    mnuFile.add( this.mnuRemoveAll );
    mnuFile.addSeparator();

    this.mnuClose = new JMenuItem( Main.getText( "action.close" ) );
    mnuFile.add( this.mnuClose );

    JMenuBar mnuBar = new JMenuBar();
    mnuBar.add( mnuFile );
    setJMenuBar( mnuBar );


    // Window content
    setLayout( new BorderLayout( 5, 5 ) );

    JPanel panelWest = new JPanel( new BorderLayout( 0, 0 ) );
    panelWest.setBorder( BorderFactory.createTitledBorder(
				Main.getText( "title.area.num" ) ) );
    add( panelWest, BorderLayout.WEST );

    
    this.list = new JList<>( getSortedDocNumVector() );
    this.list.setDragEnabled( false );
    this.list.setLayoutOrientation( JList.VERTICAL );
    this.list.setPrototypeCellValue( 888 );
    this.list.setSelectionMode( ListSelectionModel.SINGLE_SELECTION );
    panelWest.add( new JScrollPane( this.list ), BorderLayout.WEST );

    JPanel panelCenter = new JPanel( new BorderLayout( 0, 0 ) );
    panelCenter.setBorder( BorderFactory.createTitledBorder(
				Main.getText( "title.area.content" ) ) );
    add( panelCenter, BorderLayout.CENTER );

    this.textArea = new JTextArea( 10,40 );
    this.textArea.setEditable( false );
    panelCenter.add( new JScrollPane( this.textArea ), BorderLayout.CENTER );


    // Window position
    setResizable( true );
    if( !UIUtil.restoreWindowBounds(
				this,
				Main.getProperties(),
				PROP_PREFIX ) )
    {
      pack();
      setLocationByPlatform( true );
    }


    // Etc
    updRemoveAllAction();
    EventQueue.invokeLater( ()->initFld() );
  }


  public void doClose()
  {
    setVisible( false );
    dispose();
  }


  public void printDocAdded( int docNum )
  {
    this.list.setListData( getSortedDocNumVector() );
    printDocChanged( docNum );
    if( getListSize() == 1 ) {
      showDoc( docNum );
    } else {
      updContentFld();
    }
    updRemoveAllAction();
  }


  public void printDocChanged( int docNum )
  {
    if( docNum == this.curDocNum )
      showDoc( docNum );
  }


  @Override
  public void addNotify()
  {
    super.addNotify();
    if( !this.notified ) {
      this.notified = true;
      this.list.addListSelectionListener( this );
      this.mnuClose.addActionListener( this );
      this.mnuPrint.addActionListener( this );
      this.mnuPrintOptions.addActionListener( this );
      this.mnuRemove.addActionListener( this );
      this.mnuRemoveAll.addActionListener( this );
      this.mnuSaveAs.addActionListener( this );
    }
  }


  @Override
  public void removeNotify()
  {
    if( this.notified ) {
      this.notified = false;
      this.list.removeListSelectionListener( this );
      this.mnuClose.removeActionListener( this );
      this.mnuPrint.removeActionListener( this );
      this.mnuPrintOptions.removeActionListener( this );
      this.mnuRemove.removeActionListener( this );
      this.mnuRemoveAll.removeActionListener( this );
      this.mnuSaveAs.removeActionListener( this );
    }
    super.removeNotify();
  }


	/* --- ActionListener --- */

  @Override
  public void actionPerformed( ActionEvent e )
  {
    Object src = e.getSource();
    if( src == this.mnuClose ) {
      doClose();
    } else if( src == this.mnuPrint ) {
      doPrint();
    } else if( src == this.mnuPrintOptions ) {
      PrintOptionsDlg.open( this, false );
    } else if( src == this.mnuRemove ) {
      doRemove();
    } else if( src == this.mnuRemoveAll ) {
      doRemoveAll();
    } else if( src == this.mnuSaveAs ) {
      doSaveAs();
    }
  }


	/* --- ListSelectionListener --- */

  @Override
  public void valueChanged( ListSelectionEvent e )
  {
    if( e.getSource() == this.list )
      updContentFld();
  }


	/* --- Private Methods --- */

  private void doPrint()
  {
    Document doc = this.textArea.getDocument();
    if( doc != null ) {
      if( doc.getLength() > 0 ) {
	UIUtil.doPrint( this.textArea, null );
      }
    }
  }


  private void doRemove()
  {
    Integer docNum = this.list.getSelectedValue();
    if( docNum != null ) {
      this.printDocs.remove( docNum );
      showDoc( null );
      this.list.setListData( getSortedDocNumVector() );
    }
  }


  private void doRemoveAll()
  {
    this.printDocs.clear();
    this.list.setListData( new Integer[ 0 ] );
    showDoc( null );
  }


  private void doSaveAs()
  {
    String text = this.textArea.getText();
    if( text != null ) {
      if( !text.isEmpty() ) {
	File file = UIUtil.showSaveFileDlg(
			this,
			Main.getText( "title.save_text_file" ),
			this.recentDirFile,
			null,
			new FileNameExtensionFilter(
				Main.getText( "filefilter.txt" )
							+ " (*.txt)",
				"txt" ) );
	if( file != null ) {
	  try {
	    BufferedReader reader = null;
	    BufferedWriter writer = null;
	    try {
	      reader = new BufferedReader( new StringReader( text ) );
	      writer = new BufferedWriter(
				new OutputStreamWriter(
					new FileOutputStream( file ),
					"UTF-8" ) );

	      String line = reader.readLine();
	      while( line != null ) {
		writer.write( line );
		writer.newLine();
		line = reader.readLine();
	      }
	      writer.close();
	      writer = null;
	      this.recentDirFile = file.getParentFile();
	    }
	    finally {
	      IOUtil.closeSilently( reader );
	      IOUtil.closeSilently( writer );
	    }
	  }
	  catch( IOException ex ) {
	    UIUtil.showErrorMsg( this, null, ex );
	  }
	}
      }
    }
  }


  private int getListSize()
  {
    ListModel<Integer> model = this.list.getModel();
    return model != null ? model.getSize() : 0;
  }


  private Vector<Integer> getSortedDocNumVector()
  {
    Vector<Integer> vector = new Vector<>();
    vector.addAll( this.printDocs.keySet() );
    try {
      Collections.sort( vector );
    }
    catch( ClassCastException ex ) {}
    return vector;
  }


  private void initFld()
  {
    this.textArea.setColumns( 0 );
    this.textArea.setRows( 0 );
    if( getListSize() == 1 ) {
      for( Integer docNum : this.printDocs.keySet() ) {
	showDoc( docNum );
	break;
      }
    }
  }


  private void showDoc( Integer docNum )
  {
    String text  = null;
    Object value = null;
    if( docNum != null ) {
      value = this.printDocs.get( docNum );
      if( value != null ) {
	text = value.toString();
      }
      this.curDocNum = docNum.intValue();
    } else {
      this.curDocNum = -1;
    }
    if( text == null ) {
      text = "";
    }
    this.textArea.setText( text );
    try {
      this.textArea.setCaretPosition( 0 );
    }
    catch( IllegalArgumentException ex ) {}

    boolean state = !text.isEmpty();
    this.mnuPrint.setEnabled( state );
    this.mnuSaveAs.setEnabled( state );
  }


  private void updContentFld()
  {
    boolean state  = false;
    Integer docNum = this.list.getSelectedValue();
    if( docNum != null ) {
      state = true;
    }
    showDoc( docNum );
    this.mnuRemove.setEnabled( state );
  }


  private void updRemoveAllAction()
  {
    this.mnuRemoveAll.setEnabled( getListSize() > 0 );
  }
}
