/*
 * (c) 2025 Jens Mueller
 *
 * JBASICODE
 *
 * Editor Component
 */

package jbasicode.ui;

import java.awt.BorderLayout;
import java.awt.EventQueue;
import java.awt.Font;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.dnd.DropTarget;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DropTargetEvent;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.dnd.DropTargetListener;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.Collection;
import java.util.Iterator;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.Writer;
import java.text.CharacterIterator;
import java.text.MessageFormat;
import javax.swing.JOptionPane;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.Timer;
import javax.swing.event.CaretEvent;
import javax.swing.event.CaretListener;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.event.UndoableEditEvent;
import javax.swing.event.UndoableEditListener;
import javax.swing.filechooser.FileFilter;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;
import javax.swing.text.Segment;
import javax.swing.undo.UndoManager;
import jbasicode.Main;
import jbasicode.etc.IOUtil;


public class TextEditFld extends JPanel
				implements
					ActionListener,
					CaretListener,
					DocumentListener,
					DropTargetListener,
					MouseListener,
					UndoableEditListener
{
  private static final int MAX_FILE_SIZE = (0x7FFF - 1000) * 62;

  private static final byte[] bomUTF8 = new byte[] {
						(byte) 0xEF,
						(byte) 0xBB,
						(byte) 0xBF };
  private static final byte[] bomUTF16BE = new byte[] {
						(byte) 0xFE,
						(byte) 0xFF };
  private static final byte[] bomUTF16LE = new byte[] {
						(byte) 0xFF,
						(byte) 0xFE };

  private MainFrm     mainFrm;
  private boolean     notified;
  private boolean     textChanged;
  private boolean     bom;
  private String      encoding;
  private String      eol;
  private String      caretPosStatusText;
  private String      fileName;
  private File        file;
  private long        fileMillis;
  private FileFilter  fileFilter;
  private JPopupMenu  popupMenu;
  private JMenuItem   popupCopy;
  private JMenuItem   popupCut;
  private JMenuItem   popupPaste;
  private JMenuItem   popupToUpper;
  private JMenuItem   popupSelectAll;
  private JTextArea   textArea;
  private Document    document;
  private UndoManager undoMngr;
  private Timer       unselectTimer;
  private DropTarget  dropTargetScrollPane;
  private DropTarget  dropTargetTextArea;


  public TextEditFld( MainFrm mainFrm )
  {
    this.mainFrm            = mainFrm;
    this.textChanged        = false;
    this.bom                = false;
    this.encoding           = "UTF-8";
    this.eol                = "\r\n";
    this.caretPosStatusText = Main.getText( "status.cursor_position" );
    this.file               = null;
    this.fileName           = null;
    this.fileMillis         = -1;
    this.undoMngr           = new UndoManager();
    this.unselectTimer      = new Timer( 300, this );
    this.unselectTimer.setRepeats( false );

    StringBuilder buf = new StringBuilder( 128 );
    buf.append( Main.getText( "filefilter.bc" ) );
    buf.append( " (*.asc; *.bc; *.bc2; *.bc3; *.txt)" );
    this.fileFilter = new FileNameExtensionFilter(
				buf.toString(),
				"asc", "bas", "bc", "bc2", "bc3", "txt" );

    this.popupMenu = new JPopupMenu();

    this.popupCut = new JMenuItem( Main.getText( "action.edit.cut" ) );
    this.popupMenu.add( this.popupCut );

    this.popupCopy = new JMenuItem( Main.getText( "action.edit.copy" ) );
    this.popupMenu.add( this.popupCopy );

    this.popupPaste = new JMenuItem( Main.getText( "action.edit.paste" ) );
    this.popupMenu.add( this.popupPaste );
    this.popupMenu.addSeparator();

    this.popupToUpper = new JMenuItem(
				Main.getText( "action.edit.to_upper" ) );
    this.popupMenu.add( this.popupToUpper );
    this.popupMenu.addSeparator();

    this.popupSelectAll = new JMenuItem(
				Main.getText( "action.edit.select_all" ) );
    this.popupMenu.add( this.popupSelectAll );


    setLayout( new BorderLayout() );
    this.textArea = new JTextArea( 28, 60 );
    this.textArea.addCaretListener( this );
    Font font     = this.textArea.getFont();
    this.textArea.setFont(
		new Font(
			Font.MONOSPACED,
			Font.PLAIN,
			font != null ? font.getSize() : 13 ) );
    JScrollPane scrollPane = new JScrollPane( this.textArea );
    add( scrollPane, BorderLayout.CENTER );

    this.document = this.textArea.getDocument();
    if( this.document != null ) {
      this.document.addDocumentListener( this );
      this.document.addUndoableEditListener( this );
    }

    this.dropTargetScrollPane = new DropTarget( this.textArea, this );
    this.dropTargetTextArea = new DropTarget( scrollPane, this );
    this.dropTargetScrollPane.setActive( true );
    this.dropTargetTextArea.setActive( true );
  }


  public boolean canUndo()
  {
    return this.undoMngr.canUndo();
  }


  public static void checkFileSize( long fileSize ) throws IOException
  {
    if( fileSize > MAX_FILE_SIZE )
      throw new IOException( Main.getText( "error.file_too_large" ) );
  }


  public boolean confirmTextSaved()
  {
    boolean rv = !this.textChanged;
    if( this.textChanged ) {
      Object[] options = new Object[] {
		Main.getText( "option.save" ),
		Main.getText( "option.dismiss" ),
		Main.getText( "option.cancel" ) };
      JOptionPane pane = new JOptionPane(
		Main.getText( "confirm.text_not_saved" ),
		JOptionPane.WARNING_MESSAGE );
      pane.setOptions( options );
      pane.createDialog(
		this.mainFrm,
		Main.getText( "title.text_not_saved" ) ).setVisible( true );
      Object option = pane.getValue();
      if( option != null ) {
	if( option == options[ 0 ] ) {
	  rv = doSave( false );
	} else if( option == options[ 1 ] ) {
	  rv = true;
	}
      }
    }
    return rv;
  }


  public void doFind()
  {
    TextFinder textFinder = this.mainFrm.getTextFinder();
    textFinder.openFindAndReplaceDlg( this.textArea );
    if( textFinder.hasSearchText() ) {
      this.mainFrm.findNextPrevStateChanged( true );
    }
  }


  public void doFindNext()
  {
    this.mainFrm.getTextFinder().findNext( this.textArea );
    this.textArea.requestFocus();
  }


  public void doFindPrev()
  {
    this.mainFrm.getTextFinder().findPrev( this.textArea );
    this.textArea.requestFocus();
  }


  public void doGotoLine()
  {
    int lineNum = -1;
    while( lineNum < 1 ) {
      String text = JOptionPane.showInputDialog(
			this.mainFrm,
			Main.getText( "label.line_num" ),
			Main.getText( "title.goto_line" ),
			JOptionPane.QUESTION_MESSAGE );
      if( text == null ) {
	break;
      }
      try {
	lineNum = Integer.parseInt( text.trim() );
      }
      catch( NumberFormatException ex ) {}
    }
    if( lineNum > 0 ) {
      fireGotoLine( lineNum );
    }
  }


  public void doNew()
  {
    if( confirmTextSaved() ) {
      this.textChanged = false;
      this.bom         = false;
      this.encoding    = "UTF-8";
      this.eol         = "\r\n";
      this.file        = null;
      this.fileName    = null;
      this.fileMillis  = -1;
      this.textArea.setText( "" );
      this.undoMngr.discardAllEdits();
      this.mainFrm.undoStateChanged( false );
      fireTextUnchanged();
      this.mainFrm.fireShowDefaultStatusText();
      this.mainFrm.fireFileStatusChanged( MainFrm.STATUS_OTHER_FILE
						| MainFrm.STATUS_SAVE_ACTION
						| MainFrm.STATUS_TITLE );
    }
  }


  public void doOpen()
  {
    if( confirmTextSaved() ) {
      File file = UIUtil.showOpenFileDlg(
			this.mainFrm,
			Main.getText( "title.open_prg_file" ),
			getPresetPrgDir(),
			null,
			this.fileFilter );
      if( file != null ) {
	openFile( file );
      }
    }
  }


  public void doPrint()
  {
    UIUtil.doPrint( this.textArea, this.fileName );
  }


  public boolean doSave( boolean forceFileDlg )
  {
    boolean rv    = false;
    boolean state = true;
    if( !forceFileDlg && (this.file != null) ) {
      long millis = readLastModified( this.file );
      if( (millis > 0) && (this.fileMillis > 0)
	  && (millis != this.fileMillis) )
      {
	if( JOptionPane.showConfirmDialog(
		  this.mainFrm,
		  Main.getText( "confirm.file_changed" ),
		  Main.getText( "title.file_changed" ),
		  JOptionPane.WARNING_MESSAGE,
		  JOptionPane.YES_NO_OPTION ) == JOptionPane.YES_OPTION )
	{
	  state = true;
	}
      }
    }
    if( state ) {
      File file = this.file;
      if( forceFileDlg ) {
	file = null;
      }
      if( file == null ) {
	file = UIUtil.showSaveFileDlg(
			this.mainFrm,
			Main.getText( "title.save_prg_file" ),
			getPresetPrgDir(),
			this.fileName,
			this.fileFilter );
      }
      if( file != null ) {
	rv = saveFile( file );
      }
    }
    return rv;
  }


  public void doToUpper()
  {
    try {
      String text = this.textArea.getSelectedText();
      if( text != null ) {
	this.textArea.replaceSelection( text.toUpperCase() );
	this.textArea.requestFocus();
      }
    }
    catch( IllegalArgumentException ex ) {}
  }


  public void doUndo()
  {
    if( this.undoMngr.canUndo() ) {
      this.undoMngr.undo();
      this.mainFrm.undoStateChanged( this.undoMngr.canUndo() );
      this.textArea.requestFocus();
    }
  }


  public void fireGotoLine( int lineNum )
  {
    EventQueue.invokeLater( ()->gotoLine( lineNum ) );
  }


  public File getFile()
  {
    return this.file;
  }


  public String getFileName()
  {
    return this.fileName;
  }


  public String getText()
  {
    return this.textArea.getText();
  }


  public JTextArea getTextArea()
  {
    return this.textArea;
  }


  public boolean hasTextChanged()
  {
    return this.textChanged;
  }


  public boolean openFile(
			BufferedInputStream in,
			File                file,
			String              fileName )
  {
    boolean rv = false;
    try {
      long millis = -1;
      if( file != null ) {
	millis = file.lastModified();
      }
      Reader reader = null;
      try {
	in.mark( 3 );

	// check byte order mark
	String enc = "UTF-8";
	int    b0  = in.read();
	int    b1  = in.read();
	if( (b0 >= 0) && (b1 >= 0) ) {
	  if( ((byte) b0 == bomUTF16BE[ 0 ])
	      && ((byte) b1 == bomUTF16BE[ 1 ]) )
	  {
	    enc = "UTF-16BE";
	    bom = true;
	  } else if( ((byte) b0 == bomUTF16LE[ 0 ])
		     && ((byte) b1 == bomUTF16LE[ 1 ]) )
	  {
	    enc = "UTF-16LE";
	    bom = true;
	  } else if( ((byte) b0 == bomUTF8[ 0 ])
		     && ((byte) b1 == bomUTF8[ 1 ]) )
	  {
	    int b = in.read();
	    if( (b >= 0) && ((byte) b == bomUTF8[ 2 ]) ) {
	      bom = true;
	    }
	  }
	}
	if( !bom ) {
	  in.reset();
	}
	reader = new InputStreamReader( in, enc );

	// read file
	StringBuilder buf = new StringBuilder( 0x4000 );
	String        eol = null;
	boolean       cr  = false;
	int           ch  = reader.read();
	while( ch >= 0 ) {
	  if( ch == '\r' ) {
	    if( cr ) {
	      buf.append( '\n' );
	      if( eol == null ) {
		eol = "\r";
	      }
	    }
	    cr = true;
	  } else if( ch == '\n' ) {
	    buf.append( '\n' );
	    if( eol == null ) {
	      if( cr ) {
		eol = "\r\n";
	      } else {
		eol = "\n";
	      }
	    }
	    cr = false;
	  } else {
	    if( cr ) {
	      buf.append( '\n' );
	      if( eol == null ) {
		eol = "\r";
	      }
	      cr = false;
	    }
	    buf.append( (char) ch );
	  }
	  ch = reader.read();
	}
	if( cr ) {
	  buf.append( '\n' );
	}
	this.textArea.setText( buf.toString() );
	try {
	  this.textArea.setCaretPosition( 0 );
	}
	catch( IllegalArgumentException ex ) {}
	this.textChanged = false;
	this.bom         = bom;
	this.encoding    = enc;
	this.file        = file;
	this.fileName    = fileName;
	this.fileMillis  = millis;
	if( eol != null ) {
	  this.eol = eol;
	}
	this.undoMngr.discardAllEdits();
	this.mainFrm.undoStateChanged( false );
	fireTextUnchanged();
	this.mainFrm.fireShowStatusTextByKey( "status.file_opened" );
	this.mainFrm.fireFileStatusChanged( MainFrm.STATUS_OTHER_FILE
						| MainFrm.STATUS_SAVE_ACTION
						| MainFrm.STATUS_TITLE );
	Main.createSettingsDirIfNotExists();
	if( file != null ) {
	  IOUtil.writePathToFile(
			Main.getProgramPathFile(),
			file.getParentFile() );
	}
	rv = true;
      }
      finally {
	IOUtil.closeSilently( reader );
      }
    }
    catch( IOException ex ) {
      UIUtil.showErrorMsg( this.mainFrm, null, ex );
    }
    return rv;
  }


  public boolean openFile( File file )
  {
    boolean done = false;
    if( file != null ) {
      try {
	checkFileSize( file.length() );
	try( BufferedInputStream in = new BufferedInputStream(
					  new FileInputStream( file ) ) )
	{
	  done = openFile( in, file, file.getName() );
	}
      }
      catch( IOException ex ) {
	UIUtil.showErrorMsg( this.mainFrm, null, ex );
      }
    }
    return done;
  }


  public void resetPrefSize()
  {
    this.textArea.setColumns( 0 );
    this.textArea.setRows( 0 );
  }


  @Override
  public void addNotify()
  {
    super.addNotify();
    if( !this.notified ) {
      this.notified = true;
      if( this.textArea != null ) {
	this.popupCopy.addActionListener( this );
	this.popupCut.addActionListener( this );
	this.popupPaste.addActionListener( this );
	this.popupSelectAll.addActionListener( this );
	this.popupToUpper.addActionListener( this );
	this.textArea.addMouseListener( this );
      }
    }
  }


  @Override
  public void removeNotify()
  {
    if( this.notified ) {
      this.notified = false;
      if( this.textArea != null ) {
	this.popupCopy.removeActionListener( this );
	this.popupCut.removeActionListener( this );
	this.popupPaste.removeActionListener( this );
	this.popupSelectAll.removeActionListener( this );
	this.popupToUpper.removeActionListener( this );
	this.textArea.removeMouseListener( this );
      }
    }
    super.removeNotify();
  }


	/* --- ActionListener --- */

  @Override
  public void actionPerformed( ActionEvent e )
  {
    Object src = e.getSource();
    if( src == this.popupCopy ) {
      doCopy();
    } else if( src == this.popupCut ) {
      doCut();
    } else if( src == this.popupPaste ) {
      doPaste();
    } else if( src == this.popupSelectAll ) {
      doSelectAll();
    } else if( src == this.popupToUpper ) {
      doToUpper();
    } else if( src == this.unselectTimer ) {
      unselectText();
    }
  }


	/* --- CaretListener --- */

  @Override
  public void caretUpdate( CaretEvent e )
  {
    if( e.getSource() == this.textArea ) {
      try {
	int pos = textArea.getCaretPosition();
	int row = textArea.getLineOfOffset( pos );
	int bol = textArea.getLineStartOffset( row );
	int col = pos - bol;
	if( col > 0 ) {
	  Document doc     = textArea.getDocument();
	  int      tabSize = textArea.getTabSize();
	  if( (doc != null) && (tabSize > 0) ) {
	    Segment seg = new Segment();
	    seg.setPartialReturn( false );
	    doc.getText( bol, col, seg );
	    col = 0;
	    int ch = seg.first();
	    while( ch != CharacterIterator.DONE ) {
	      if( ch == '\t' ) {
		col = ((col / tabSize) + 1) * tabSize;
	      } else {
		col++;
	      }
	      ch = seg.next();
	    }
	  }
	}
	this.mainFrm.setStatusText(
			MessageFormat.format(
				this.caretPosStatusText,
				String.valueOf( row + 1 ),
				String.valueOf( col + 1 ) ) );
      }
      catch( BadLocationException ex ) {}
      this.mainFrm.caretUpdate( e );
    }
  }


	/* --- DocumentListener --- */

  @Override
  public void changedUpdate( DocumentEvent e )
  {
    docChanged( e );
  }


  @Override
  public void insertUpdate( DocumentEvent e )
  {
    docChanged( e );
  }


  @Override
  public void removeUpdate( DocumentEvent e )
  {
    docChanged( e );
  }


	/* --- DropTargetListener --- */

  @Override
  public void dragEnter( DropTargetDragEvent e )
  {
    if( !isFileDrop( e ) )
      e.rejectDrag();
  }


  @Override
  public void dragExit( DropTargetEvent e )
  {
    // empty
  }


  @Override
  public void dragOver( DropTargetDragEvent e )
  {
    // empty
  }


  @Override
  public void drop( DropTargetDropEvent e )
  {
    File file = fileDrop( e );
    if( file != null ) {
      EventQueue.invokeLater( ()->checkAndOpenFile( file, e ) );
    }
  }


  @Override
  public void dropActionChanged( DropTargetDragEvent e )
  {
    if( !isFileDrop( e ) )
      e.rejectDrag();
  }


	/* --- MouseListener --- */

  @Override
  public void mouseClicked( MouseEvent e )
  {
    if( e.isPopupTrigger() ) {
      showPopupMenu( e );
      e.consume();
    }
  }


  @Override
  public void mouseEntered( MouseEvent e )
  {
    // empty
  }


  @Override
  public void mouseExited( MouseEvent e )
  {
    // empty
  }


  @Override
  public void mousePressed( MouseEvent e )
  {
    if( e.isPopupTrigger() ) {
      showPopupMenu( e );
      e.consume();
    }
  }


  @Override
  public void mouseReleased( MouseEvent e )
  {
    if( e.isPopupTrigger() ) {
      showPopupMenu( e );
      e.consume();
    }
  }


	/* --- UndoableEditListener --- */

  @Override 
  public void undoableEditHappened( UndoableEditEvent e )
  {
    this.undoMngr.undoableEditHappened( e );
    this.mainFrm.undoStateChanged( this.undoMngr.canUndo() );
  }


	/* --- Private Methods --- */

  private void checkAndOpenFile( File file, DropTargetDropEvent e )
  {
    if( confirmTextSaved() ) {
      if( !Downloader.checkAndStart( this, file, e ) ) {
	openFile( file );
      }
    }
  }


  private void docChanged( DocumentEvent e )
  {
    if( e.getDocument() == this.document )
      setTextChanged( true );
  }


  private void doCopy()
  {
    this.textArea.copy();
    this.textArea.requestFocus();
  }


  private void doCut()
  {
    this.textArea.cut();
    this.textArea.requestFocus();
  }


  private void doPaste()
  {
    this.textArea.paste();
    this.textArea.requestFocus();
  }


  private void doSelectAll()
  {
    this.textArea.selectAll();
    this.textArea.requestFocus();
  }


  private static File fileDrop( DropTargetDropEvent e )
  {
    File file = null;
    if( isFileDrop( e ) ) {
      e.acceptDrop( DnDConstants.ACTION_COPY );    // don't delete source
      Transferable t = e.getTransferable();
      if( t != null ) {
	try {
	  Object o = t.getTransferData( DataFlavor.javaFileListFlavor );
	  if( o != null ) {
	    if( o instanceof Collection ) {
	      Iterator iter = ((Collection) o).iterator();
	      if( iter != null ) {
		while( iter.hasNext() ) {
		  o = iter.next();
		  if( o != null ) {
		    if( o instanceof File ) {
		      file = (File) o;
		    } else {
		      String s = o.toString();
		      if( s != null ) {
			if( !s.isEmpty() ) {
			  file = new File( s );
			  break;
			}
		      }
		    }
		  }
		}
	      }
	    }
	  }
	}
	catch( Exception ex ) {}
      }
      e.dropComplete( true );
    } else {
      e.rejectDrop();
    }
    return file;
  }


  private void fireTextUnchanged()
  {
    EventQueue.invokeLater( ()->setTextChanged( false ) );
  }


  private File getPresetPrgDir()
  {
    return this.file != null ?
		this.file.getParentFile()
		: IOUtil.readPathFromFile( Main.getProgramPathFile() );
  }


  private void gotoLine( int lineNum )
  {
    try {
      --lineNum;
      int startOffs = this.textArea.getLineStartOffset( lineNum );
      int endOffs   = this.textArea.getLineEndOffset( lineNum );
      if( (startOffs >= 0) && (endOffs >= startOffs) ) {
	this.textArea.setCaretPosition( endOffs );
	this.textArea.moveCaretPosition( startOffs );
	this.textArea.requestFocus();
	this.unselectTimer.restart();
      }
    }
    catch( BadLocationException ex ) {}
  }


  private static boolean isFileDrop( DropTargetDragEvent e )
  {
    boolean rv = false;
    int action = e.getDropAction();
    if( (action == DnDConstants.ACTION_COPY)
	|| (action == DnDConstants.ACTION_MOVE)
	|| (action == DnDConstants.ACTION_COPY_OR_MOVE)
	|| (action == DnDConstants.ACTION_LINK) )
    {
      rv = e.isDataFlavorSupported( DataFlavor.javaFileListFlavor );
    }
    return rv;
  }


  public static boolean isFileDrop( DropTargetDropEvent e )
  {
    boolean rv = false;
    int action = e.getDropAction();
    if( (action == DnDConstants.ACTION_COPY)
	|| (action == DnDConstants.ACTION_MOVE)
	|| (action == DnDConstants.ACTION_COPY_OR_MOVE)
	|| (action == DnDConstants.ACTION_LINK) )
    {
      rv = e.isDataFlavorSupported( DataFlavor.javaFileListFlavor );
    }
    return rv;
  }


  private long readLastModified( File file )
  {
    return (new File( file.getPath() )).lastModified();
  }


  private boolean saveFile( File file )
  {
    boolean rv = false;
    try {
      OutputStream out    = null;
      Writer       writer = null;
      try {
	out = new BufferedOutputStream( new FileOutputStream( file ) );
	if( this.bom ) {
	  switch( this.encoding ) {
	    case "URF-8":
	      out.write( bomUTF8 );
	      break;
	    case "UTF-16BE":
	      out.write( bomUTF16BE );
	      break;
	    case "UTF-16LE":
	      out.write( bomUTF16LE );
	      break;
	  }
	}
	writer = new OutputStreamWriter( out, this.encoding );

	String text = this.textArea.getText();
	if( text != null ) {
	  int len = text.length();
	  if( len > 0 ) {
	    for( int i = 0; i < len; i++ ) {
	      char ch = text.charAt( i );
	      if( ch == '\n' ) {
		writer.write( this.eol );
	      } else {
		writer.write( ch );
	      }
	    }
	  }
	}
	writer.close();
	writer          = null;
	this.file       = file;
	this.fileName   = file.getName();
	this.fileMillis = readLastModified( file );
	fireTextUnchanged();
	this.mainFrm.fireShowStatusTextByKey( "status.file_saved" );
	this.mainFrm.fireFileStatusChanged(
			MainFrm.STATUS_SAVE_ACTION | MainFrm.STATUS_TITLE );
	Main.createSettingsDirIfNotExists();
	IOUtil.writePathToFile(
			Main.getProgramPathFile(),
			file.getParentFile() );
	rv = true;
      }
      finally {
	IOUtil.closeSilently( writer );
	IOUtil.closeSilently( out );
      }
    }
    catch( IOException ex ) {
      UIUtil.showErrorMsg( this.mainFrm, null, ex );
    }
    return rv;
  }


  private void setTextChanged( boolean state )
  {
    if( state != this.textChanged ) {
      this.textChanged = state;
      this.mainFrm.fireFileStatusChanged( MainFrm.STATUS_SAVE_ACTION );
    }
  }


  private void showPopupMenu( MouseEvent e )
  {
    boolean state = UIUtil.hasSelection( this.textArea );
    this.popupCopy.setEnabled( state );
    this.popupCut.setEnabled( state );
    this.popupToUpper.setEnabled( state );
    this.popupMenu.show(
		e.getComponent(),
		e.getX(),
		e.getY() );
  }


  private void unselectText()
  {
    int pos = this.textArea.getCaretPosition();
    this.textArea.setSelectionStart( pos );
    this.textArea.setSelectionEnd( pos );
    this.textArea.requestFocus();
    this.unselectTimer.stop();
  }
}
