/*
 * (c) 2025 Jens Mueller
 *
 * JBASICODE
 *
 * Searching and replacing text in a JTextArea
 */

package jbasicode.ui;

import java.awt.Component;
import java.awt.Window;
import java.text.MessageFormat;
import java.util.concurrent.atomic.AtomicInteger;
import javax.swing.JTextArea;
import javax.swing.JOptionPane;
import jbasicode.Main;
import jbasicode.ui.UIUtil;


public class TextFinder
{
  private String  searchText;
  private String  internalSearchText;
  private String  replaceText;
  private boolean caseSensitive;


  public TextFinder()
  {
    this.searchText         = null;
    this.caseSensitive      = false;
    this.replaceText        = null;
    this.internalSearchText = null;
  }


  private TextFinder(
		String  searchText,
		boolean caseSensitive,
		String  replaceText )
  {
    setValues( searchText, caseSensitive, replaceText );
  }


  public boolean findNext( JTextArea textArea )
  {
    return find( textArea, -1, false, true );
  }


  public boolean findPrev( JTextArea textArea )
  {
    return find( textArea, -1, true, true );
  }


  public String getReplaceText()
  {
    return this.replaceText;
  }


  public boolean hasReplaceText()
  {
    return (this.replaceText != null) && !this.replaceText.isEmpty();
  }


  public boolean hasSearchText()
  {
    return (this.searchText != null) && !this.searchText.isEmpty();
  }


  public void openFindDlg( JTextArea textArea )
  {
    FindAndReplaceDlg dlg = FindAndReplaceDlg.createFindDlg(
					UIUtil.getWindow( textArea ),
					getPresetSearchText( textArea ) );
    dlg.setVisible( true );

    String searchText = dlg.getSearchText();
    if( searchText != null ) {
      if( !searchText.isEmpty() ) {
	setValues( searchText, false, null );
	find( textArea, -1, false, true );
      }
    }
  }


  public void openFindAndReplaceDlg( JTextArea textArea )
  {
    Window window = UIUtil.getWindow( textArea );
    if( window != null ) {
      FindAndReplaceDlg dlg = FindAndReplaceDlg.createFindAndReplaceDlg(
				window,
				getPresetSearchText( textArea ),
				this.caseSensitive,
				this.replaceText );
      dlg.setVisible( true );

      String searchText = dlg.getSearchText();
      if( searchText != null ) {
	if( !searchText.isEmpty() ) {
	  switch( dlg.getAction() ) {
	    case FIND_NEXT:
	      setValues(
			searchText,
			dlg.getCaseSensitive(),
			dlg.getReplaceText() );
	      find( textArea, -1, false, true );
	      break;

	    case REPLACE_ALL:
	      setValues(
			searchText,
			dlg.getCaseSensitive(),
			dlg.getReplaceText() );
	      replaceAll(
			textArea,
			searchText,
			dlg.getCaseSensitive(),
			dlg.getReplaceText() );
	      break;
	  }
	}
      }
    }
  }


	/* --- Private Methods --- */

  /*
   * Actual search:
   * If startPos < 0 the search continues at the selected text
   * or at the current cursor position.
   */
  private boolean find(
			JTextArea textArea,
			int       startPos,
			boolean   backward,
			boolean   interactive )
  {
    boolean rv = false;
    if( this.internalSearchText != null ) {
      if( !this.internalSearchText.isEmpty() ) {
	String  baseText = textArea.getText();
	if( baseText == null ) {
	  baseText = "";
	}
	if( !this.caseSensitive ) {
	  baseText = baseText.toUpperCase();
	}
	if( startPos < 0 ) {
	  startPos = textArea.getCaretPosition();
	}
	int len      = baseText.length();
	int foundPos = -1;
	if( backward ) {
	  if( UIUtil.hasSelection( textArea ) ) {
	    startPos = Math.min(
				textArea.getSelectionStart(),
				textArea.getSelectionEnd() );
	  }
	  --startPos;
	  for( int i = 0; i < 2; i++ ) {
	    if( startPos < 1 ) {
	      startPos = len - 1;
	    } else if( startPos > (len - 1) ) {
	      startPos = len - 1;
	    }
	    if( startPos < 0 ) {
	      startPos = 0;
	    }
	    foundPos = baseText.lastIndexOf(
				this.internalSearchText,
				startPos );
	    if( foundPos >= 0 ) {
	      break;
	    }
	    startPos = len - 1;
	  }
	} else {
	  if( UIUtil.hasSelection( textArea ) ) {
	    startPos = Math.max(
				textArea.getSelectionStart(),
				textArea.getSelectionEnd() );
	  }
	  for( int i = 0; i < 2; i++ ) {
	    if( (startPos < 0) || (startPos >= len) ) {
	      startPos = 0;
	    }
	    foundPos = baseText.indexOf(
				this.internalSearchText,
				startPos );
	    if( foundPos >= 0 ) {
	      break;
	    }
	    startPos = 0;
	  }
	}
	if( (foundPos >= 0) && (foundPos < len) ) {
	  Window window = UIUtil.getWindow( textArea );
	  if( window != null ) {
	    window.toFront();
	  }
	  textArea.requestFocus();
	  if( interactive ) {
	    textArea.setCaretPosition( foundPos );
	  }
	  textArea.select( foundPos, foundPos + this.searchText.length() );
	  rv = true;
	} else {
	  if( interactive ) {
	    showTextNotFound( textArea );
	  }
	}
      }
    }
    return rv;
  }


  private String getPresetSearchText( JTextArea textArea )
  {
    String text = textArea.getSelectedText();
    if( text != null ) {
      if( text.isEmpty() ) {
	text = null;
      }
    }
    return text != null ? text : this.searchText;
  }


  private void replaceAll(
			JTextArea textArea,
			String    searchText,
			boolean   caseSensitive,
			String    replaceText )
  {
    int    nReplaced = 0;
    String baseText  = textArea.getText();
    if( baseText != null ) {
      String pattern = searchText;
      if( !caseSensitive ) {
	pattern = searchText.toUpperCase();
      }
      int patternLen = pattern.length();
      int baseLen    = baseText.length();
      if( (patternLen > 0) && (baseLen > 0) ) {
	StringBuilder buf = new StringBuilder( baseLen );
	int baseIdx = 0;
	while( baseIdx < baseLen  ) {
	  boolean found = true;
	  if( (baseIdx + patternLen) > baseLen ) {
	    found = false;
	  } else {
	    for( int i = 0; i < patternLen; i++ ) {
	      char ch = baseText.charAt( baseIdx + i );
	      if( !caseSensitive ) {
		ch = Character.toUpperCase( ch );
	      }
	      if( pattern.charAt( i ) != ch ) {
		found = false;
		break;
	      }
	    }
	  }
	  if( found ) {
	    buf.append( replaceText );
	    baseIdx += patternLen;
	    nReplaced++;
	  } else {
	    buf.append( baseText.charAt( baseIdx ) );
	    baseIdx++;
	  }
	}
        if( nReplaced > 0 ) {
	  textArea.setText( buf.toString() );
	  try {
	    textArea.setCaretPosition( 0 );
	  }
	  catch( IllegalArgumentException ex ) {}
	  JOptionPane.showMessageDialog(
		UIUtil.getWindow( textArea ),
		nReplaced == 1 ?
			Main.getText( "msg.one_text_replacement" )
			: MessageFormat.format(
				Main.getText( "msg.n_text_replacements" ),
				String.valueOf( nReplaced ) ),
		Main.getText( "title.replace_text" ),
		JOptionPane.INFORMATION_MESSAGE );
	} else {
	  showTextNotFound( textArea );
	}
      }
    }
  }


  private void setValues(
		String  searchText,
		boolean caseSensitive,
		String  replaceText )
  {
    this.searchText         = searchText;
    this.caseSensitive      = caseSensitive;
    this.replaceText        = replaceText;
    this.internalSearchText = searchText;
    if( !caseSensitive && (searchText != null) ) {
      this.internalSearchText = searchText.toUpperCase();
    }
  }


  private static void showInfoDlg(
			Component owner,
			String    msg,
			String    title )
  {
    JOptionPane.showMessageDialog(
			UIUtil.getWindow( owner ),
			msg,
			title,
			JOptionPane.INFORMATION_MESSAGE );
  }


  private static void showTextNotFound( Component owner )
  {
    JOptionPane.showMessageDialog(
			UIUtil.getWindow( owner ),
			Main.getText( "msg.text_not_found" ),
			Main.getText( "title.find_text" ),
			JOptionPane.INFORMATION_MESSAGE );
  }
}
