/*
 * (c) 2025 Jens Mueller
 *
 * JBASICODE
 *
 * Utility Functions for User Interface
 */

package jbasicode.ui;

import java.awt.Component;
import java.awt.Dialog;
import java.awt.EventQueue;
import java.awt.Frame;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.Image;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.event.InputEvent;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.print.PrinterException;
import java.awt.print.PrinterJob;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Properties;
import javax.print.attribute.HashPrintRequestAttributeSet;
import javax.print.attribute.PrintRequestAttributeSet;
import javax.print.attribute.standard.Copies;
import javax.print.attribute.standard.JobName;
import javax.print.attribute.standard.MediaSizeName;
import javax.print.attribute.standard.OrientationRequested;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JTextArea;
import javax.swing.KeyStroke;
import javax.swing.filechooser.FileFilter;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.swing.text.JTextComponent;
import jbasicode.Main;


public class UIUtil
{
  private static final String PROP_WINDOW_X      = "window.x";
  private static final String PROP_WINDOW_Y      = "window.y";
  private static final String PROP_WINDOW_WIDTH  = "window.width";
  private static final String PROP_WINDOW_HEIGHT = "window.height";

  private static final String[] iconImageResources = {
					"/images/app/jbasicode_20x20.png",
					"/images/app/jbasicode_24x24.png",
					"/images/app/jbasicode_32x32.png",
					"/images/app/jbasicode_48x48.png" };

  private static Image                    largestIconImage = null;
  private static List<Image>              iconImages       = null;
  private static PrintRequestAttributeSet printAttrs       = null;

  private static Method menuShortcutKeyMaskMethod = null;
  private static Method modelToViewMethod         = null;
  private static Method viewToModelMethod         = null;


  public static JButton createImageButton(
				Component c,
				String    resource,
				String    text )
  {
    JButton btn = null;
    Image   img = null;
    URL     url = Main.class.getResource( resource );
    if( url != null ) {
      img = c.getToolkit().getImage( url );
    }
    if( img != null ) {
      btn = new JButton( new ImageIcon( img ) );
      if( !text.isEmpty() ) {
	btn.setToolTipText( text );
      }
    } else {
      btn = new JButton( text );
    }
    return btn;
  }


  public static void doPrint( JTextArea textArea, String fileName )
  {
    try {
      PrintRequestAttributeSet attrs = getPrintRequestAttributeSet();
      attrs.add( new Copies( 1 ) );
      String jobName = fileName;
      if( jobName == null ) {
	jobName = Main.APPNAME;
      }
      attrs.add( new JobName( jobName, Locale.getDefault() ) );
      attrs.add( MediaSizeName.ISO_A4 );
      attrs.add( OrientationRequested.PORTRAIT );

      PrinterJob pj = PrinterJob.getPrinterJob();
      pj.setCopies( 1 );
      pj.setJobName( jobName );
      if( pj.printDialog( attrs ) ) {
	pj.setPrintable( new PlainTextPrintable( textArea, fileName ) );
	pj.print( attrs );
      }
    }
    catch( PrinterException ex ) {
      showErrorMsg( textArea, null, ex );
    }
  }


  public static void fireShowErrorMsg( Component owner, String msg )
  {
    fireShowErrorMsg( owner, msg, null );
  }


  public static void fireShowErrorMsg(
				Component owner,
				String    msg,
				Exception ex )
  {
    EventQueue.invokeLater( ()->showErrorMsg( owner, msg, ex ) );
  }


  public static Image getLargestIconImage()
  {
    return largestIconImage;
  }


  public static Window getWindow( Component c )
  {
    while( c != null ) {
      if( c instanceof Window ) {
        return (Window) c;
      }
      c = c.getParent();
    }
    return null;
  }


  public static void getWindowBounds(
				Window     window,
				Properties props,
				String     propPrefix )
  {
    getWindowBounds(
		window.getBounds(),
		isResizable( window ),
		props,
		propPrefix );
  }


  public static void getWindowBounds(
				Rectangle  windowBounds,
				boolean    resizable,
				Properties props,
				String     propPrefix )
  {
    if( windowBounds != null ) {
      props.setProperty(
		propPrefix + PROP_WINDOW_X,
		String.valueOf( windowBounds.x ) );
      props.setProperty(
		propPrefix + PROP_WINDOW_Y,
		String.valueOf( windowBounds.y ) );
      if( resizable ) {
	props.setProperty(
		propPrefix + PROP_WINDOW_WIDTH,
		String.valueOf( windowBounds.width ) );
	props.setProperty(
		propPrefix + PROP_WINDOW_HEIGHT,
		String.valueOf( windowBounds.height ) );
      }
    }
  }


  public static boolean hasSelection( JTextComponent textComp )
  {
    boolean rv = false;
    if( textComp != null ) {
      int e = textComp.getSelectionEnd();
      rv    = (textComp.getSelectionStart() < e) && (e > 0);
    }
    return rv;
  }


  /*
   * Since Java 9 the method JTextComponent#modelToView(...)
   * has been replaced by JTextComponent#modelToView2D(...).
   * For this reason, the functionality is encapsulated in a wrapper method.
   */
  public static Rectangle2D modelToView( JTextComponent textComp, int pos )
  {
    Rectangle2D rv = null;
    if( (textComp != null) && (pos >= 0) ) {
      if( modelToViewMethod == null ) {
	try {
	  modelToViewMethod = textComp.getClass().getMethod(
							"modelToView2D",
							int.class );
	  if( !isPublic( modelToViewMethod ) ) {
	    modelToViewMethod = null;
	  }
	}
	catch( NoSuchMethodException ex ) {}
      }
      if( modelToViewMethod == null ) {
	try {
	  modelToViewMethod = textComp.getClass().getMethod(
							"modelToView",
							int.class );
	  if( !isPublic( modelToViewMethod ) ) {
	    modelToViewMethod = null;
	  }
	}
	catch( NoSuchMethodException ex ) {}
      }
      if( modelToViewMethod != null ) {
	try {
	  Object obj = modelToViewMethod.invoke( textComp, pos );
	  if( obj != null ) {
	    if( obj instanceof Rectangle2D ) {
	      rv = (Rectangle2D) obj;
	    }
	  }
	}
	catch( Exception ex ) {}
      }
    }
    return rv;
  }


  // This method is called in the BCRuntime thread
  public static File openDataFileDlg(
				MainFrm mainFrm,
				boolean forSave,
				File    dirFile,
				String  fileName )
						throws InterruptedException
  {
    StringBuilder outFile = new StringBuilder();
    synchronized( outFile ) {
      EventQueue.invokeLater(
		()->showDataFileDlg(
				mainFrm,
				forSave,
				dirFile,
				fileName,
				outFile ) );
      try {
	outFile.wait();
      }
      catch( IllegalMonitorStateException ex ) {}
    }
    return outFile.length() > 0 ? new File( outFile.toString() ) : null;
  }


  public static boolean restoreWindowBounds(
				Window     window,
				Properties props,
				String     propPrefix )
  {
    boolean rv = false;
    if( props != null ) {
      try {
	int x = getInt( props, propPrefix + PROP_WINDOW_X );
	int y = getInt( props, propPrefix + PROP_WINDOW_Y );
	int w = 0;
	int h = 0;
	if( isResizable( window ) ) {
	  w = getInt( props, propPrefix + PROP_WINDOW_WIDTH );
	  h = getInt( props, propPrefix + PROP_WINDOW_HEIGHT );
	  if( (w > -10) && (h > -10) ) {
	    window.setSize( w, h );
	  }
	} else {
	  w = window.getWidth();
	  h = window.getHeight();
	}

	// check whether upper left corner of window is visible
	boolean visible = false;
	w /= 4;
	h /= 4;
	for( GraphicsDevice gd : GraphicsEnvironment
					.getLocalGraphicsEnvironment()
					.getScreenDevices() )
	{
	  for( GraphicsConfiguration gc : gd.getConfigurations() ) {
	    Rectangle r = gc.getBounds();
	    if( r != null ) {
	      if( r.contains( x + 10, y + 10, w, h ) ) {
		visible = true;
		break;
	      }
	    }
	  }
	  if( visible ) {
	    break;
	  }
	}
	if( visible ) {
	  window.setLocation( x, y );
	  rv = true;
	}
      }
      catch( NullPointerException ex ) {}
      catch( NumberFormatException ex ) {}
    }
    return rv;
  }


  public static void setIconImagesAt( Window window )
  {
    if( iconImages == null ) {
      iconImages = new ArrayList<>();
      for( String resource : iconImageResources ) {
	try {
	  URL url = window.getClass().getResource( resource );
	  if( url != null ) {
	    Image image = window.getToolkit().createImage( url );
	    if( image != null ) {
	      largestIconImage = image;
	      iconImages.add( image );
	    }
	  }
	}
	catch( Exception ex ) {}
      }
    }
    if( !iconImages.isEmpty() ) {
      window.setIconImages( iconImages );
    }
  }


  public static void setShortcutAt(
				JMenuItem item,
				int       keyCode,
				boolean   withShift )
  {
    int modifiers = getMenuShortcutKeyMask( item );
    if( modifiers != 0 ) {
      if( withShift ) {
	modifiers |= InputEvent.SHIFT_DOWN_MASK;
      }
      item.setAccelerator( KeyStroke.getKeyStroke( keyCode, modifiers ) );
    }
  }


  public static void setParentCentered( Component c )
  {
    Component p = c.getParent();
    if( p != null ) {
      int x = p.getX() + ((p.getWidth() - c.getWidth()) / 2);
      int y = p.getY() + ((p.getHeight() - c.getHeight()) / 2);
      c.setLocation( x > 0 ? x : 0, y > 0 ? y : 0 );
    }
  }


  public static void showErrorMsg(
				Component owner,
				String    msg,
				Exception ex )
  {
    String exMsg = null;
    if( ex != null ) {
      exMsg = ex.getMessage();
      if( exMsg != null ) {
	if( exMsg.isEmpty() ) {
	  exMsg = null;
	}
      }
      if( ex instanceof IOException ) {
	if( (msg == null) && (exMsg == null) ) {
	  msg = Main.getText( "error.io" );
	}
      } else if( ex instanceof RuntimeException ) {
	if( exMsg == null ) {
	  exMsg = ex.getClass().getName();
	}
	if( msg == null ) {
	  msg = Main.getText( "error.internal" );
	}
      }
    }
    if( msg != null ) {
      if( exMsg != null ) {
	msg = msg + ":\n\n" + exMsg;
      }
    } else if( exMsg != null ) {
      msg = exMsg;
    }
    showErrorMsg( owner, msg );
  }


  public static void showErrorMsg( Component owner, String msg )
  {
    // Error message should be shown window centered
    while( owner != null ) {
      if( owner instanceof Window ) {
	break;
      }
      owner = owner.getParent();
    }
    JOptionPane.showMessageDialog(
		getWindow( owner ),
		msg != null ? msg : Main.getText( "error.unknown" ),
		Main.getText( "title.error" ),
		JOptionPane.ERROR_MESSAGE );
  }


  public static File showOpenFileDlg(
			Component  owner,
			String     title,
			File       dirFile,
			String     fileName,
			FileFilter fileFilter )
  {
    JFileChooser fc = createJFileChooser(
					title,
					dirFile,
					fileName,
					fileFilter );
    return fc.showOpenDialog( getWindow( owner ) )
				== JFileChooser.APPROVE_OPTION ?
						fc.getSelectedFile()
						: null;
  }


  public static File showSaveFileDlg(
			Component        owner,
			String           title,
			File             dirFile,
			String           fileName,
			FileFilter       fileFilter )
  {
    File         rv = null;
    JFileChooser fc = createJFileChooser(
					title,
					dirFile,
					fileName,
					fileFilter );
    if( fc.showDialog(
		getWindow( owner ),
		Main.getText( "action.file.save" ) )
				== JFileChooser.APPROVE_OPTION )
    {
      File file = fc.getSelectedFile();
      if( file != null ) {

	// ggf. fehlende Dateiendung auomatisch hinzufuegen
	FileFilter fFilter = fc.getFileFilter();
	String     fName   = file.getName();
	if( (fFilter != null) && (fName != null) ) {
	  if( (fFilter instanceof FileNameExtensionFilter)
	      && (fName.indexOf( '.' ) < 0) )
	  {
	    String[] e = ((FileNameExtensionFilter) fFilter).getExtensions();
	    if( e != null ) {
	      if( e.length == 1 ) {
		fName  = fName + '.' + e[ 0 ].toLowerCase();
		File d = file.getParentFile();
		if( d != null ) {
		  file = new File( d, fName );
		} else {
		  file = new File( fName );
		}
	      }
	    }
	  }
	}

	// Test auf Vorhandensein der Datei
	boolean status = true;
	if( file.exists() ) {
	  StringBuilder buf = new StringBuilder( 256 );
	  if( fName != null ) {
	    buf.append( fName );
	    buf.append( ":\n" );
	  }
	  buf.append( Main.getText( "confirm.override_existing_file" ) );
	  if( JOptionPane.showConfirmDialog(
				getWindow( owner ),
				buf.toString(),
				Main.getText( "title.confirm" ),
				JOptionPane.YES_NO_OPTION,
				JOptionPane.WARNING_MESSAGE )
					== JOptionPane.YES_OPTION )
	  {
	    status = true;
	  }
	}
	if( status ) {
	  rv = file;
	}
      }
    }
    return rv;
  }


  /*
   * Since Java 9 the method JTextComponent#viewToModel(...)
   * has been replaced by JTextComponent#viewToModel2D(...).
   * For this reason, the functionality is encapsulated in a wrapper method.
   */
  public static int viewToModel( JTextComponent textComp, Point point )
  {
    int rv = -1;
    if( (textComp != null) && (point != null) ) {
      if( viewToModelMethod == null ) {
	try {
	  viewToModelMethod = textComp.getClass().getMethod(
							"viewToModel2D",
							Point2D.class );
	  if( !isPublic( viewToModelMethod ) ) {
	    viewToModelMethod = null;
	  }
	}
	catch( NoSuchMethodException ex ) {}
      }
      if( viewToModelMethod == null ) {
	try {
	  viewToModelMethod = textComp.getClass().getMethod(
							"viewToModel",
							Point.class );
	  if( !isPublic( viewToModelMethod ) ) {
	    viewToModelMethod = null;
	  }
	}
	catch( NoSuchMethodException ex ) {}
      }
      if( viewToModelMethod != null ) {
	try {
	  Object pos = viewToModelMethod.invoke( textComp, point );
	  if( pos != null ) {
	    if( pos instanceof Number ) {
	      rv = ((Number) pos).intValue();
	    }
	  }
	}
	catch( Exception ex ) {}
      }
    }
    return rv;
  }


	/* --- private Methoden --- */

  private static JFileChooser createJFileChooser(
					String     title,
					File       dirFile,
					String     fileName,
					FileFilter fileFilter )
  {
    JFileChooser fc = new JFileChooser();
    if( title != null ) {
      fc.setDialogTitle( title );
    }
    if( dirFile != null ) {
      if( !dirFile.isDirectory() ) {
	dirFile = dirFile.getParentFile();
      }
      if( dirFile != null ) {
	fc.setCurrentDirectory( dirFile );
      }
    }
    if( fileName != null ) {
      if( dirFile != null ) {
	fc.setSelectedFile( new File( dirFile, fileName ) );
      } else {
	fc.setSelectedFile( new File( fileName ) );
      }
    }
    if( fileFilter != null ) {
      fc.addChoosableFileFilter( fileFilter );
      fc.setFileFilter( fileFilter );
    }
    return fc;
  }


  private static int getInt( Properties props, String keyName )
  {
    if( props == null ) {
      throw new NullPointerException();
    }
    String value = props.getProperty( keyName );
    if( value == null ) {
      throw new NullPointerException();
    }
    return Integer.parseInt( value );
  }


  /*
   * Since Java 10 the method Toolkiit#getMenuShortcutKeyMask(...)
   * has been replaced by Toolkit#getMenuShortcutKeyMaskEx(...).
   * For this reason, the functionality is encapsulated in a wrapper method.
   */
  private static int getMenuShortcutKeyMask( Component c )
  {
    int     rv = 0;
    Toolkit tk = c.getToolkit();
    if( tk != null ) {
      if( menuShortcutKeyMaskMethod == null ) {
	try {
	  menuShortcutKeyMaskMethod = tk.getClass().getMethod(
						"getMenuShortcutKeyMaskEx" );
	  if( !isPublic( menuShortcutKeyMaskMethod ) ) {
	    menuShortcutKeyMaskMethod = null;
	  }
	}
	catch( NoSuchMethodException ex ) {}
      }
      if( menuShortcutKeyMaskMethod == null ) {
	try {
	  menuShortcutKeyMaskMethod = tk.getClass().getMethod(
						"getMenuShortcutKeyMask" );
	  if( !isPublic( menuShortcutKeyMaskMethod ) ) {
	    menuShortcutKeyMaskMethod = null;
	  }
	}
	catch( NoSuchMethodException ex ) {}
      }
      if( menuShortcutKeyMaskMethod != null ) {
	try {
	  Object obj = menuShortcutKeyMaskMethod.invoke( tk );
	  if( obj != null ) {
	    if( obj instanceof Integer ) {
	      rv = ((Integer) obj).intValue();
	    }
	  }
	}
	catch( Exception ex ) {}
      }
    }
    return rv;
  }


  private static PrintRequestAttributeSet getPrintRequestAttributeSet()
  {
    if( printAttrs == null ) {
      printAttrs = new HashPrintRequestAttributeSet();
    }
    return printAttrs;
  }


  private static boolean isPublic( Method method )
  {
    boolean rv = false;
    if( method != null ) {
      if( (method.getModifiers() & Modifier.PUBLIC) != 0 ) {
	rv = true;
      }
    }
    return rv;
  }


  private static boolean isResizable( Window window )
  {
    boolean rv = false;
    if( window instanceof Dialog ) {
      rv = ((Dialog) window).isResizable();
    } if( window instanceof Frame ) {
      rv = ((Frame) window).isResizable();
    }
    return rv;
  }


  private static void showDataFileDlg(
			MainFrm       mainFrm,
			boolean       forSave,
			File          dirFile,
			String        fileName,
			StringBuilder outFile )
  {
    File file = null;
    if( forSave ) {
      file = showSaveFileDlg(
			mainFrm,
			Main.getText( "title.save_data_file" ),
			dirFile,
			fileName,
			null );
    } else {
      file = showOpenFileDlg(
			mainFrm,
			Main.getText( "title.open_data_file" ),
			dirFile,
			fileName,
			null );
    }
    synchronized( outFile ) {
      if( file != null ) {
	mainFrm.setRecentDataFile( file );
	outFile.append( file.getPath() );
      }
      try {
	outFile.notifyAll();
      }
      catch( IllegalMonitorStateException ex ) {}
    }
  }


	/* --- Constructor --- */

  private UIUtil()
  {
    // empty
  }
}
